#######################################################################
# Copyright: DAASI International GmbH 2018. All rights reserved.
#
# This is Open Source Software
# License: Artistic License 2.0
#
# Author: Peter Gietz, DAASI International GmbH, www.daasi.de
# For questions please mail to peter.gietz@daasi.de
#######################################################################
#!/usr/bin/perl -w


package DAASIlib::ACL;

use strict;
use warnings;
use vars qw($VERSION);
our $VERSION = "0.1";

use Log::Log4perl qw(:levels);
my $logger = Log::Log4perl->get_logger(''); 
use DAASIlib::Data;
my $data = new DAASIlib::Data;


my @tokens = qw ( userbase groupbase userattr groupattr user_dn 
		  searchpermission nosearchpermissionstext);

my @fields = qw ( who tree scope attr perm );

#my @permissions = qw ( read write search entryadd );
my @permissions = qw ( search read write entryadd );
#my @permissions = qw ( entryadd write read search );



sub new {
    my $class;
    my $self;

    my ( $this , $param ) = @_;
    $class = ref($this) || $this;

    $self={
	data => {},
    };

    bless($self , $class );

    return $self;

}

sub readConfig {
    my $self = shift;
    my ($filename) = @_;

    open FH, "<$filename";

    $self->{data}{acls} = {};
    $self->{data}{aclcount} = 0;
    $self->{data}{linecount} = 0;
    $self->{data}{groupcount} = 0;
    $self->{data}{addedgroups} = 0;

    while ( <FH> ) {
	$self->{data}{linecount}++;
	s/^\s*//;
	s/\s*$//;

	if ( /^\#/ || /^\s*$/) {
	    next;
	}
	if ( /\#/ ) {
	    $self->parseLine($_);
	}
	elsif ( /\=/ ) {
	    my ( $token, $value ) = split /\s*=\s*/, $_, 2;

	    $value =~ s/^\"//;
	    $value =~ s/\"$//;

	    my $hit = grep($token, @tokens);
	    if ( $hit ) {
		$self->{data}{$token}=$value;
	    } 
	    else {
		$logger->error("acl::readConfig(): wrong token $token in configfile $filename");
	    }
	} 
	else {
	    $logger->error("acl::readConfig(): syntax error in in configfile $filename, line: $_");
	}
    }
    close FH;


}

sub parseLine {
    my $self = shift;
    my ( $line ) = @_;

#    $logger->info("ACL::parseLine()");
    $self->{data}{aclcount}++;
    my $num = $self->{data}{aclcount};
    my ($who, $tree, $scope, $attrs, $perm ) = split /\s*\#\s*/, $line;
   
    $self->setWho($who, $num);
    $self->{data}{acls}{$num}{'tree'}=$tree;
    $self->{data}{acls}{$num}{'scope'}=$scope;
    $self->{data}{acls}{$num}{'attrs'}=$attrs;
    $self->{data}{acls}{$num}{'perm'}=$perm;

}

sub setWho {
    my $self = shift;
    my ( $who, $num ) = @_;

    if ( $who =~ /:/ ) {
	my ($type,$val) = split /:/, $who, 2;
	if ( $val eq '*') {
	    $who = $val;
	}
	elsif ($type eq 'u') {
	    if (  $self->{data}{userattr} &&  $self->{data}{userbase}) {
		$who = $self->{data}{userattr}.'='.$val.','.$self->{data}{userbase};
	    }
	    elsif ( $self->{data}{userbase} ) {
		$who = $val.','.$self->{data}{userbase};
	    }
	    else {
		$who = $val;
	    }
	}
	elsif ($type eq 'g') {
	    if (  $self->{data}{groupattr} &&  $self->{data}{groupbase}) {
		$who = $self->{data}{groupattr}.'='.$val.','.$self->{data}{groupbase};
	    }
	    elsif ( $self->{data}{groupbase} ) {
		$who = $val.','.$self->{data}{groupbase};
	    }
	    else {
		$who = $val;
	    }
	    $self->{data}{groupcount}++;
	}
	else {
	    $logger->error("wrong type in definition of who ($who)");
	    return;
	}

	$self->{data}{acls}{$num}{'whotype'}=$type;
    }
    else {
	$self->{data}{acls}{$num}{'whotype'}='?';
    }

    $self->{data}{acls}{$num}{'who'}=$who;

}

sub setLdap {
    my $self = shift;
    my ( $ldap, $user_dn ) = @_;

    $logger->debug("**********  setting ldapserver");

    $self->{data}{ldap} = $ldap;
    $self->{data}{user_dn} = $user_dn;
}

sub showAcls {
    my $self = shift;

    $logger->info("ACL::showAcls()");
    foreach my $token ( @tokens ) {
	if ( $self->{data}{$token} ) {
	    $logger->info("Token $token:".$self->{data}{$token});
	}
    }
    
    if ( $self->{data}{usergroups} ) {
	foreach my $group ( @{$self->{data}{usergroups}} ) {
	    $logger->info("active user belongs to group: $group");
	}
    }
    
    foreach my $num ( sort keys %{$self->{data}{acls}} ) {
	$logger->info("$num, wtype: ".$self->{data}{acls}{$num}{'whotype'});
	$logger->info("$num,   who: ".$self->{data}{acls}{$num}{'who'});
	$logger->info("$num,  tree: ".$self->{data}{acls}{$num}{'tree'});
	$logger->info("$num, scope: ".$self->{data}{acls}{$num}{'scope'});
	$logger->info("$num, attrs: ".$self->{data}{acls}{$num}{'attrs'});
	$logger->info("$num,  perm: ".$self->{data}{acls}{$num}{'perm'});
    }
}

sub isAccess {
    my $self = shift;
    my ( $user, $dn, $attr, $perm ) = @_;

    if ( ! $self->{data}{addedgroups} ) {
	$self->addLdapGroups();
#	$self->showAcls();
    }

    $logger->debug("isAccess( user: $user, dn: $dn, attr: $attr, perm: $perm)");

#    $self->{data}{goodacls} = ()
    my @whatgood = $self->findWhat(lc($dn));
    $logger->debug("acls with right what-part: @whatgood ($dn)");

    my @whogood = $self->findWho($user, \@whatgood);
    $logger->debug("acls with right what- and who-part: @whogood ($user)");

    my @attrgood = $self->findAttr($attr, \@whogood);
    $logger->debug("acls with right what- and who- and attrpart: @attrgood ($attr)");

    my @permgood = $self->findPerm($perm, \@attrgood);
    my $numok = $#permgood+1;
    $logger->debug("$numok acls allow permission (@permgood) ($perm)");

    
    if ( $numok ) {
	return 1;
    }
    else {
	return 0;
    }
}


sub findPerm {
    my $self = shift;
    my ( $wantperm, $ra_acls ) = @_;


    my @fittingacls = ();

    foreach my $num ( @{$ra_acls} ) {
	my $perm = $self->{data}{acls}{$num}{'perm'};
	$logger->debug("acl nr $num has perm: $perm");
	if ( $perm eq '*' ) {
	   push @fittingacls, $num; 
	}
	else {
	    my $found = 0;
	    my @includedperms = ();
	    foreach my $permission ( @permissions ) {
		$logger->debug("adding perm $permission to included");
		push @includedperms, $permission;
		if ( $perm eq $permission ) {
		    $logger->debug("found permission $perm");
		    $found = 1;
		    last;
		}
	    }
	    if ( $found ) {
		$logger->debug("permission $perm seems valid");
		foreach my $includedperm (@includedperms) {
		    if ( $wantperm eq $includedperm ) {
			$logger->debug("wanted perm $wantperm is included in acl");
			push @fittingacls, $num;
		    }
		}
	    }
	} 

    }

    return @fittingacls;
}


sub findAttr {
    my $self = shift;
    my ( $attribute, $ra_acls ) = @_;


    my @fittingacls = ();

    foreach my $num ( @{$ra_acls} ) {
	my $attrs = $self->{data}{acls}{$num}{'attrs'};
	if ( $attribute eq '*') {
	    push @fittingacls, $num;
	}
	if ( $attrs eq '*' ) {
	    push @fittingacls, $num;
	    $logger->debug("Found attr $attribute in acl with *");

	}
	else  {
	    my @attrs = split /\s*,\s*/, $attrs;
	    foreach my $attr ( @attrs ) {
		if ( lc($attribute) eq lc($attr) ) {
		    $logger->debug("Found attr $attribute in acl");
		    push @fittingacls, $num;
		    last;
		}
	    }
	}

    }

    return @fittingacls;
}

sub findWho {
    my $self = shift;
    my ( $user, $ra_acls ) = @_;


    my @fittingacls = ();

    if ( $self->{data}{userbase} 
	 && $self->{data}{userattr}
	 && $user !~ /,\w+=/ ) {
	$user = $self->{data}{userattr}."=".$user.",".$self->{data}{userbase};
#	$logger->info("changed user to $user");
    }

    foreach my $num ( @{$ra_acls} ) {
	my $type = $self->{data}{acls}{$num}{'whotype'};
	my $who = $self->{data}{acls}{$num}{'who'};
	if ( $who eq '*' ) {
	    push @fittingacls, $num;
	}
	elsif ( $type eq 'u' ) {
	    if ( $who eq $user ) {
		push @fittingacls, $num;
	    }
	    else {
#		$logger->info("no match for whopart $who with $user"); 
	    }
	}
	elsif ( $type eq 'g' ) {
	    foreach my $group ( @{$self->{data}{usergroups}} ) {
		if ( $who eq $group ) {
		    push @fittingacls, $num;
		    last;
		}
	    }
	    
	}
	else {
	    $logger->error("findWho(): $type is no allowed type");
	}

    }

    return @fittingacls;
}

sub findWhat {
    my $self = shift;
    my ( $dn ) = @_;

    my @fittingacls = ();

    foreach  my $num ( sort keys %{$self->{data}{acls}} ) {
	my $scope = $self->{data}{acls}{$num}{'scope'};
	my $tree = $self->{data}{acls}{$num}{'tree'};
	if ( $dn =~ /$tree$/ ) {
	    if ( $scope eq 'sub' ) {
#		$logger->info("found DN: $dn");
		push @fittingacls, $num;
	    }
	    elsif ( $scope eq 'one' ) {
		my $onelevel = $dn;
		$onelevel =~ s/\w*?=\w*,(.*)/$1/;
#		$logger->info("Cutt $dn to $onelevel (tree is $tree)");
		if ( $onelevel eq $tree ) {
		    push @fittingacls, $num;
		}
	    }
	    elsif ( $scope eq 'base' && $dn eq $tree ) {
		push @fittingacls, $num;
	    }
	    else {
#		$logger->info("didn't match DN $dn in tree $tree (scope: $scope)"); 
	    }
	}
	else {
#	   $logger->info("No $tree in $dn"); 
	}
    }

    return ( @fittingacls );
}


sub addLdapGroups {
    my $self = shift;

    $self->{data}{usergroups} = ();

    $logger->debug("**** adding groups of user");
    $logger->debug("**** ".$self->{data}{groupcount}." group(s) have been found");

    if ( $self->{data}{groupcount} ) {
	my $filter = "(".$self->{data}{memberattr}.'='
	    .$self->{data}{user_dn}. ")";
	
	my $mesg = $self->{data}{ldap}->search( filter => $filter,
						scope => "sub",
						base => $self->{data}{groupbase});

	if ($mesg->count) {
	    foreach my $entry ($mesg->entries ) {
		push @{$self->{data}{usergroups}}, $entry->dn;
	    }
	}
    }

    $self->{data}{addedgroups} = 1;

}


sub setLastAllowed () {
    my $self = shift;
    my ( $count ) = @_;

    $self->{data}{lastallowedcountisset} = 1;
    $self->{data}{lastallowedcount} = $count;

} 



sub getNoSearchPermissionsText () {
    my $self = shift;

    my $result = undef;
    if ( $self->{data}{nosearchpermissionstext} ) {
	$result = $self->{data}{nosearchpermissionstext};
    }
    else {
	$result = "NO results that you may see";
    }

    return $result;
}

sub getSearchPermission () {
    my $self = shift;

    my $result = undef;
    if ( $self->{data}{searchpermission} ) {
	$result = $self->{data}{searchpermission};
    }

    return $result;
}

sub getLastAllowed () {
    my $self = shift;

    my $result = undef;

    if ( $self->{data}{lastallowedcountisset} ) {

	$result = $self->{data}{lastallowedcount};
	$self->{data}{lastallowedcountisset} = 0;
    }
    return $result ;

} 




#	foreach my $num ( sort keys %{$self->{data}{acls}} ) {
#	    if ( $self->{data}{acls}{$num}{'whotype'} eq 'g') {


1;
