#######################################################################
# Copyright: DAASI International GmbH 2018. All rights reserved.
#
# This is Open Source Software
# License: Artistic License 2.0
#
# Author: Peter Gietz, DAASI International GmbH, www.daasi.de
# For questions please mail to peter.gietz@daasi.de
#######################################################################
package DAASIlib::Convert;


use 5.008001;
use strict;
use warnings;
use vars qw($VERSION);

$VERSION = "0.60";

use Log::Log4perl qw(:levels);
use Log::Log4perl::Level;

my $logger = Log::Log4perl->get_logger('');

use DAASIlib::DBI;
my $libdbi = new  DAASIlib::DBI;
use DAASIlib::LDAP;
my $libldap = new DAASIlib::LDAP;
use DAASIlib::Data;
my $data = new DAASIlib::Data;


use Text::Iconv;
#use Test::utf8;
use MIME::Base64;
use utf8;

use Encode;
use DateTime;
use URI::Escape;

# Preloaded methods go here.


sub new {
    my $class;
    my $self;

    my ( $this , $param ) = @_;
    $class = ref($this) || $this;

    $self={
	data => {},
	uniqueidlist => [],
    };

    bless($self , $class );

    return $self;
}


sub init {

    my $self = shift;
    my ($codesfile, $isrealutf8, $dictionaryfile) = @_;
    $logger->info("Init with codesfile $codesfile") if $codesfile;
    $self->{data}{'codetabisloaded'} = 0;
    $self->{data}{'utfcodetabisloaded'} = 0;
    $self->{data}{'codesfile'} = $codesfile;
    $self->{data}{'dictionaryfileisloaded'} = 0;
    $self->{data}{'abbreviationfileisloaded'} = 0;
    $self->{data}{'dictionaryfile'} = $dictionaryfile;
    $self->{data}{'isrealutf8'} = $isrealutf8;
    $self->loadiconvtable();

#    $self->{data}{is_UUID_module} = undef;
#    if ( ! defined( $is_UUID_module ) ) {
    eval ("use UUID;");
    if ( $@ ) { 
	$logger->warn("Couldn't load module UUID, Thus looking for uuidgen");
	$self->{data}{is_UUID_module} = 0;
    }
    else {
	$self->{data}{is_UUID_module} = 1;
    }
    
}

sub is_dictionary {
    my $self = shift;
    return $self->{data}{'dictionaryfile'};
    
}


sub loadiconvtable {
    my $self = shift;

    my @array =
	("UTF", "ISO", "HTML",
	 "8859_1", "8859_2", "8859_3", "8859_4", "8859_5", "8859_6",
	 "8859_7", "8859_8", "8859_9", 
	 "CP037", "CP038", "CP273", "CP274", "CP275", "CP278",
	 "CP280", "CP281", "CP282", "CP284", "CP285", "CP290",
	 "CP297", "CP367", "CP420", "CP423", "CP424", "CP437",
	 "CP500", "CP737", "CP775", "CP813", "CP819", "CP850",
	 "CP851", "CP852", "CP855", "CP856", "CP857", "CP860",
	 "CP861", "CP862", "CP863", "CP864", "CP865", "CP866",
	 "CP866NAV", "CP868", "CP869", "CP870", "CP871", "CP874",
	 "CP875", "CP880", "CP891", "CP903", "CP904", "CP905",
	 "CP912", "CP915", "CP916", "CP918", "CP920", "CP922",
	 "CP930", "CP932", "CP933", "CP935", "CP936", "CP937",
	 "CP939", "CP949", "CP950", "CP1004", "CP1026", "CP1046",
	 "CP1047", "CP1070", "CP1079", "CP1081", "CP1084", "CP1089",
	 "CP1124", "CP1125", "CP1129", "CP1132", "CP1133", "CP1160",
	 "CP1161", "CP1162", "CP1163", "CP1164", "CP1250", "CP1251", 
	 "CP1252", "CP1253", "CP1254", "CP1255", "CP1256", "CP1257",
	 "CP1258", "CP1361", "CP10007", "CPIBM861", 
	 "ISO-8859-1", "ISO-8859-2", "ISO-8859-3", "ISO-8859-4",
	 "ISO-8859-5", "ISO-8859-6", "ISO-8859-7", "ISO-8859-8",
	 "ISO-8859-9", "ISO-8859-10", "ISO-8859-11", "ISO-8859-13",
	 "ISO-8859-14", "ISO-8859-15", "ISO-8859-16",
	 "LATIN-9", "LATIN-GREEK-1", "LATIN-GREEK", "LATIN1",
	 "LATIN2", "LATIN3", "LATIN4", "LATIN5", "LATIN6", "LATIN7",
	 "LATIN8", "LATIN10", "LATINGREEK", "LATINGREEK1",
	 "T.61-8BIT", "T.61", "T.618BIT", 
	 "UNICODE", "UTF-7", "UTF-8", "UTF-16", "UTF-16BE", "UTF-16LE",
	 "UTF-32", "UTF-32BE", "UTF-32LE", "UTF7", "UTF8", "UTF16",
	 "UTF16BE", "UTF16LE", "UTF32", "UTF32BE", "UTF32LE",
	 );

    $self->{data}{iconvtable} = \@array;

}


sub dispatchfunction {
    my $self = shift;
    my ($string, $is_multival, $rh_functionconf) = @_;

#    $logger->debug("dispatcher loaded with: [$string], multival: $is_multival");
#    print "This is the dispatcher\n";
    my $conv;
    my $converted="";
    my $function;
    my @vals;

### not really faster:
#    return("") if ( $string =~ /&&([\w-]+)\(\)/ );

    if ($is_multival) {
	@vals = split ( / \$ /, $string );
    } else {
	$vals[0] = $string;
    }
    $string = "";

#    print "Vals: @vals\n";

    my $oristring;

    foreach my $val ( @vals) {
	if ($val =~ /&&([\w-]+)\((.*)\)/ ) {
	    $function = $1;
	    $oristring= $conv=$2;
#	    print "\$function: [$function]  \$converted: [$conv]\n";
	    if ( $conv ) {
		$conv= $self->dispatchfunction($conv, $is_multival, $rh_functionconf);
#		print "sending [$conv] to function $function\n";
		$logger->debug("sending [$conv] to function $function");
		$conv = $self->convert($function,$conv, $rh_functionconf->{$function});
#		print "got back: |$conv|\n";
		if ( $conv) {
		    $logger->debug("function $function returned: [$conv]");
		}
		else {
		    $logger->warn("call to ${function}($oristring) returned nothing!!");
		}
	    }
	} else {
	    #nomore function
	    $conv = $val;
	}   
	if ( $converted) {
	    $converted = $converted . " \$ ". $conv;
	} else {
	    $converted = $conv;
	}

    }
    
#   print "returning: [$converted]\n";
    return($converted);
   
}


sub convert {

    my $self = shift;
    my ( $mode, $text, $ra_paras) = @_;

    my $converted;
    my $is_html = 0;
    my $funcref;
    my ($from, $to);

    
#    print "convert: mode: $mode, text: $text\n";
#    use Data::Dumper;
#    print Dumper($ra_paras);
#   exit();
    

 #   print "This is convert\n";
    $logger->debug("convert got function $mode and text $text");
    $text = $self->redoMakeSafe($text);

    $logger->debug("convert after redomakesafe got function $mode and text $text");


    if ( $mode =~ /^[xX]/ ) {
#	print "internal function wanted: $mode\n";
	eval { $self->$mode() };
	if (! $@) {
#	    print "internal function $mode\n";
	    $logger->debug("convert now calls function $mode with text $text");
	    $converted = $self->$mode($text, $ra_paras);
#	    print "|$mode| got back with $converted\n";
	} else {
#	    print "Ignoring unknown internal function $mode(): $@\n";
	    $logger->error("Ignoring unknown internal function $mode(): $@");
	    $converted = $text;
	}
    } elsif ( $mode =~ /\-to\-/i ) {
#	print "iconv function wanted\n";
	$mode = uc($mode);
	($from, $to) = split (/-TO-/, $mode);
	
	if ( $from =~ /^HTML$/ ) {
	    $text = $self->html_to_utf8($text);
	    $from = "UTF-8"; 
	}
	
	if ( $to  =~ /^HTML$/ ) {
	    $is_html = 1;
	    $to = "UTF-8";
	}
	
	$from =~ s/^ISO$/ISO-8859-1/;
	$to =~ s/^ISO$/ISO-8859-1/;
	$from =~ s/^UTF$/UTF-8/;
	$to =~ s/^UTF$/UTF-8/;
	
	my $to_is_ok = 0;
	my $from_is_ok = 0;
	
	foreach my $coding (@{$self->{data}{iconvtable}}) {
	    if (!$to_is_ok && ( $to eq $coding ) ) {
		$to_is_ok = 1;
	    }
	    if (!$from_is_ok && ( $from eq $coding ) ) {
		$from_is_ok = 1;
	    }
	    if ($to_is_ok && $from_is_ok ) {
		last;
	    }
	} 

	if ( ! $to_is_ok || ! $from_is_ok ) {
### TODO LOG ERROR here
	    print "unknown Iconv encoding function wanted: $from -> $to\n";
	    print "To: $to_is_ok from: $from_is_ok\n";
	    $converted = $text;
	} else {
#	    print "converting $text from $from to $to\n";
	    my $converter = Text::Iconv->new($from, $to);
	    $converter->raise_error(1);
	    eval { $converted = $converter->convert($text) };
#	    $converted = $converter->convert($text); 

	    if ( $text && !$converted ) {
		if ($@) { 
		    $@ =~ s/(.*) at [\.\/\\].*/$1/;
		    $logger->error("### Error in Iconv ($from -> $to) of [$text]:  $@"); 
		} else {
		    $logger->error("### unknown Error in Iconv conversion ($from -> $to) of [$text]");
#### TODO log error here
		}
		$converted = $text;
	    }
	    
	    if ($is_html) {
		$converted = $self->utf8_to_html($converted);
	    }
	}
	
    } else {
	
### TODO LOG ERROR here
#	print "unknown function $mode [$from] -> [$to] wanted\n";
	$converted = $text;
    }

    return($converted);
}

sub functiongrep {
    my $self = shift;
    my ( $searchvalue, $searchfunction, $ra_values ) = @_;

#    use Data::Dumper;
#    print Dumper($ra_values);
#    exit();
    my $is_found = 0;

    foreach ( @{$ra_values} ) {
#	print "### processing value $_\n";
	if ( $searchfunction ) {
	    if ( $self->$searchfunction($_) eq $searchvalue ) {
		$is_found = 1;
		last;
	    }
	} else {
	    if ( $_ eq $searchvalue ) {
		$is_found = 1;
		last;
	    }
	}
    }

    return ($is_found);
}

sub xdoesexist {
    my $self = shift;
    my ( $string ) = @_;

    if ( $string) {
	return "TRUE";
    }
    return "FALSE";
}


sub xtrim {
    my $self = shift;
    my ( $string ) = @_;

#    print "xtrim with string [$string]\n";
    if ( $string) {
#	$string =~ s/^ *(.*) *$/$1/;
	$string =~ s/^\s*//;
	$string =~ s/\s*$//;
    }

    return ($string);
}

sub xdequote {
    my $self = shift;
    my ( $string ) = @_;

    if ( $string) {
	$string =~ s/^\s*\"//;
	$string =~ s/\"\s*$//;
    }

    return ($string);
}

sub xremovequotes {
    my $self = shift;
    my ( $string ) = @_;

    if ( $string) {
	$string =~ s/\"//g;
    }

    return ($string);
}

sub xcreateuid {
    my $self = shift;
    my ( $string ) = @_;

    
    if ($string) { 
## TODO LOG notify
#	print "creating uid from [$string]\n";
	$string = $self->xnormalize($string);
#	print "creating uid from normalized [$string]\n";
	my @cnparts = split ( / /, $string);
	my $uid =lc(substr($cnparts[0],0,1)).
	    lc(substr($cnparts[$#cnparts],0,4)).'01' ;
#    print "uid: $uid\n";
	return ($uid);
    } else {
	return ("");
    }
}

sub xcreaterandomstring {
    my $self = shift;
    my ( $string ) = @_;
    my $randstr='';
    
    if ($string) { 
	for ( my $i = 0; $i< length($string); $i++ ) {
	    $randstr .= $data->getrandchar();
	} 
	return $randstr;
    } else {
	return ("");
    }

}

sub xchecksn {
    my $self = shift;
    my ( $string ) = @_;


    if ($string) { 
	return $string;
    } else {
	return ("Anonym");
    }
}

sub xadd20k {
    my $self = shift;
    my ( $num ) = @_;

    if ( $num) {
	return (20000+$num);
    }
    else {
	return ("");
    }
}

sub xadd10g {
    my $self = shift;
    my ( $num ) = @_;

    if ( $num) {
	return (10000000+$num);
    }
    else {
	return ("");
    }
}

sub xboolreverse {
    my $self = shift;
    my ( $bool ) = @_;

    if ( $bool) {
	if ( $bool eq 'FALSE') {
	    return ('TRUE');
	}
	elsif ( $bool eq 'TRUE' ) {
	    return ('FALSE');
	}
	else {
	    $logger->error("a string supposed to be an LDAP bool, but contains neither TRUE nor FALSE, but: $bool");
	    return ('');
	}
    }
    else {
	return ('');
    }
}

sub xnum2bool {
    my $self = shift;
    my ( $num ) = @_;
    $logger->debug("xnum2bool called with value $num");
    my $retval='FALSE';
    if ( $num ) {
	$retval = 'TRUE';
    }
    return $retval;

}

sub xgerman2iso3361 {

    my $self = shift;
    my ( $string ) = @_;

#    print "xtrim with string [$string]\n";
    if ( $string) {
	no warnings;
	my %german2iso3361 = (
"Brit. Jungferninseln" => "AI",
"Afghanistan" => "AF",
"Ägypten" => "EG",
"Albanien" => "AL",
"Algerien" => "DZ",
"Andorra" => "AD",
"Angola" => "AO",
"Antigua und Barbuda" => "AG",
"Äquatorialguinea" => "GQ",
"Argentinien" => "AR",
"Armenien" => "AM",
"Aserbaidschan" => "AZ",
"Äthiopien" => "ET",
"Australien" => "AU",
"Bahamas" => "BS",
"Bahrain" => "BH",
"Bangladesch" => "BD",
"Barbados" => "BB",
"Belgien" => "BE",
"Belize" => "BZ",
"Benin(Dahome)" => "BJ",
"Bhutan" => "BT",
"Birma/Myanmar" => "MM",
"Bolivien" => "BO",
"Bosnien und Herzegowina" => "BA",
"Bosnien-Herzegowina" => "BA",
"Botswana" => "BW",
"Brasilien" => "BR",
"Brunei Darussalam" => "BN",
"Bulgarien" => "BG",
"Burkina Faso" => "BF",
"Burundi" => "BI",
"Chile" => "CL",
"China" => "CN",
"China (VR)(einschl.Tibet)" => "CN",
"Cookinseln" => "CK",
"Costa Rica" => "CR",
"Dänemark" => "DK",
"Deutschland" => "DE",
"Dominica (Winward Inseln)" => "DM",
"Dominikanische Republik" => "DO",
"Dschibuti" => "DJ",
"Ecuador" => "EC",
#"ehemalige Sowjetunion" => "",
"El Salvador" => "SV",
"Elfenbeinküste" => "CI",
#"Eritrea" => "",
"Eritrea" => "ER",
"Estland" => "EE",
"Fidschi-Inseln" => "FJ",
"Finnland" => "FI",
"Frankreich" => "FR",
"Gabun" => "GA",
"Gambia" => "GM",
"Georgien" => "GE",
"Ghana" => "GH",
"Gibraltar" => "GI",
"Grenada" => "GD",
"Grenadinen (St.Vincent)" => "GD",
"Griechenland" => "GR",
"Guatemala" => "GT",
"Guayana" => "GY",
"Guinea" => "GN",
"Guinea-Bissau" => "GW",
"Haiti" => "HT",
"Honduras" => "HN",
"Hongkong" => "HK",
"Indien" => "IN",
"Indonesien" => "ID",
"Irak" => "IQ",
"Iran,Islamische Republik" => "IR",
"Irland" => "IE",
"Island" => "IS",
"Israel" => "IL",
"Italien" => "IT",
"Jamaika" => "JM",
"Japan" => "JP",
"Jemen" => "YE",
"Jordanien" => "JO",
"Kambodscha" => "KH",
"Kamerun" => "CM",
"Kanada" => "CA",
"Kap Verde" => "CV",
"Kasachstan" => "KZ",
"Katar" => "QA",
"Kenia" => "KE",
"Kirgisien" => "KG",
"Kiribati" => "KI",
"Kolumbien" => "CO",
"Komoren" => "KM",
"Kongo, Volksrepublik" => "CG",
"Kongo (Dem.Republik)" => "CD",
"Korea, Republik" => "KR",
"Korea, Nord, Demokr. VR" => "KP",
"Kroatien" => "HR",
"Kuba" => "CU",
"Kuwait" => "KW",
"Laotische Dem.Volksrep." => "LA",
"Lesotho" => "LS",
"Lettland" => "LV",
"Libanon" => "LB",
"Liberia" => "LR",
"Libyen" => "LY",
"Liechtenstein" => "LI",
"Litauen" => "LT",
"Luxemburg" => "LU",
"Madagaskar" => "MG",
"Malawi" => "MW",
"Malaysia" => "MY",
"Malediven" => "MV",
"Mali" => "ML",
"Malta" => "MT",
"Marokko" => "MA",
"Marshall-Inseln" => "MH",
"Mauretanien" => "MR",
"Mauritius" => "MU",
"Mazedonien(ehem.jug.Rep.)" => "MK",
"Mazedonien" => "MK",
"Mexiko" => "MX",
"Mikronesien" => "FM",
"Moldawien" => "MD",
"Monaco" => "MC",
"Mongolei" => "MN",
"Mosambik" => "MZ",
"Niederl.Antillen" => "AN",
"Namibia" => "NA",
"Nauru" => "NR",
"Nepal" => "NP",
"Neuseeland" => "NZ",
"Nicaragua" => "NI",
"Niederlande" => "NL",
"Niger" => "NE",
"Nigeria" => "NG",
"Niue" => "NU",
"Norwegen" => "NO",
"Ohne Angabe" => "00",
"Oman" => "OM",
"Österreich" => "AT",
"Pakistan,Islamische Rep." => "PK",
"Pakistan" => "PK",
"Palau" => "PW",
"Panama" => "PA",
"Papua-Neuguinea" => "PG",
"Paraguay" => "PY",
"Peru" => "PE",
"Philippinen" => "PH",
"Pitcairn-Inseln" => "PN",
"Polen" => "PL",
"Portugal" => "PT",
"Ruanda" => "RW",
"Rumänien" => "RO",
"Russische Foederation" => "RU",
"Sao Tome und Principe" => "ST",
"Salomonen" => "SB",
"Sambia" => "ZM",
"San Marino" => "SM",
"Saudi-Arabien" => "SA",
"Schweden" => "SE",
"Schweiz" => "CH",
"Senegal" => "SN",
"Serbien und Montenegro" => "CS",
"Serbien u. Montenegro" => "CS",
"Seychellen" => "SC",
"Sierra Leone" => "SL",
"Simbabwe" => "ZW",
"Singapur" => "SG",
"Slowakei" => "SK",
"Slowenien" => "SI",
"Somalia" => "SO",
"Spanien" => "ES",
"Sri Lanka(Ceylon)" => "LK",
"St.Helena einschl.Ascens." => "SH",
"St.Kitts und Nevis" => "KN",
"Staatenlos" => "0",
"Südafrika" => "ZA",
"Sudan" => "SD",
"Surinam" => "SR",
"Swasiland" => "SZ",
"Arabische Republ.Syrien" => "SY",
"Tadschikistan" => "TJ",
"Taiwan" => "TW",
"Tansania" => "TZ",
"Thailand" => "TH",
"Thailand" => "TH",
"Timor" => "TL",
"Togo" => "TG",
"Tonga" => "TO",
"Trinidad und Tobago" => "TT",
"Tschad" => "TD",
"Tschechische Republik" => "CZ",
"Tschechei" => "CZ",
"Tunesien" => "TN",
"Türkei" => "TR",
"Turkmenien" => "TM",
"Tuvalu" => "TV",
"Uganda" => "UG",
"Ukraine" => "UA",
"Ungarn" => "HU",
"Ungeklärt" => "0",
"Uruguay" => "UY",
"USA" => "US",
"Usbekistan" => "ZU",
"Vanuatu" => "VU",
"Vatikanstadt" => "VA",
"Venezuela" => "VE",
"Vereinigte Arab. Emirate" => "AE",
"Vereinigtes Königreich" => "GB",
"Vietnam" => "VN",
"Weissrussland(Belarus)" => "BY",
"Weissrussland" => "BY",
"Westsamoa" => "WS",
"Zentralafrik. Republik" => "CF",
"Zypern" => "CY",
			      );	

	use warnings;
	my $iso = $german2iso3361{$string};

	if ( ! $iso ) {
	    my $converter = Text::Iconv->new('UTF-8' , 'ISO-8859-1');
#	    $converter->raise_error(1);
	    eval { $string = $converter->convert($string) };
	    if ($string) {
		$iso = $german2iso3361{$string};
	    }
	}

	return($iso);
    } else {
	return("");
    }

#	if ( $german2iso3361{$string} ) {
#	    return ( $german2iso3361{$string} );
#	} else {
#	    my $converter = Text::Iconv->new('UTF-8' , 'ISO-8859-1');
#	    $converter->raise_error(1);
#	    eval { $string = $converter->convert($string) };
#	    if ( $string) {
#		if ( $german2iso3361{$string} ) {
#		    return ( $german2iso3361{$string} );
#		} else {
#		    $logger->error("unknown countryname $string");
#		    return ("");
#		}
#	    } 
#	}
#    } else {
#	return("");
#    } 
}


sub xgerman2iso3361_old {

    my $self = shift;
    my ( $string ) = @_;

#    print "xtrim with string [$string]\n";
    if ( $string) {
	no warnings;
	my %german2iso3361 = (
			      "Afghanistan" => "AF",
			      "Albanien" => "AL",
			      "Armenien" => "AM",
			      "Arabische Republ.Syrien" => "SY",
			      "Arabische Republ. Syrien" => "SY",
			      "Arabische Republik Syrien" => "SY",
			      "Argentinien" => "AR",
			      "Ägypten" => "EG",
			      "\xc3\x84gypten" => "EG",
			      "Äthiopien" => "ET",
			      "\xc3\x84thiopien" => "ET",
			      "Benin(Dahome)" => "BJ",
			      "Benin" => "BJ",
			      "Bosnien und Herzegowina" => "BA",
			      "Bosnien-Herzegowina" => "BA",
			      "Bulgarien" => "BG",
			      "Burkina Faso" => "BF",
			      "Chile" => "CL",
			      "China (VR)(einschl.Tibet)" => "CN",
			      "Deutschland" => "DE",
			      "Ecoador" => "EC",
			      "Estland" => "EE",
			      "Frankreich" => "FR",
			      "Finnland" => "FI",
			      "Georgien" => "GE",
			      "Ghana" => "GH",
			      "Griechenland" => "GR",
			      "Indien" => "IN",
			      "Indonesien" => "ID",
			      "Irak" => "IQ",
			      "Iran,Islamische Republik" => "IR",
			      "Iran" => "IR",
			      "Irland" => "IE",
			      "Israel" => "IL",
			      "Italien" => "IT",
			      "Kambodscha" => "KH",
			      "Kamerun" => "CM",
			      "Kasachstan" => "KZ",
			      "Kenia" => "KE",
			      "Kirgisien" => "KG",
			      "Kolumbien" => "CO",
			      "Korea, Nord, Demokr. VR" => "KP",
			      "Korea, Republik" => "KR",
			      "Kroatien" => "HR",
			      "Kuba" => "CU",
			      "Lettland" => "LV",
			      "Libanon" => "LB",
			      "Litauen" => "LT",
			      "Malaysia" => "MY",
			      "Mauretanien" => "MR",
			      "Marokko" => "MA",
			      "Mazedonien" => "MK",
			      "Moldawien" => "MD",
			      "Mongolei" => "MN",
			      "Niederlande" => "NL",
			      "Ãsterreich" => "AT",
			      "Österreich" => "AT",
			      "\xc3\x96sterreich" => "AT",
			      "Pakistan,Islamische Rep." => "PK",
			      "Pakistan" => "PK",
			      "Paraguay" => "PY",
			      "Polen" => "PL",
			      "Ruanda" => "RW",
			      "RumÃ¤nien" => "RO",
			      "Rumänien" => "RO",
			      "Russische Foederation" => "RU",
			      "Schweiz" => "CH",
			      "Senegal" => "SN",
			      "Serbien und Montenegro" => "CS",
			      "Serbien" => "CS",
			      "Singapur" => "SG",
			      "Slowakai" => "SK",
			      "Slowenien" => "SI",
			      "Spanien" => "ES",
			      "Staatenlos" => "00",
			      "Syrien" => "SY",
			      "Sri Lanka(Ceylon)" => "LK",
			      "Thailand" => "TH",
			      "Tschechische Republik" => "CZ",
			      "Tunesien" => "TN",
			      "TÃ¼rkei" => "TR",
			      "Türkei" => "TR",
			      "Turkmenien" => "TM",
			      "Ukraine" => "UA",
			      "Ungarn" => "HU",
			      "Usbekistan" => "UZ",
			      "Vereinigtes KÃ¶nigreich" => "GB",
			      "Vereinigtes Königreich" => "GB",
			      "Vietnam" => "VN",
			      "Weissrussland(Belarus)" => "BY",
			      "Weissrussland" => "BY",
		      );
	use warnings;
	if ( $german2iso3361{$string} ) {
	    return ( $german2iso3361{$string} );
	} else {
	    $logger->error("unknown countryname $string");
	    return ("");
	}
    } else {
	return("");
    } 
}

sub searchValue {
    my $self = shift;
    my ( $mapvalue, $uri, $attr, $mapattr, $searchtype, $pwtoken ) = @_;
    my $searchedvalue="";

    if ( $uri =~ /^ldap/ ) {
	$searchedvalue=$libldap->searchValueFromLdap($attr, $mapvalue, 
						     $mapattr, $uri, 
						     $searchtype, $pwtoken);
    } elsif  ( $uri =~ /^dbi/ ) {
	$searchedvalue=$libdbi->searchValueFromSql($attr, $mapvalue, $mapattr,
						   $uri, $searchtype, $pwtoken);
    } 

}


sub loadValues {
    my $self = shift;
    my ( $ra_list, $uri, $attr, $pwtoken ) = @_;
    $logger->debug("Loading $attr values from $uri");
#print "Loading values (attr $attr from $uri)\n";
    if ( $uri =~ /^ldap/ ) {
	$libldap->loadValuesFromLdap($ra_list, $uri, $attr, $pwtoken);
    } elsif  ( $uri =~ /^dbi/ ) {
	$libdbi->loadValuesFromSql($ra_list, $uri, $attr, $pwtoken);
    }
    my $valuecount = @{$ra_list};
    if (! $valuecount ) {
#	print "no values found\n";
	$logger->warn("no values found");
    }
    else {
#	print "found $valuecount values\n";
	$logger->debug("found $valuecount values");
    }
}


sub xgeneralizedTime2date {
    my $self = shift;
    my ( $string ) = @_;
    
    my $date_str ="";
    
    if ( $string =~ /^(\d\d\d\d)(\d\d)(\d\d)(\d\d)(\d\d)(\d\d)Z$/ ) { 
	$date_str= "$1-$2-$3 $4:$5:$6"; 
    }
    
    return ($date_str);
}

# changes: PG 2017-04-24: made this more defensive
sub base64decodedsmlfile {
    my $self = shift;
    my ( $filename, $outfilename ) = @_;

    $logger->debug("base64decodedsmlfile(): called with input: $filename and output: $outfilename");
    #    my ( $fpin, $fpout);
    open (FPIN, "<$filename") or $logger->error("base64decodedsmlfile(): error opening file $filename for read: $!");
    open (FPOUT, ">$outfilename") or $logger->error("base64decodedsmlfile(): error opening file $outfilename for write: $!");
    my $count = 0;
    while ( <FPIN> ) {
	$count++;
	if ( $_ =~ m|<dsml:value encoding='base64'>(.+)</dsml:value>| ) {
	    my $base64val = $1;
	    my $val = decode_base64($base64val);
	    print FPOUT "<dsml:value>$val</dsml:value>";
	    $logger->debug("base64decodedsmlfile(): found 64encoded value: $val");
	}
	else {
#	    $logger->debug("base64decodedsmlfile(): no 64encoded value: $_");
	    print FPOUT $_;
	}
    }

    close(FPIN);
    close(FPOUT);
    
    if ( $count ) {
	$logger->debug("base64decodedsmlfile(): processed $count lines");
    }
    else {
	$logger->debug("base64decodedsmlfile(): NO LINES processed");
    }	

}

sub xunixtimestamp2generalizedTime {
    my $self = shift;
    my ( $string ) = @_;
    $string = $self->xtrim($string);
    if ($string) { 
	my ( $sek, $min, $hour, $day, $mon, $year ) = localtime($string);
	my $timestamp = sprintf("%4d%02d%02d%02d%02d%02dZ", $year+1900, 
				$mon+1, $day, $hour, $min, $sek);	    
	return $timestamp;
    }
    else {
	return '';
    }
}

sub xgeneralizedTime2unixtimestamp {
    my $self = shift;
    my ( $string ) = @_;

    my $epoch = "";

    if ( $string && $string =~ /^(\d\d\d\d)(\d\d)(\d\d)(\d\d)(\d\d)(\d\d)(\w+)$/ ) { 
	if ($7 ne 'Z') {
	    $logger->warn("xgeneralizedTime2unixtimestamp: LDAP timestamp not in UTC (Z) time format");
	}
	my $dt = DateTime->new(
	    year   => $1,
	    month  => $2,
	    day    => $3,
	    hour   => $4,
	    minute => $5,
	    second => $6,
	    );
	$epoch = $dt->epoch();
    }
    return $epoch;
}



sub xdate2generalizedTime {
    my $self = shift;
    my ( $string ) = @_;
    
    my $gtime;

    $string = $self->xtrim($string);
    if ($string) {
	my ($val, $sep, $datesep, $timesep) =  split /;/, $string, 4 ;
	$sep = ' ' if ( ! $sep || $sep eq 'BLANK');
	$datesep = '-' if ! $datesep;
	$timesep = ':' if ! $timesep;
	$logger->debug("xdate2generalizedTime() sep: |$sep|, datesep: |$datesep|, timesep: |$timesep|"); 
	my ($date, $time ) = split /$sep /, $val, 2;

	my ($year, $month, $day) = split /$datesep/, $date;

	if (! $year ) {
	    $logger->error("no year found in datastring $year");
	    $year = "";
	}

	if (! $month ) {
	    $logger->error("no month found in datastring $date");
	    $month = "";
	}

	if (! $day ) {
	    $logger->error("no day found in datastring $date");
	    $day="";
	}

	if ( $time ) {
	    my ($hour,$min,$sek) = split /$timesep/, $time;
	    $gtime = "$year$month$day$hour$min${sek}Z";
	} else {
	    $gtime = "$year$month${day}000000Z";
	}
	return ($gtime);
    } else {
	return ("");
    }
   
}

sub xcurrentDateAsGeneralizedTime {

    my $self = shift;

    my ( $sek, $min, $hour, $day, $mon, $year ) = localtime();
    my $timestamp = sprintf("%4d%02d%02d%02d%02d%02dZ", $year+1900, 
			    $mon+1, $day, $hour, $min, $sek);	    

    return ($timestamp);

}


sub elementExists {
    my $self = shift;
    my ($string, @list) = @_;

    my $is_found =0;
    foreach my $tmpstr ( @list ) {
	if ( $string eq $tmpstr )  {
	    $is_found = 1;
	    last;
	}

    }

    return($is_found);
}


sub xcreateuuid {
    my $self = shift;
    my $Uuid="";
    my $string = "";

    if ( ! $self->{data}{is_UUID_module} ) {
	$string = `uuidgen -t`;
	if ( $string) {
	    chomp($string);
	} else {
	    $logger->error("Couldn't create UUID");
	}
    } else {
	UUID::generate($Uuid);
	  UUID::unparse($Uuid, $string);
      }

    return ($string);

}

sub makeRegexpSafe {
    my $self = shift;
    my ($string) = @_;

#    my $tobequoted = "\)\(\]\[\#\-\*\+\.\{\}\\\/";
#    my $tobequoted = ')(\]\[#\-*+.{}\\/';

#    $string =~ s!([$tobequoted])!\\$1!g;
#    foreach my $char ( split '', $tobequoted) {
#	$string =~ s!(\$char)!\\$1!g;
#    }

    $string =~ s/"/\\"/g;
    $string = quotemeta($string); 
#    print "made regexpsafe: $string\n";
    return ($string);
}

sub makeSafe {
    my $self = shift;
    my ($string, $char ) = @_;

    if ( ! $char || $char == '#') {
	$string =~ s/\#/SZ\{FIS\}/g; 
    }
    if ( ! $char || $char == '(') {
	$string =~ s/\(/SZ\{RKA\}/g;
    }
    if ( ! $char || $char == ')') {
	$string =~ s/\)/SZ\{RKZ\}/g;
    }
    if ( ! $char || $char == '$') {
	$string =~ s/\$/SZ\{DOL\}/g;
    }
 
    return ($string);
}

sub redoMakeSafe {
    my $self = shift;
    my ($string, $char ) = @_;

    if ( ! $char || $char == '#') {
	$string =~ s/SZ\{FIS\}/\#/g;
    }
    if ( ! $char || $char == '(') {
	$string =~ s/SZ\{RKA\}/\(/g;
    }
    if ( ! $char || $char == ')') {
	$string =~ s/SZ\{RKZ\}/\)/g;
    }
    if ( ! $char || $char == '$') {
	$string =~ s/SZ\{DOL\}/\$/g;
    }
    return ($string);
}

### normalize and camelize deleting blanks and hyphens
sub xcamelize {
    my $self = shift;
    my ( $string ) = @_;
    my $normalized;

    if ( $string ) {
	$string = $self->xnormalize($string);
	
	if ( $string =~/^(.)/ ) {
	    my $char = uc($1);
	    $string =~ s/^(.)/$char/;
	}
	while ( $string =~ / (.)/ ) {
	    my $char = uc($1);
	    $string =~ s/ (.)/$char/;
	}
	while ( $string =~ /-(.)/ ) {
	    my $char = uc($1);
	    $string =~ s/-(.)/$char/;
	}
    }


    $normalized = $string;
    

    return $normalized;
}


##################################################################
# name: xnormalize
#
# Description: Replaces special non-ASCII charakters with ASCII descriptions
#              Lowercases all characters
#              Replaces Umlauts with two-letter equivalents (ü>ue,ß>ss,...)
#              Replaces accented characters with non-accented ones 
#              
#
# Input: $string: a String for replacement
#
# Output: $normalized: normalization
#
# Created: PG 201X
# Changes: 
#
######################################################################

sub xnormalize {
    my $self = shift;
    my ( $string ) = @_;
    my $normalized;

## both these experiments didn't work:
#    use Text::Normalize::NACO qw( naco_normalize );
#    use Unicode::Normalize 'normalize';
#    if ($string) {
#	print "normalize with string [$string]\n";
#    
#	$normalized = naco_normalize( $string );
#	$normalized = normalize('KD', $string );
#    }

    if ($string) {
#	print "normalizing: [$string]\n";
	my @chararray = split //, $string;

	my $is_wait_for_next =0;

	foreach my $char (@chararray) {
#	    print "testing char [$char]\n"; 
	    if ( ord($char) > 31 && ord($char) < 127 ) {
		$normalized .=lc($char);
	    } else { 
		if ( ord($char) == 195 ) {
		    $is_wait_for_next = 1;
		} 
		if ($is_wait_for_next ) {
		    $is_wait_for_next = 1;
		    if (ord($char) == 128 ) { $normalized .='a'; }
		    elsif (ord($char) > 128 && ord($char) < 132 ) { 
			$normalized .='a'; 
		    }
		    elsif (ord($char) == 132 ) { $normalized .='ae'; }
		    elsif (ord($char) == 133 ) { $normalized .='a'; }
		    elsif (ord($char) == 134 ) { $normalized .='ae'; }
		    elsif (ord($char) == 135 ) { $normalized .='c'; }
		    elsif (ord($char) > 135 && ord($char) < 140 ) { 
			$normalized .='e'; 
		    }
		    elsif (ord($char) > 139 && ord($char) < 144 ) { 
			$normalized .='i'; 
		    }
		    elsif (ord($char) == 144 ) { $normalized .='d'; }
		    elsif (ord($char) == 145 ) { $normalized .='n'; }
		    elsif (ord($char) > 145 && ord($char) < 150 ) { 
			$normalized .='o'; 
		    }
		    elsif (ord($char) == 150 ) { $normalized .='oe'; }
		    elsif (ord($char) > 152 && ord($char) < 156 ) { 
			$normalized .='u'; 
		    }
		    elsif (ord($char) == 156 ) { $normalized .='ue'; }
		    elsif (ord($char) == 159 ) { $normalized .='ss'; }
		    elsif (ord($char) > 159 && ord($char) < 164 ) { 
			$normalized .='a'; 
		    }
		    elsif (ord($char) == 164 ) { $normalized .='ae'; }
		    elsif (ord($char) == 165 ) { $normalized .='a'; }
		    elsif (ord($char) == 166 ) { $normalized .='ae'; }
		    elsif (ord($char) == 167 ) { $normalized .='c'; }
		    elsif (ord($char) > 167 && ord($char) < 172 ) { 
			$normalized .='e'; 
		    }
		    elsif (ord($char) > 171 && ord($char) < 175 ) { 
			$normalized .='i'; 
		    }
		    elsif (ord($char) == 177 ) { $normalized .='n'; }
		    elsif (ord($char) > 177 && ord($char) < 182 ) { 
			$normalized .='o'; 
		    }
		    elsif (ord($char) == 182 ) { $normalized .='oe'; }
		    elsif (ord($char) > 184 && ord($char) < 188 ) { 
			$normalized .='u'; 
		    }
		    elsif (ord($char) == 188 ) { $normalized .='ue'; }
		    elsif (ord($char) == 189 ) { $normalized .='y'; }
		    elsif (ord($char) == 191 ) { $normalized .='y'; }
		} elsif ( ord($char) == 228 ) { $normalized .='ae'; }
		elsif ( ord($char) == 246 ) { $normalized .='oe'; } 
		elsif ( ord($char) == 252 ) { $normalized .='ue'; } 
		elsif ( ord($char) == 196 ) { $normalized .='ae'; } 
		elsif ( ord($char) == 214 ) { $normalized .='oe'; } 
		elsif ( ord($char) == 220 ) { $normalized .='ue'; } 
		elsif ( ord($char) == 223 ) { $normalized .='ss'; } 

#		print "found non ascii char [$char] Nr. ". ord($char)."in String [$string]\n";
	    }
	}
	
    }

#   print "normalize returning: $normalized\n";
    return ($normalized);

}



sub xreversename {
    my $self = shift;
    my ( $string ) = @_;
 
    my ($lastname, $givenname) = split /,/, $string;

    return ("$givenname $lastname");

}


sub xdecapitalizename {
    my $self = shift;
    my ( $string ) = @_;

    if ( $string ) {
	$string =~ s/-/ - /g;
	my $newstring = "";
	foreach my $namepart ( split( / /, $string, -1 ) ) {
	    if ( ! $namepart ) { next; }
	    $namepart = lc($namepart);
	    $namepart = "\u$namepart";
	    if ( $newstring ) {
		$newstring = "$newstring ";
	    }
	    $newstring .= $namepart;
	}
	$newstring =~ s/ - /-/g;
	return($newstring);
    } else {
	return("");
    }

}


sub xsybase_to_utf8 {
    my $self = shift;
    my ( $str ) =@_;

    my %converts = (
		    "\xc3" => "\xc3\x9c", # Uuml
		    "\xc1" => "\xc3\x84", # Auml
		    "\xc2" => "\xc3\x96", # Ouml
		    "\xd1" => "\xc3\xa4", # auml
		    "\xd2" => "\xc3\xb6", # ouml
		    "\xd3" => "\xc3\xbc", # uuml 
		    "\xbe" => "\xc3\x9f", # szlig
		    "\xe2" => "\xc3\xa9", # eacute
		    );
## \xe1
## \xe3
## \xcc

    my $dontknow = '[??]';

    my $retstr = '';
    if ($str ) {
#	my @chars = unpack("U*", $str);
	my @chars = split //, $str;
	foreach my $c (@chars) {
	    if (ord($c) > ord('~') ) {
#		print "konvertiere $c\n";
		if ( $converts{$c} ) {
		    $retstr .= $converts{$c};
		}
		else {
		    $retstr .= $dontknow;
		    $logger->error("Sybase_to_utf8 doesn't know what to do with character $c (dec. ".ord($c).") in string $str");
		}	    
	    }
	    else {
		$retstr .= $c;
	    }
	}

	return($retstr);
    } 
    else {
	return("");
    }

}

sub xsybase_to_utf8_old {
    my $self = shift;
    my ( $str ) =@_;

    if ($str ) {

	$str =~ s/\xc3/\xc3\x9c/g; # Uuml
	$str =~ s/\xc1/\xc3\x84/g; # Auml
	$str =~ s/\xc2/\xc3\x96/g; # Ouml
	$str =~ s/\xd1/\xc3\xa4/g; # auml
	$str =~ s/\xd2/\xc3\xb6/g; # ouml
	$str =~ s/\xd3/\xc3\xbc/g; # uuml 
	$str =~ s/\xbe/\xc3\x9f/g; # szlig
	
	$str =~ s/\xe1/[???]/g; # eacute
	$str =~ s/\xe2/\xc3\xa9/g; # eacute
	$str =~ s/\xe2/[???]/g; # eacute
#	$str =~ s/\xe3/[???]/g; # eacute

	if ( $str =~ /([\x7f-\xff]+)/ ) {
###	    if ( $1 != \xc3
	    my $char =  ord($1);
	    $logger->error("Sybase_to_utf8 doesn't know what to do with character $1 dec. ($char) in string $str");
	    $str =~ s/$1/[???]/g;
	}
	
# $str =~ s//\xc3\xa8/g; # egrave
# $str =~ s/&ecirc;/\xc3\xaa/g; # ecirc

    
	return($str);
    } else {
	return("");
    }
}

sub html_to_utf8 {
    my $self = shift;
    my ( $str ) =@_;

$str =~ s/&uuml;/\xc3\xbc/ig;
$str =~ s/&nbsp;/\xc2\xa0/g;
$str =~ s/&iexcl;/\xc2\xa1/g;
$str =~ s/&cent;/\xc2\xa2/g;
$str =~ s/&pound;/\xc2\xa3/g;
$str =~ s/&curren;/\xc2\xa4/g;
$str =~ s/&yen;/\xc2\xa5/g;
$str =~ s/&brvbar;/\xc2\xa6/g;
$str =~ s/&sect;/\xc2\xa7/g;
$str =~ s/&uml;/\xc2\xa8/g;
$str =~ s/&copy;/\xc2\xa9/g;
$str =~ s/&ordf;/\xc2\xaa/g;
$str =~ s/&laquo;/\xc2\xab/g;
$str =~ s/&not;/\xc2\xac/g;
$str =~ s/&shy;/\xc2\xad/g;
$str =~ s/&reg;/\xc2\xae/g;
$str =~ s/&macr;/\xc2\xaf/g;
$str =~ s/&deg;/\xc2\xb0/g;
$str =~ s/&plusmn;/\xc2\xb1/g;
$str =~ s/&sup2;/\xc2\xb2/g;
$str =~ s/&sup3;/\xc2\xb3/g;
$str =~ s/&acute;/\xc2\xb4/g;
$str =~ s/&micro;/\xc2\xb5/g;
$str =~ s/&para;/\xc2\xb6/g;
$str =~ s/&middot;/\xc2\xb7/g;
$str =~ s/&cedil;/\xc2\xb8/g;
$str =~ s/&supl;/\xc2\xb9/g;
$str =~ s/&ordm;/\xc2\xba/g;
$str =~ s/&raquo;/\xc2\xbb/g;
$str =~ s/&frac14;/\xc2\xbc/g;
$str =~ s/&frac12;/\xc2\xbd/g;
$str =~ s/&frac34;/\xc2\xbe/g;
$str =~ s/&iquest;/\xc2\xbf/g;
$str =~ s/&Agrave;/\xc3\x80/g;
$str =~ s/&Aacute;/\xc3\x81/g;
$str =~ s/&Acirc;/\xc3\x82/g;
$str =~ s/&Atilde;/\xc3\x83/g;
$str =~ s/&Auml;/\xc3\x84/g;
$str =~ s/&Aring;/\xc3\x85/g;
$str =~ s/&AElig;/\xc3\x86/g;
$str =~ s/&Ccedil;/\xc3\x87/g;
$str =~ s/&Egrave;/\xc3\x88/g;
$str =~ s/&Eacute;/\xc3\x89/g;
$str =~ s/&Ecirc;/\xc3\x8a/g;
$str =~ s/&Euml;/\xc3\x8b/g;
$str =~ s/&Igrave;/\xc3\x8c/g;
$str =~ s/&Iacute;/\xc3\x8d/g;
$str =~ s/&Icirc;/\xc3\x8e/g;
$str =~ s/&Iuml;/\xc3\x8f/g;
$str =~ s/&ETH;/\xc3\x90/g;
$str =~ s/&Ntilde;/\xc3\x91/g;
$str =~ s/&Ograve;/\xc3\x92/g;
$str =~ s/&Oacute;/\xc3\x93/g;
$str =~ s/&Ocirc;/\xc3\x94/g;
$str =~ s/&Otilde;/\xc3\x95/g;
$str =~ s/&Ouml;/\xc3\x96/g;
$str =~ s/&times;/\xc3\x97/g;
$str =~ s/&Oslash;/\xc3\x98/g;
$str =~ s/&Ugrave;/\xc3\x99/g;
$str =~ s/&Uacute;/\xc3\x9a/g;
$str =~ s/&Ucirc;/\xc3\x9b/g;
$str =~ s/&Uuml;/\xc3\x9c/g;
$str =~ s/&Yacute;/\xc3\x9d/g;
$str =~ s/&THORN;/\xc3\x9e/g;
$str =~ s/&szlig;/\xc3\x9f/g;
$str =~ s/&agrave;/\xc3\xa0/g;
$str =~ s/&aacute;/\xc3\xa1/g;
$str =~ s/&acirc;/\xc3\xa2/g;
$str =~ s/&atilde;/\xc3\xa3/g;
$str =~ s/&auml;/\xc3\xa4/g;
$str =~ s/&aring;/\xc3\xa5/g;
$str =~ s/&aelig;/\xc3\xa6/g;
$str =~ s/&ccedil;/\xc3\xa7/g;
$str =~ s/&egrave;/\xc3\xa8/g;
$str =~ s/&eacute;/\xc3\xa9/g;
$str =~ s/&ecirc;/\xc3\xaa/g;
$str =~ s/&euml;/\xc3\xab/g;
$str =~ s/&igrave;/\xc3\xac/g;
$str =~ s/&iacute;/\xc3\xad/g;
$str =~ s/&icirc;/\xc3\xae/g;
$str =~ s/&iuml;/\xc3\xaf/g;
$str =~ s/&eth;/\xc3\xb0/g;
$str =~ s/&ntilde;/\xc3\xb1/g;
$str =~ s/&ograve;/\xc3\xb2/g;
$str =~ s/&oacute;/\xc3\xb3/g;
$str =~ s/&ocirc;/\xc3\xb4/g;
$str =~ s/&otilde;/\xc3\xb5/g;
$str =~ s/&ouml;/\xc3\xb6/g;
$str =~ s/&divide;/\xc3\xb7/g;
$str =~ s/&oslash;/\xc3\xb8/g;
$str =~ s/&ugrave;/\xc3\xb9/g;
$str =~ s/&uacute;/\xc3\xba/g;
$str =~ s/&ucirc;/\xc3\xbb/g;
$str =~ s/&yacute;/\xc3\xbd/g;
$str =~ s/&thorn;/\xc3\xbe/g;
$str =~ s/&yuml;/\xc3\xbf/g;
$str =~ s/&OElig;/\xc5\x92/g;
$str =~ s/&oelig;/\xc5\x93/g;
$str =~ s/&Scaron;/\xc5\xa0/g;
$str =~ s/&scaron;/\xc5\xa1/g;
$str =~ s/&Yuml;/\xc5\xb8/g;
$str =~ s/&fnof;/\xc6\x92/g;
$str =~ s/&circ;/\xcb\x86/g;
$str =~ s/&tilde;/\xcb\x9c/g;
$str =~ s/&ndash;/\xe2\x80\x93/g;
$str =~ s/&mdash;/\xe2\x80\x94/g;
$str =~ s/&lsquo;/\xe2\x80\x98/g;
$str =~ s/&rsquo;/\xe2\x80\x99/g;
$str =~ s/&sbquo;/\xe2\x80\x9a/g;
$str =~ s/&ldquo;/\xe2\x80\x9c/g;
$str =~ s/&rdquo;/\xe2\x80\x9d/g;
$str =~ s/&bdquo;/\xe2\x80\x9e/g;
$str =~ s/&dagger;/\xe2\x80\xa0/g;
$str =~ s/&Dagger;/\xe2\x80\xa1/g;
$str =~ s/&bull;/\xe2\x80\xa2/g;
$str =~ s/&hellip;/\xe2\x80\xa6/g;
$str =~ s/&permil;/\xe2\x80\xb0/g;
$str =~ s/&lsaquo;/\xe2\x80\xb9/g;
$str =~ s/&rsaquo;/\xe2\x80\xba/g;
$str =~ s/&euro;/\xe2\x82\xac/g;
$str =~ s/&trade;/\xe2\x84\xa2/g;

    return ($str);
}


sub utf8_to_html {
    my $self = shift;
    my ( $str ) =@_;

$str =~ s/\xc3\xbc/&uuml;/ig;
$str =~ s/\xc2\xa0/&nbsp;/g;
$str =~ s/\xc2\xa1/&iexcl;/g;
$str =~ s/\xc2\xa2/&cent;/g;
$str =~ s/\xc2\xa3/&pound;/g;
$str =~ s/\xc2\xa4/&curren;/g;
$str =~ s/\xc2\xa5/&yen;/g;
$str =~ s/\xc2\xa6/&brvbar;/g;
$str =~ s/\xc2\xa7/&sect;/g;
$str =~ s/\xc2\xa8/&uml;/g;
$str =~ s/\xc2\xa9/&copy;/g;
$str =~ s/\xc2\xaa/&ordf;/g;
$str =~ s/\xc2\xab/&laquo;/g;
$str =~ s/\xc2\xac/&not;/g;
$str =~ s/\xc2\xad/&shy;/g;
$str =~ s/\xc2\xae/&reg;/g;
$str =~ s/\xc2\xaf/&macr;/g;
$str =~ s/\xc2\xb0/&deg;/g;
$str =~ s/\xc2\xb1/&plusmn;/g;
$str =~ s/\xc2\xb2/&sup2;/g;
$str =~ s/\xc2\xb3/&sup3;/g;
$str =~ s/\xc2\xb4/&acute;/g;
$str =~ s/\xc2\xb5/&micro;/g;
$str =~ s/\xc2\xb6/&para;/g;
$str =~ s/\xc2\xb7/&middot;/g;
$str =~ s/\xc2\xb8/&cedil;/g;
$str =~ s/\xc2\xb9/&supl;/g;
$str =~ s/\xc2\xba/&ordm;/g;
$str =~ s/\xc2\xbb/&raquo;/g;
$str =~ s/\xc2\xbc/&frac14;/g;
$str =~ s/\xc2\xbd/&frac12;/g;
$str =~ s/\xc2\xbe/&frac34;/g;
$str =~ s/\xc2\xbf/&iquest;/g;
$str =~ s/\xc3\x80/&Agrave;/g;
$str =~ s/\xc3\x81/&Aacute;/g;
$str =~ s/\xc3\x82/&Acirc;/g;
$str =~ s/\xc3\x83/&Atilde;/g;
$str =~ s/\xc3\x84/&Auml;/g;
$str =~ s/\xc3\x85/&Aring;/g;
$str =~ s/\xc3\x86/&AElig;/g;
$str =~ s/\xc3\x87/&Ccedil;/g;
$str =~ s/\xc3\x88/&Egrave;/g;
$str =~ s/\xc3\x89/&Eacute;/g;
$str =~ s/\xc3\x8a/&Ecirc;/g;
$str =~ s/\xc3\x8b/&Euml;/g;
$str =~ s/\xc3\x8c/&Igrave;/g;
$str =~ s/\xc3\x8d/&Iacute;/g;
$str =~ s/\xc3\x8e/&Icirc;/g;
$str =~ s/\xc3\x8f/&Iuml;/g;
$str =~ s/\xc3\x90/&ETH;/g;
$str =~ s/\xc3\x91/&Ntilde;/g;
$str =~ s/\xc3\x92/&Ograve;/g;
$str =~ s/\xc3\x93/&Oacute;/g;
$str =~ s/\xc3\x94/&Ocirc;/g;
$str =~ s/\xc3\x95/&Otilde;/g;
$str =~ s/\xc3\x96/&Ouml;/g;
$str =~ s/\xc3\x97/&times;/g;
$str =~ s/\xc3\x98/&Oslash;/g;
$str =~ s/\xc3\x99/&Ugrave;/g;
$str =~ s/\xc3\x9a/&Uacute;/g;
$str =~ s/\xc3\x9b/&Ucirc;/g;
$str =~ s/\xc3\x9c/&Uuml;/g;
$str =~ s/\xc3\x9d/&Yacute;/g;
$str =~ s/\xc3\x9e/&THORN;/g;
$str =~ s/\xc3\x9f/&szlig;/g;
$str =~ s/\xc3\xa0/&agrave;/g;
$str =~ s/\xc3\xa1/&aacute;/g;
$str =~ s/\xc3\xa2/&acirc;/g;
$str =~ s/\xc3\xa3/&atilde;/g;
$str =~ s/\xc3\xa4/&auml;/g;
$str =~ s/\xc3\xa5/&aring;/g;
$str =~ s/\xc3\xa6/&aelig;/g;
$str =~ s/\xc3\xa7/&ccedil;/g;
$str =~ s/\xc3\xa8/&egrave;/g;
$str =~ s/\xc3\xa9/&eacute;/g;
$str =~ s/\xc3\xaa/&ecirc;/g;
$str =~ s/\xc3\xab/&euml;/g;
$str =~ s/\xc3\xac/&igrave;/g;
$str =~ s/\xc3\xad/&iacute;/g;
$str =~ s/\xc3\xae/&icirc;/g;
$str =~ s/\xc3\xaf/&iuml;/g;
$str =~ s/\xc3\xb0/&eth;/g;
$str =~ s/\xc3\xb1/&ntilde;/g;
$str =~ s/\xc3\xb2/&ograve;/g;
$str =~ s/\xc3\xb3/&oacute;/g;
$str =~ s/\xc3\xb4/&ocirc;/g;
$str =~ s/\xc3\xb5/&otilde;/g;
$str =~ s/\xc3\xb6/&ouml;/g;
$str =~ s/\xc3\xb7/&divide;/g;
$str =~ s/\xc3\xb8/&oslash;/g;
$str =~ s/\xc3\xb9/&ugrave;/g;
$str =~ s/\xc3\xba/&uacute;/g;
$str =~ s/\xc3\xbb/&ucirc;/g;
$str =~ s/\xc3\xbd/&yacute;/g;
$str =~ s/\xc3\xbe/&thorn;/g;
$str =~ s/\xc3\xbf/&yuml;/g;
$str =~ s/\xc5\x92/&OElig;/g;
$str =~ s/\xc5\x93/&oelig;/g;
$str =~ s/\xc5\xa0/&Scaron;/g;
$str =~ s/\xc5\xa1/&scaron;/g;
$str =~ s/\xc5\xb8/&Yuml;/g;
$str =~ s/\xc6\x92/&fnof;/g;
$str =~ s/\xcb\x86/&circ;/g;
$str =~ s/\xcb\x9c/&tilde;/g;
$str =~ s/\xe2\x80\x93/&ndash;/g;
$str =~ s/\xe2\x80\x94/&mdash;/g;
$str =~ s/\xe2\x80\x98/&lsquo;/g;
$str =~ s/\xe2\x80\x99/&rsquo;/g;
$str =~ s/\xe2\x80\x9a/&sbquo;/g;
$str =~ s/\xe2\x80\x9c/&ldquo;/g;
$str =~ s/\xe2\x80\x9d/&rdquo;/g;
$str =~ s/\xe2\x80\x9e/&bdquo;/g;
$str =~ s/\xe2\x80\xa0/&dagger;/g;
$str =~ s/\xe2\x80\xa1/&Dagger;/g;
$str =~ s/\xe2\x80\xa2/&bull;/g;
$str =~ s/\xe2\x80\xa6/&hellip;/g;
$str =~ s/\xe2\x80\xb0/&permil;/g;
$str =~ s/\xe2\x80\xb9/&lsaquo;/g;
$str =~ s/\xe2\x80\xba/&rsaquo;/g;
$str =~ s/\xe2\x82\xac/&euro;/g;
$str =~ s/\xe2\x84\xa2/&trade;/g;

   return($str);
}


sub utf8_to_xml {
    my $self = shift;
    my ( $str ) =@_;

$str =~ s/\xc3\xbc/&uuml;/ig;
$str =~ s/\xc2\xa0/&nbsp;/g;
$str =~ s/\xc2\xa1/&iexcl;/g;
$str =~ s/\xc2\xa2/&cent;/g;
$str =~ s/\xc2\xa3/&pound;/g;
$str =~ s/\xc2\xa4/&curren;/g;
$str =~ s/\xc2\xa5/&yen;/g;
$str =~ s/\xc2\xa6/&brvbar;/g;
$str =~ s/\xc2\xa7/&sect;/g;
$str =~ s/\xc2\xa8/&uml;/g;
$str =~ s/\xc2\xa9/&copy;/g;
$str =~ s/\xc2\xaa/&ordf;/g;
$str =~ s/\xc2\xab/&laquo;/g;
$str =~ s/\xc2\xac/&not;/g;
$str =~ s/\xc2\xad/&shy;/g;
$str =~ s/\xc2\xae/&reg;/g;
$str =~ s/\xc2\xaf/&macr;/g;
$str =~ s/\xc2\xb0/&deg;/g;
$str =~ s/\xc2\xb1/&plusmn;/g;
$str =~ s/\xc2\xb2/&sup2;/g;
$str =~ s/\xc2\xb3/&sup3;/g;
$str =~ s/\xc2\xb4/&acute;/g;
$str =~ s/\xc2\xb5/&micro;/g;
$str =~ s/\xc2\xb6/&para;/g;
$str =~ s/\xc2\xb7/&middot;/g;
$str =~ s/\xc2\xb8/&cedil;/g;
$str =~ s/\xc2\xb9/&supl;/g;
$str =~ s/\xc2\xba/&ordm;/g;
$str =~ s/\xc2\xbb/&raquo;/g;
$str =~ s/\xc2\xbc/&frac14;/g;
$str =~ s/\xc2\xbd/&frac12;/g;
$str =~ s/\xc2\xbe/&frac34;/g;
$str =~ s/\xc2\xbf/&iquest;/g;
$str =~ s/\xc3\x80/&Agrave;/g;
$str =~ s/\xc3\x81/&Aacute;/g;
$str =~ s/\xc3\x82/&Acirc;/g;
$str =~ s/\xc3\x83/&Atilde;/g;
$str =~ s/\xc3\x84/&Auml;/g;
$str =~ s/\xc3\x85/&Aring;/g;
$str =~ s/\xc3\x86/&AElig;/g;
$str =~ s/\xc3\x87/&Ccedil;/g;
$str =~ s/\xc3\x88/&Egrave;/g;
$str =~ s/\xc3\x89/&Eacute;/g;
$str =~ s/\xc3\x8a/&Ecirc;/g;
$str =~ s/\xc3\x8b/&Euml;/g;
$str =~ s/\xc3\x8c/&Igrave;/g;
$str =~ s/\xc3\x8d/&Iacute;/g;
$str =~ s/\xc3\x8e/&Icirc;/g;
$str =~ s/\xc3\x8f/&Iuml;/g;
$str =~ s/\xc3\x90/&ETH;/g;
$str =~ s/\xc3\x91/&Ntilde;/g;
$str =~ s/\xc3\x92/&Ograve;/g;
$str =~ s/\xc3\x93/&Oacute;/g;
$str =~ s/\xc3\x94/&Ocirc;/g;
$str =~ s/\xc3\x95/&Otilde;/g;
$str =~ s/\xc3\x96/&Ouml;/g;
$str =~ s/\xc3\x97/&times;/g;
$str =~ s/\xc3\x98/&Oslash;/g;
$str =~ s/\xc3\x99/&Ugrave;/g;
$str =~ s/\xc3\x9a/&Uacute;/g;
$str =~ s/\xc3\x9b/&Ucirc;/g;
$str =~ s/\xc3\x9c/&Uuml;/g;
$str =~ s/\xc3\x9d/&Yacute;/g;
$str =~ s/\xc3\x9e/&THORN;/g;
$str =~ s/\xc3\x9f/&szlig;/g;
$str =~ s/\xc3\xa0/&agrave;/g;
$str =~ s/\xc3\xa1/&aacute;/g;
$str =~ s/\xc3\xa2/&acirc;/g;
$str =~ s/\xc3\xa3/&atilde;/g;
$str =~ s/\xc3\xa4/&#xE4;/g;
$str =~ s/\xc3\xa5/&aring;/g;
$str =~ s/\xc3\xa6/&aelig;/g;
$str =~ s/\xc3\xa7/&ccedil;/g;
$str =~ s/\xc3\xa8/&egrave;/g;
$str =~ s/\xc3\xa9/&eacute;/g;
$str =~ s/\xc3\xaa/&ecirc;/g;
$str =~ s/\xc3\xab/&euml;/g;
$str =~ s/\xc3\xac/&igrave;/g;
$str =~ s/\xc3\xad/&iacute;/g;
$str =~ s/\xc3\xae/&icirc;/g;
$str =~ s/\xc3\xaf/&iuml;/g;
$str =~ s/\xc3\xb0/&eth;/g;
$str =~ s/\xc3\xb1/&ntilde;/g;
$str =~ s/\xc3\xb2/&ograve;/g;
$str =~ s/\xc3\xb3/&oacute;/g;
$str =~ s/\xc3\xb4/&ocirc;/g;
$str =~ s/\xc3\xb5/&otilde;/g;
$str =~ s/\xc3\xb6/&ouml;/g;
$str =~ s/\xc3\xb7/&divide;/g;
$str =~ s/\xc3\xb8/&oslash;/g;
$str =~ s/\xc3\xb9/&ugrave;/g;
$str =~ s/\xc3\xba/&uacute;/g;
$str =~ s/\xc3\xbb/&ucirc;/g;
$str =~ s/\xc3\xbd/&yacute;/g;
$str =~ s/\xc3\xbe/&thorn;/g;
$str =~ s/\xc3\xbf/&yuml;/g;
$str =~ s/\xc5\x92/&OElig;/g;
$str =~ s/\xc5\x93/&oelig;/g;
$str =~ s/\xc5\xa0/&Scaron;/g;
$str =~ s/\xc5\xa1/&scaron;/g;
$str =~ s/\xc5\xb8/&Yuml;/g;
$str =~ s/\xc6\x92/&fnof;/g;
$str =~ s/\xcb\x86/&circ;/g;
$str =~ s/\xcb\x9c/&tilde;/g;
$str =~ s/\xe2\x80\x93/&ndash;/g;
$str =~ s/\xe2\x80\x94/&mdash;/g;
$str =~ s/\xe2\x80\x98/&lsquo;/g;
$str =~ s/\xe2\x80\x99/&rsquo;/g;
$str =~ s/\xe2\x80\x9a/&sbquo;/g;
$str =~ s/\xe2\x80\x9c/&ldquo;/g;
$str =~ s/\xe2\x80\x9d/&rdquo;/g;
$str =~ s/\xe2\x80\x9e/&bdquo;/g;
$str =~ s/\xe2\x80\xa0/&dagger;/g;
$str =~ s/\xe2\x80\xa1/&Dagger;/g;
$str =~ s/\xe2\x80\xa2/&bull;/g;
$str =~ s/\xe2\x80\xa6/&hellip;/g;
$str =~ s/\xe2\x80\xb0/&permil;/g;
$str =~ s/\xe2\x80\xb9/&lsaquo;/g;
$str =~ s/\xe2\x80\xba/&rsaquo;/g;
$str =~ s/\xe2\x82\xac/&euro;/g;
$str =~ s/\xe2\x84\xa2/&trade;/g;

   return($str);
}


sub utf8_to_ascii {
    my $self = shift;

    my ( $str ) =@_;
    my $ori = $str;
#    print "utf8_to_ascii() got |$str|\n";

$str =~ s/\xc3\x80/A/g;
$str =~ s/\xc3\x81/A/g;
$str =~ s/\xc3\x82/A/g;
$str =~ s/\xc3\x83/A/g;
$str =~ s/\xc3\x84/A/g;
$str =~ s/\xc3\x85/A/g;
$str =~ s/\xc3\x87/C/g;
$str =~ s/\xc3\x88/E/g;
$str =~ s/\xc3\x89/E/g;
$str =~ s/\xc3\x8a/E/g;
$str =~ s/\xc3\x8b/E/g;
$str =~ s/\xc3\x8c/I/g;
$str =~ s/\xc3\x8d/I/g;
$str =~ s/\xc3\x8e/I/g;
$str =~ s/\xc3\x8f/I/g;
$str =~ s/\xc3\x91/N/g;
$str =~ s/\xc3\x92/O/g;
$str =~ s/\xc3\x93/O/g;
$str =~ s/\xc3\x94/O/g;
$str =~ s/\xc3\x95/O/g;
$str =~ s/\xc3\x96/O/g;
$str =~ s/\xc3\x98/O/g;
$str =~ s/\xc3\x99/U/g;
$str =~ s/\xc3\x9a/U/g;
$str =~ s/\xc3\x9b/U/g;
$str =~ s/\xc3\x9c/U/g;
$str =~ s/\xc3\x9d/Y/g;
$str =~ s/\xc3\xa0/a/g;
$str =~ s/\xc3\xa1/a/g;
$str =~ s/\xc3\xa2/a/g;
$str =~ s/\xc3\xa3/a/g;
$str =~ s/\xc3\xa4/a/g;
$str =~ s/\xc3\xa5/a/g;
$str =~ s/\xc3\xa7/c/g;
$str =~ s/\xc3\xa8/e/g;
$str =~ s/\xc3\xa9/e/g;
$str =~ s/\xc3\xaa/e/g;
$str =~ s/\xc3\xab/e/g;
$str =~ s/\xc3\xac/i/g;
$str =~ s/\xc3\xad/i/g;
$str =~ s/\xc3\xae/i/g;
$str =~ s/\xc3\xaf/i/g;
$str =~ s/\xc3\xb1/n/g;
$str =~ s/\xc3\xb2/o/g;
$str =~ s/\xc3\xb3/o/g;
$str =~ s/\xc3\xb4/o/g;
$str =~ s/\xc3\xb5/o/g;
$str =~ s/\xc3\xb6/o/g;
$str =~ s/\xc3\xb8/o/g;
$str =~ s/\xc3\xb9/u/g;
$str =~ s/\xc3\xba/u/g;
$str =~ s/\xc3\xbb/u/g;
$str =~ s/\xc3\xbc/u/g;
$str =~ s/\xc3\xbd/y/g;
$str =~ s/\xc3\xbf/y/g;
$str =~ s/\xc4\x80/A/g;
$str =~ s/\xc4\x81/a/g;
$str =~ s/\xc4\x82/A/g;
$str =~ s/\xc4\x83/a/g;
$str =~ s/\xc4\x84/A/g;
$str =~ s/\xc4\x85/a/g;
$str =~ s/\xc4\x86/C/g;
$str =~ s/\xc4\x87/c/g;
$str =~ s/\xc4\x88/C/g;
$str =~ s/\xc4\x89/c/g;
$str =~ s/\xc4\x8a/C/g;
$str =~ s/\xc4\x8b/c/g;
$str =~ s/\xc4\x8c/C/g;
$str =~ s/\xc4\x8d/c/g;
$str =~ s/\xc4\x8e/D/g;
$str =~ s/\xc4\x8f/d/g;
$str =~ s/\xc4\x90/D/g;
$str =~ s/\xc4\x91/d/g;
$str =~ s/\xc4\x92/E/g;
$str =~ s/\xc4\x93/e/g;
$str =~ s/\xc4\x94/E/g;
$str =~ s/\xc4\x95/e/g;
$str =~ s/\xc4\x96/E/g;
$str =~ s/\xc4\x97/e/g;
$str =~ s/\xc4\x98/E/g;
$str =~ s/\xc4\x99/e/g;
$str =~ s/\xc4\x9a/E/g;
$str =~ s/\xc4\x9b/e/g;
$str =~ s/\xc4\x9c/G/g;
$str =~ s/\xc4\x9d/g/g;
$str =~ s/\xc4\x9e/G/g;
$str =~ s/\xc4\x9f/g/g;
$str =~ s/\xc4\xa0/G/g;
$str =~ s/\xc4\xa1/g/g;
$str =~ s/\xc4\xa2/G/g;
$str =~ s/\xc4\xa3/g/g;
$str =~ s/\xc4\xa4/H/g;
$str =~ s/\xc4\xa5/h/g;
$str =~ s/\xc4\xa6/H/g;
$str =~ s/\xc4\xa7/h/g;
$str =~ s/\xc4\xa8/I/g;
$str =~ s/\xc4\xa9/i/g;
$str =~ s/\xc4\xaa/I/g;
$str =~ s/\xc4\xab/i/g;
$str =~ s/\xc4\xac/I/g;
$str =~ s/\xc4\xad/i/g;
$str =~ s/\xc4\xae/I/g;
$str =~ s/\xc4\xaf/i/g;
$str =~ s/\xc4\xb0/I/g;
$str =~ s/\xc4\xb4/J/g;
$str =~ s/\xc4\xb5/j/g;
$str =~ s/\xc4\xb6/K/g;
$str =~ s/\xc4\xb7/k/g;
$str =~ s/\xc4\xb9/L/g;
$str =~ s/\xc4\xba/l/g;
$str =~ s/\xc4\xbb/L/g;
$str =~ s/\xc4\xbc/l/g;
$str =~ s/\xc4\xbd/L/g;
$str =~ s/\xc4\xbe/l/g;
$str =~ s/\xc4\xbf/L/g;
$str =~ s/\xc5\x80/l/g;
$str =~ s/\xc5\x81/L/g;
$str =~ s/\xc5\x82/l/g;
$str =~ s/\xc5\x83/N/g;
$str =~ s/\xc5\x84/n/g;
$str =~ s/\xc5\x85/N/g;
$str =~ s/\xc5\x86/n/g;
$str =~ s/\xc5\x87/N/g;
$str =~ s/\xc5\x88/n/g;
$str =~ s/\xc5\x89/n/g;
$str =~ s/\xc5\x8c/O/g;
$str =~ s/\xc5\x8d/o/g;
$str =~ s/\xc5\x8e/O/g;
$str =~ s/\xc5\x8f/o/g;
$str =~ s/\xc5\x90/O/g;
$str =~ s/\xc5\x91/o/g;
$str =~ s/\xc5\x94/R/g;
$str =~ s/\xc5\x95/r/g;
$str =~ s/\xc5\x96/R/g;
$str =~ s/\xc5\x97/r/g;
$str =~ s/\xc5\x98/R/g;
$str =~ s/\xc5\x99/r/g;
$str =~ s/\xc5\x9a/S/g;
$str =~ s/\xc5\x9b/s/g;
$str =~ s/\xc5\x9c/S/g;
$str =~ s/\xc5\x9d/s/g;
$str =~ s/\xc5\x9e/S/g;
$str =~ s/\xc5\x9f/s/g;
$str =~ s/\xc5\xa0/S/g;
$str =~ s/\xc5\xa1/s/g;
$str =~ s/\xc5\xa2/T/g;
$str =~ s/\xc5\xa3/t/g;
$str =~ s/\xc5\xa4/T/g;
$str =~ s/\xc5\xa5/t/g;
$str =~ s/\xc5\xa6/T/g;
$str =~ s/\xc5\xa7/t/g;
$str =~ s/\xc5\xa8/U/g;
$str =~ s/\xc5\xa9/u/g;
$str =~ s/\xc5\xaa/U/g;
$str =~ s/\xc5\xab/u/g;
$str =~ s/\xc5\xac/U/g;
$str =~ s/\xc5\xad/u/g;
$str =~ s/\xc5\xae/U/g;
$str =~ s/\xc5\xaf/u/g;
$str =~ s/\xc5\xb0/U/g;
$str =~ s/\xc5\xb1/u/g;
$str =~ s/\xc5\xb2/U/g;
$str =~ s/\xc5\xb3/u/g;
$str =~ s/\xc5\xb4/W/g;
$str =~ s/\xc5\xb5/w/g;
$str =~ s/\xc5\xb6/Y/g;
$str =~ s/\xc5\xb7/y/g;
$str =~ s/\xc5\xb8/Y/g;
$str =~ s/\xc5\xb9/Z/g;
$str =~ s/\xc5\xba/z/g;
$str =~ s/\xc5\xbb/Z/g;
$str =~ s/\xc5\xbc/z/g;
$str =~ s/\xc5\xbd/Z/g;
$str =~ s/\xc5\xbe/z/g;
$str =~ s/\xc6\x80/b/g;
$str =~ s/\xc6\x81/B/g;
$str =~ s/\xc6\x82/B/g;
$str =~ s/\xc6\x83/b/g;
$str =~ s/\xc6\x87/C/g;
$str =~ s/\xc6\x88/c/g;
$str =~ s/\xc6\x8a/D/g;
$str =~ s/\xc6\x8b/D/g;
$str =~ s/\xc6\x8c/d/g;
$str =~ s/\xc6\x91/F/g;
$str =~ s/\xc6\x92/f/g;
$str =~ s/\xc6\x93/G/g;
$str =~ s/\xc6\x97/I/g;
$str =~ s/\xc6\x98/K/g;
$str =~ s/\xc6\x99/k/g;
$str =~ s/\xc6\x9a/l/g;
$str =~ s/\xc6\x9d/N/g;
$str =~ s/\xc6\x9e/n/g;
$str =~ s/\xc6\x9f/O/g;
$str =~ s/\xc6\xa0/O/g;
$str =~ s/\xc6\xa1/o/g;
$str =~ s/\xc6\xa4/P/g;
$str =~ s/\xc6\xa5/p/g;
$str =~ s/\xc6\xab/t/g;
$str =~ s/\xc6\xac/T/g;
$str =~ s/\xc6\xad/t/g;
$str =~ s/\xc6\xae/T/g;
$str =~ s/\xc6\xaf/U/g;
$str =~ s/\xc6\xb0/u/g;
$str =~ s/\xc6\xb2/V/g;
$str =~ s/\xc6\xb3/Y/g;
$str =~ s/\xc6\xb4/y/g;
$str =~ s/\xc6\xb5/Z/g;
$str =~ s/\xc6\xb6/z/g;
$str =~ s/\xc7\x85/D/g;
$str =~ s/\xc7\x88/L/g;
$str =~ s/\xc7\x8b/N/g;
$str =~ s/\xc7\x8d/A/g;
$str =~ s/\xc7\x8e/a/g;
$str =~ s/\xc7\x8f/I/g;
$str =~ s/\xc7\x90/i/g;
$str =~ s/\xc7\x91/O/g;
$str =~ s/\xc7\x92/o/g;
$str =~ s/\xc7\x93/U/g;
$str =~ s/\xc7\x94/u/g;
$str =~ s/\xc7\x95/U/g;
$str =~ s/\xc7\x96/u/g;
$str =~ s/\xc7\x97/U/g;
$str =~ s/\xc7\x98/u/g;
$str =~ s/\xc7\x99/U/g;
$str =~ s/\xc7\x9a/u/g;
$str =~ s/\xc7\x9b/U/g;
$str =~ s/\xc7\x9c/u/g;
$str =~ s/\xc7\x9e/A/g;
$str =~ s/\xc7\x9f/a/g;
$str =~ s/\xc7\xa0/A/g;
$str =~ s/\xc7\xa1/a/g;
$str =~ s/\xc7\xa4/G/g;
$str =~ s/\xc7\xa5/g/g;
$str =~ s/\xc7\xa6/G/g;
$str =~ s/\xc7\xa7/g/g;
$str =~ s/\xc7\xa8/K/g;
$str =~ s/\xc7\xa9/k/g;
$str =~ s/\xc7\xaa/O/g;
$str =~ s/\xc7\xab/o/g;
$str =~ s/\xc7\xac/O/g;
$str =~ s/\xc7\xad/o/g;
$str =~ s/\xc7\xb0/j/g;
$str =~ s/\xc7\xb2/D/g;
$str =~ s/\xc7\xb4/G/g;
$str =~ s/\xc7\xb5/g/g;
$str =~ s/\xc7\xb8/N/g;
$str =~ s/\xc7\xb9/n/g;
$str =~ s/\xc7\xba/A/g;
$str =~ s/\xc7\xbb/a/g;
$str =~ s/\xc7\xbe/O/g;
$str =~ s/\xc7\xbf/o/g;
$str =~ s/\xe1\xb8\x80/A/g;
$str =~ s/\xe1\xb8\x81/a/g;
$str =~ s/\xe1\xb8\x82/B/g;
$str =~ s/\xe1\xb8\x83/b/g;
$str =~ s/\xe1\xb8\x84/B/g;
$str =~ s/\xe1\xb8\x85/b/g;
$str =~ s/\xe1\xb8\x86/B/g;
$str =~ s/\xe1\xb8\x87/b/g;
$str =~ s/\xe1\xb8\x88/C/g;
$str =~ s/\xe1\xb8\x89/c/g;
$str =~ s/\xe1\xb8\x8a/D/g;
$str =~ s/\xe1\xb8\x8b/d/g;
$str =~ s/\xe1\xb8\x8c/D/g;
$str =~ s/\xe1\xb8\x8d/d/g;
$str =~ s/\xe1\xb8\x8e/D/g;
$str =~ s/\xe1\xb8\x8f/d/g;
$str =~ s/\xe1\xb8\x90/D/g;
$str =~ s/\xe1\xb8\x91/d/g;
$str =~ s/\xe1\xb8\x92/D/g;
$str =~ s/\xe1\xb8\x93/d/g;
$str =~ s/\xe1\xb8\x94/E/g;
$str =~ s/\xe1\xb8\x95/e/g;
$str =~ s/\xe1\xb8\x96/E/g;
$str =~ s/\xe1\xb8\x97/e/g;
$str =~ s/\xe1\xb8\x98/E/g;
$str =~ s/\xe1\xb8\x99/e/g;
$str =~ s/\xe1\xb8\x9a/E/g;
$str =~ s/\xe1\xb8\x9b/e/g;
$str =~ s/\xe1\xb8\x9c/E/g;
$str =~ s/\xe1\xb8\x9d/e/g;
$str =~ s/\xe1\xb8\x9e/F/g;
$str =~ s/\xe1\xb8\x9f/f/g;
$str =~ s/\xe1\xb8\xa0/G/g;
$str =~ s/\xe1\xb8\xa1/g/g;
$str =~ s/\xe1\xb8\xa2/H/g;
$str =~ s/\xe1\xb8\xa3/h/g;
$str =~ s/\xe1\xb8\xa4/H/g;
$str =~ s/\xe1\xb8\xa5/h/g;
$str =~ s/\xe1\xb8\xa6/H/g;
$str =~ s/\xe1\xb8\xa7/h/g;
$str =~ s/\xe1\xb8\xa8/H/g;
$str =~ s/\xe1\xb8\xa9/h/g;
$str =~ s/\xe1\xb8\xaa/H/g;
$str =~ s/\xe1\xb8\xab/h/g;
$str =~ s/\xe1\xb8\xac/I/g;
$str =~ s/\xe1\xb8\xad/i/g;
$str =~ s/\xe1\xb8\xae/I/g;
$str =~ s/\xe1\xb8\xaf/i/g;
$str =~ s/\xe1\xb8\xb0/K/g;
$str =~ s/\xe1\xb8\xb1/k/g;
$str =~ s/\xe1\xb8\xb2/K/g;
$str =~ s/\xe1\xb8\xb3/k/g;
$str =~ s/\xe1\xb8\xb4/K/g;
$str =~ s/\xe1\xb8\xb5/k/g;
$str =~ s/\xe1\xb8\xb6/L/g;
$str =~ s/\xe1\xb8\xb7/l/g;
$str =~ s/\xe1\xb8\xb8/L/g;
$str =~ s/\xe1\xb8\xb9/l/g;
$str =~ s/\xe1\xb8\xba/L/g;
$str =~ s/\xe1\xb8\xbb/l/g;
$str =~ s/\xe1\xb8\xbc/L/g;
$str =~ s/\xe1\xb8\xbd/l/g;
$str =~ s/\xe1\xb8\xbe/M/g;
$str =~ s/\xe1\xb8\xbf/m/g;
$str =~ s/\xe1\xb9\x80/M/g;
$str =~ s/\xe1\xb9\x81/m/g;
$str =~ s/\xe1\xb9\x82/M/g;
$str =~ s/\xe1\xb9\x83/m/g;
$str =~ s/\xe1\xb9\x84/N/g;
$str =~ s/\xe1\xb9\x85/n/g;
$str =~ s/\xe1\xb9\x86/N/g;
$str =~ s/\xe1\xb9\x87/n/g;
$str =~ s/\xe1\xb9\x88/N/g;
$str =~ s/\xe1\xb9\x89/n/g;
$str =~ s/\xe1\xb9\x8a/N/g;
$str =~ s/\xe1\xb9\x8b/n/g;
$str =~ s/\xe1\xb9\x8c/O/g;
$str =~ s/\xe1\xb9\x8d/o/g;
$str =~ s/\xe1\xb9\x8e/O/g;
$str =~ s/\xe1\xb9\x8f/o/g;
$str =~ s/\xe1\xb9\x90/O/g;
$str =~ s/\xe1\xb9\x91/o/g;
$str =~ s/\xe1\xb9\x92/O/g;
$str =~ s/\xe1\xb9\x93/o/g;
$str =~ s/\xe1\xb9\x94/P/g;
$str =~ s/\xe1\xb9\x95/p/g;
$str =~ s/\xe1\xb9\x96/P/g;
$str =~ s/\xe1\xb9\x97/p/g;
$str =~ s/\xe1\xb9\x98/R/g;
$str =~ s/\xe1\xb9\x99/r/g;
$str =~ s/\xe1\xb9\x9a/R/g;
$str =~ s/\xe1\xb9\x9b/r/g;
$str =~ s/\xe1\xb9\x9c/R/g;
$str =~ s/\xe1\xb9\x9d/r/g;
$str =~ s/\xe1\xb9\x9e/R/g;
$str =~ s/\xe1\xb9\x9f/r/g;
$str =~ s/\xe1\xb9\xa0/S/g;
$str =~ s/\xe1\xb9\xa1/s/g;
$str =~ s/\xe1\xb9\xa2/S/g;
$str =~ s/\xe1\xb9\xa3/s/g;
$str =~ s/\xe1\xb9\xa4/S/g;
$str =~ s/\xe1\xb9\xa5/s/g;
$str =~ s/\xe1\xb9\xa6/S/g;
$str =~ s/\xe1\xb9\xa7/s/g;
$str =~ s/\xe1\xb9\xa8/S/g;
$str =~ s/\xe1\xb9\xa9/s/g;
$str =~ s/\xe1\xb9\xaa/T/g;
$str =~ s/\xe1\xb9\xab/t/g;
$str =~ s/\xe1\xb9\xac/T/g;
$str =~ s/\xe1\xb9\xad/t/g;
$str =~ s/\xe1\xb9\xae/T/g;
$str =~ s/\xe1\xb9\xaf/t/g;
$str =~ s/\xe1\xb9\xb0/T/g;
$str =~ s/\xe1\xb9\xb1/t/g;
$str =~ s/\xe1\xb9\xb2/U/g;
$str =~ s/\xe1\xb9\xb3/u/g;
$str =~ s/\xe1\xb9\xb4/U/g;
$str =~ s/\xe1\xb9\xb5/u/g;
$str =~ s/\xe1\xb9\xb6/U/g;
$str =~ s/\xe1\xb9\xb7/u/g;
$str =~ s/\xe1\xb9\xb8/U/g;
$str =~ s/\xe1\xb9\xb9/u/g;
$str =~ s/\xe1\xb9\xba/U/g;
$str =~ s/\xe1\xb9\xbb/u/g;
$str =~ s/\xe1\xb9\xbc/V/g;
$str =~ s/\xe1\xb9\xbd/v/g;
$str =~ s/\xe1\xb9\xbe/V/g;
$str =~ s/\xe1\xb9\xbf/v/g;
$str =~ s/\xe1\xba\x80/W/g;
$str =~ s/\xe1\xba\x81/w/g;
$str =~ s/\xe1\xba\x82/W/g;
$str =~ s/\xe1\xba\x83/w/g;
$str =~ s/\xe1\xba\x84/W/g;
$str =~ s/\xe1\xba\x85/w/g;
$str =~ s/\xe1\xba\x86/W/g;
$str =~ s/\xe1\xba\x87/w/g;
$str =~ s/\xe1\xba\x88/W/g;
$str =~ s/\xe1\xba\x89/w/g;
$str =~ s/\xe1\xba\x8a/X/g;
$str =~ s/\xe1\xba\x8b/x/g;
$str =~ s/\xe1\xba\x8c/X/g;
$str =~ s/\xe1\xba\x8d/x/g;
$str =~ s/\xe1\xba\x8e/Y/g;
$str =~ s/\xe1\xba\x8f/y/g;
$str =~ s/\xe1\xba\x90/Z/g;
$str =~ s/\xe1\xba\x91/z/g;
$str =~ s/\xe1\xba\x92/Z/g;
$str =~ s/\xe1\xba\x93/z/g;
$str =~ s/\xe1\xba\x94/Z/g;
$str =~ s/\xe1\xba\x95/z/g;
$str =~ s/\xe1\xba\x96/h/g;
$str =~ s/\xe1\xba\x97/t/g;
$str =~ s/\xe1\xba\x98/w/g;
$str =~ s/\xe1\xba\x99/y/g;
$str =~ s/\xe1\xba\x9a/a/g;
$str =~ s/\xe1\xba\xa0/A/g;
$str =~ s/\xe1\xba\xa1/a/g;
$str =~ s/\xe1\xba\xa2/A/g;
$str =~ s/\xe1\xba\xa3/a/g;
$str =~ s/\xe1\xba\xa4/A/g;
$str =~ s/\xe1\xba\xa5/a/g;
$str =~ s/\xe1\xba\xa6/A/g;
$str =~ s/\xe1\xba\xa7/a/g;
$str =~ s/\xe1\xba\xa8/A/g;
$str =~ s/\xe1\xba\xa9/a/g;
$str =~ s/\xe1\xba\xaa/A/g;
$str =~ s/\xe1\xba\xab/a/g;
$str =~ s/\xe1\xba\xac/A/g;
$str =~ s/\xe1\xba\xad/a/g;
$str =~ s/\xe1\xba\xae/A/g;
$str =~ s/\xe1\xba\xaf/a/g;
$str =~ s/\xe1\xba\xb0/A/g;
$str =~ s/\xe1\xba\xb1/a/g;
$str =~ s/\xe1\xba\xb2/A/g;
$str =~ s/\xe1\xba\xb3/a/g;
$str =~ s/\xe1\xba\xb4/A/g;
$str =~ s/\xe1\xba\xb5/a/g;
$str =~ s/\xe1\xba\xb6/A/g;
$str =~ s/\xe1\xba\xb7/a/g;
$str =~ s/\xe1\xba\xb8/E/g;
$str =~ s/\xe1\xba\xb9/e/g;
$str =~ s/\xe1\xba\xba/E/g;
$str =~ s/\xe1\xba\xbb/e/g;
$str =~ s/\xe1\xba\xbc/E/g;
$str =~ s/\xe1\xba\xbd/e/g;
$str =~ s/\xe1\xba\xbe/E/g;
$str =~ s/\xe1\xba\xbf/e/g;
$str =~ s/\xe1\xbb\x80/E/g;
$str =~ s/\xe1\xbb\x81/e/g;
$str =~ s/\xe1\xbb\x82/E/g;
$str =~ s/\xe1\xbb\x83/e/g;
$str =~ s/\xe1\xbb\x84/E/g;
$str =~ s/\xe1\xbb\x85/e/g;
$str =~ s/\xe1\xbb\x86/E/g;
$str =~ s/\xe1\xbb\x87/e/g;
$str =~ s/\xe1\xbb\x88/I/g;
$str =~ s/\xe1\xbb\x89/i/g;
$str =~ s/\xe1\xbb\x8a/I/g;
$str =~ s/\xe1\xbb\x8b/i/g;
$str =~ s/\xe1\xbb\x8c/O/g;
$str =~ s/\xe1\xbb\x8d/o/g;
$str =~ s/\xe1\xbb\x8e/O/g;
$str =~ s/\xe1\xbb\x8f/o/g;
$str =~ s/\xe1\xbb\x90/O/g;
$str =~ s/\xe1\xbb\x91/o/g;
$str =~ s/\xe1\xbb\x92/O/g;
$str =~ s/\xe1\xbb\x93/o/g;
$str =~ s/\xe1\xbb\x94/O/g;
$str =~ s/\xe1\xbb\x95/o/g;
$str =~ s/\xe1\xbb\x96/O/g;
$str =~ s/\xe1\xbb\x97/o/g;
$str =~ s/\xe1\xbb\x98/O/g;
$str =~ s/\xe1\xbb\x99/o/g;
$str =~ s/\xe1\xbb\x9a/O/g;
$str =~ s/\xe1\xbb\x9b/o/g;
$str =~ s/\xe1\xbb\x9c/O/g;
$str =~ s/\xe1\xbb\x9d/o/g;
$str =~ s/\xe1\xbb\x9e/O/g;
$str =~ s/\xe1\xbb\x9f/o/g;
$str =~ s/\xe1\xbb\xa0/O/g;
$str =~ s/\xe1\xbb\xa1/o/g;
$str =~ s/\xe1\xbb\xa2/O/g;
$str =~ s/\xe1\xbb\xa3/o/g;
$str =~ s/\xe1\xbb\xa4/U/g;
$str =~ s/\xe1\xbb\xa5/u/g;
$str =~ s/\xe1\xbb\xa6/U/g;
$str =~ s/\xe1\xbb\xa7/u/g;
$str =~ s/\xe1\xbb\xa8/U/g;
$str =~ s/\xe1\xbb\xa9/u/g;
$str =~ s/\xe1\xbb\xaa/U/g;
$str =~ s/\xe1\xbb\xab/u/g;
$str =~ s/\xe1\xbb\xac/U/g;
$str =~ s/\xe1\xbb\xad/u/g;
$str =~ s/\xe1\xbb\xae/U/g;
$str =~ s/\xe1\xbb\xaf/u/g;
$str =~ s/\xe1\xbb\xb0/U/g;
$str =~ s/\xe1\xbb\xb1/u/g;
$str =~ s/\xe1\xbb\xb2/Y/g;
$str =~ s/\xe1\xbb\xb3/y/g;
$str =~ s/\xe1\xbb\xb4/Y/g;
$str =~ s/\xe1\xbb\xb5/y/g;
$str =~ s/\xe1\xbb\xb6/Y/g;
$str =~ s/\xe1\xbb\xb7/y/g;
$str =~ s/\xe1\xbb\xb8/Y/g;
$str =~ s/\xe1\xbb\xb9/y/g;
$str =~ s/\xe1\xbb\xbe/Y/g;
$str =~ s/\xe1\xbb\xbf/y/g;


$str =~ s/\xc3\x9f/s/g;
#    $str =~ s/\xe2\x94\x80/-/g;

    if ( $ori eq $str && ! &containsutf8nonchar($str) && ord($str) != 204  ) {

	$logger->error("utf8_to_ascii() could not transform |$str| (". ord($str).")");
    }
   return($str);
}

sub containsutf8nonchar {
    my ($str) = @_;
    if ( $str =~ /\xe2\x94\x80/ 
	 | $str =~ /\xc2\xab/
	 | $str =~ /\xc2\xbb/
	 | $str =~ /\xca\xbf/
	 | $str =~ /\xcc\x84/
	 ) {

	return 1;
    }
    else {
	return 0;
    }

}

### just for testing purposes:

sub xreverse {
    my $self = shift;
    my ( $string ) = @_;
    
    return ( scalar reverse ($string ) );         

}

sub xsep2revclone {
    my $self = shift;
    my ( $string ) = @_;
    
    if ($string) {
	return ( "$string&&&".$self->xreverse($string) );         
    } else {
	return("");
    }
}


sub xaddblanks {
    my $self = shift;
    my ( $string ) = @_;

    if ($string) {
	return ( join (' ', split(/ */, $string) ) );
    } 
}


sub xcreatemd5 {
    my $self = shift;
    my ( $string ) = @_;

    if ($string) {
	use Digest::MD5 qw(md5 md5_hex md5_base64);
	$string = '{MD5}'.md5_base64($string).'==';
	return ( $string) ;
    } 
    else {
	return('');
    }

}

sub xdummy {
    my $self = shift;
    my ( $string ) = @_;
    if ( $string) {
	return ('dummy');
    }
    else {
	return('');
    }

}



sub xcodes2utf {
    my $self = shift;
    my ( $string ) = @_;

#   if ( $string =~ /Bhakti/ ) {
#	print  "Hare Krishna\n";
#   }


    if ( $self->{data}{'isrealutf8'} ) {
	$logger->debug("this is realutf8");
	return ( $self->xcodes2utf8($string) );
    }

## the following should be renamed to xcodes2htmlentities
    $logger->debug("this is xcodes2utf");
    
    if ( ! $self->{data}{'codetabisloaded'} ) {
	$logger->debug("Loading Code table: isloaded: $self->{data}{'codetabisloaded'}");
	$self->loadcodetable();
	$self->{data}{'codetabisloaded'} = 1;
    }
    
    if ( $string) {
	no warnings;
	if ( $string =~ /[\#\^%ÜÖÄüöäß]/ ) {
	    no warnings;
	    foreach my $code ( reverse sort keys %{$self->{data}{codes}} ) {
#		print "Code: $code\n";
		my $safecode = $self->makeRegexpSafe($code);
		$string =~ s/$safecode/$self->{data}{codes}{$code}/g;
	    }
	    use warnings;
#	    print "converted $string\n";
	}
        use warnings;

	return ($string);
    }
    else {
	return('');
    }
}
 

sub xutf82codes {
    my $self = shift;
    my ( $string ) = @_;

### didn't work: DeviP=Dev^%:A^#.<pur^%:A^%/a^^^1a

#    if ( ! $self->{data}{'utfcodetabisloaded'} ) {
#	$logger->debug("Loading UTF Code table: isloaded: $self->{data}{'utfcodetabisloaded'}");
#	$self->loadutf8codetable();
#	$self->{data}{'utfcodetabisloaded'} = 1;
#    }

    if ( $string) {
#	foreach my $code ( reverse sort keys %{$self->{data}{utf8codes}} ) {
#		print "Code: $code\n";
#	    
#	    my $safecode = $self->makeRegexpSafe($code);
#	    $string =~ s/$safecode/$self->{data}{utf8codes}{$code}/g;
#	}

	$string =~ s/\xc4\x81/%-a/g; # amacron
	$string =~ s/\xc4\xab/%-i/g; # imacron
	$string =~ s/\xc5\xab/%-u/g; # umacron
	$string =~ s/\xc5\x9b/%\/s/g; # s with acute
	$string =~ s/\xc5\x9a/%\/S/g; # S with acute
	$string =~ s/\xe1\xb9\xa3/%..s/g; # n with dor below 
	$string =~ s/\xe1\xb9\x87/%..n/g; # n with dor below 
	$string =~ s/\xe1\xb9\x9b/%..r/g; # r with dor below 
	$string =~ s/\xe1\xb9\x9a/%..R/g; # R with dor below 

	return $string;
    }
    else {
	return '';
    }
}

sub xcodes2utf8 {
    my $self = shift;
    my ( $string ) = @_;

    $logger->debug("this is xcodes2utf8");

    if ( ! $self->{data}{'utfcodetabisloaded'} ) {
	$logger->debug("Loading UTF Code table: isloaded: $self->{data}{'utfcodetabisloaded'}");
	$self->loadutf8codetable();
	$self->{data}{'utfcodetabisloaded'} = 1;
    }

    if ( $string) {
#	if ( $string =~ /[\#\^%ÜÖÄüöäß]/ ) {
        $string =~ s/\.\^-/\. --/g;
        $string =~ s/\^-/--/g;



	if ( $string =~ /[\#\^%]/ ) {
	    $logger->debug("found codes in |$string|"); 
	    foreach my $code ( reverse sort keys %{$self->{data}{codes}} ) {
#		print "Code: $code\n";
		
		my $safecode = $self->makeRegexpSafe($code);
		$string =~ s/$safecode/$self->{data}{codes}{$code}/g;
	    }
	    $logger->debug("converted string to: |$string|\n");
	}

	return ($string);
    }
    else {
	return('');
    }
}

sub loadcodetable {
    my $self = shift;

    my $fh;
    my $filename = $self->{data}{codesfile};

    if ( ! -e $filename ) {
	print "file $filename not found\n";
	exit();
    }
    open ($fh, "<$filename");  

    no warnings;

    while ( <$fh>) {
	if ( /[äöüÄÖÜß\#%\^]/) {
	    if ( /([0-9A-F]{4,4})([=><\^:])(.+)/ ) {
		my $htmlentity = "&#x$1;";
		$self->{data}{codes}{$3}=$htmlentity;
		$self->{data}{revcodes}{$1}=$3;
	    }
	}
    }
    use warnings;
    close($fh);
    return;
}


sub loadutf8codetable {
    my $self = shift;

    my $fh;
    my $filename = $self->{data}{codesfile};

    if ( ! -e $filename ) {
	print "file $filename not found\n";
	exit();
    }
    open ($fh, "<$filename");  

    while ( <$fh>) {
#	if ( /[äöüÄÖÜß\#%\^]/) {
	if ( /[\#%\^]/) {
	    if ( /([0-9A-F]{4,4})([=><\^:])(.+?)([=])(.*?)/ ) {
		my $utf8char = chr (hex($1));
		$self->{data}{codes}{$3}=$utf8char;
		$self->{data}{revcodes}{$1}=$3;
#		$self->{data}{utf8codes}{$utf8char}=$3;
		if ($5) {
		    push @{$self->{data}{altcodes}{$5}}, $utf8char;
		}
	    }
	    elsif ( /([0-9A-F]{4,4})([=><\^:])(.+)/ ) {
		my $utf8char = chr (hex($1));
		$self->{data}{codes}{$3}=$utf8char;
		$self->{data}{revcodes}{$1}=$3;
#		$self->{data}{utf8codes}{$utf8char}=$3;
	    }
	}
    }
    close($fh);
    return;
}


sub htmlutf2codes {
    my $self = shift;
    my ( $string ) = @_;

#    $logger->debug("htmlutf2codes(): starting");
    if ( ! $string ) {
	return ('');
    }

    if ( ! $self->{data}{'codetabisloaded'} ) {
	$logger->debug("Loading Code table: isloaded: $self->{data}{'codetabisloaded'}");
	$self->loadcodetable();
	$self->{data}{'codetabisloaded'} = 1;
    }

#    $string =~ s/&\#x(00)([0-9A-F]{2,2});/chr(hex($2))/ge;

    $string =~ s/&\#x([0-9A-F]{4,4});/$self->{data}{revcodes}{$1}/ge;


#    $logger->debug("htmlutf2codes(): returning $string");

    return $string;

}



sub utf82htmlutf {
    my $self = shift;
    my ( $string ) = @_;

    $logger->debug("utf82htmlutf(): starting");
    if ( ! $string ) {
	return ('');
    }

    if ( $string =~ s/([\xc2-\xe2])(.)/"&\#".uc(sprintf("%x",ord($1))).uc(sprintf("%x",ord($2))).";"/ge ) {
	my $x1 = ord($1);
	my $s1 = sprintf("%x",$x1);
	my $x2 = ord($1);
	my $s2 = sprintf("%x",$x1);
	
	$logger->debug("x1: $x1, s1: $s1, String: $string");
    }
    else {
	$string = '';
    }

    $logger->debug("utf82htmlutf(): returning $string");

    return $string;

}


sub found_in_dictionary {
    my $self = shift;
    my ( $string ) = @_;

    my $equivalent = '';

    $logger->debug("this is found_in_dictionary");

    if ( $string) {

	if ( $self->{data}{'dictionaryfile'} && 
	     (! $self->{data}{'dictionaryfileisloaded'}) 
	     ) {
	    $logger->info("Loading dictionary: $self->{data}{'dictionaryfileisloaded'}");
	    $self->loaddictionary();
	    $self->{data}{'dictionaryfileisloaded'} = 1;
	}

	
	$equivalent = $self->{data}{dictionary}{$string};
	if ( ! $equivalent ) {
	    $equivalent = $self->{data}{dictionary}{lc($string)};
	}
	if ( ! $equivalent ) {
	    $equivalent = $self->lookForSubstringsInDict(lc($string));
	}
	if ( $equivalent ) {

	    $logger->debug("found |$equivalent| from |$string|"); 
	}
	return ($equivalent);
    }
    else {
	return('');
    }
}



sub lookForSubstringsInDict { 
    my $self = shift;
    my ( $string ) = @_;

    my $equivalent = '';

    $logger->debug("this is found_in_dictionary");

    foreach my $key ( keys %{$self->{data}{dictionary}} ) {
	if ( $key =~ /^(.*)($string)(.*)$/ ) {
	    my $val = $self->{data}{dictionary}{$key};

	    $equivalent = substr($val,length($1),length($string));
	    $logger->debug("Found substring of $string in $val: equiv: $equivalent");
	    last;
			   
	} 

    }
    return ( $equivalent );
}

sub loaddictionary {
    my $self = shift;

    my $fh;
    my $filename = $self->{data}{dictionaryfile};

    if ( ! -e $filename ) {
	print "file $filename not found\n";
	exit();
    }
#    open ($fh, "<$filename");  
    open ($fh, "<:utf8", $filename);  

    while ( <$fh>) {
	if ( /^(.+?)=(.+)/ ) {
#	    $logger->debug("***************** $1 = $2");
#	    eval ( decode_utf8($1));
#	    $self->{data}{dictionary}{$2}=$1;
	    $self->{data}{dictionary}{lc($1)}=$2;
	}
    }
    close($fh);
    return;
}


sub xcreatevaluefromLDAP {
    my $self = shift;
    my ( $string, $rh_paras ) = @_;

    if ( ! $string ) {
	return ('');
    }
	
    my $searchattr;
    
    $string =~ s/\+/;/g;

    my ($searchval, $attrname, $is_iso_to_utf, $function, $is_multi) = split /;/, $string, 5 ;

#    print "xcreatevaluefromLDAP called with paras:\n   1.) $searchval\n";  
#    print "   2.) $attrname\n   3.) $otherparams\n   4.) [$rh_paras->{parameter2}]\n";

    $attrname =~ s/^'(.*)'$/$1/;
 
    $logger->debug("xcreatevaluefromLDAP called with paras:");
    $logger->debug("   1.) searchval: $searchval");  
    $logger->debug("   2.) attrname: $attrname");
    $logger->debug("   3.) is_iso_to_utf: $is_iso_to_utf") if $is_iso_to_utf;
    $logger->debug("   4.) function: $function") if $function;
    $logger->debug("   5.) is_multi: $is_multi") if $is_multi;
    $logger->debug("   6.) parameter2: [$rh_paras->{parameter2}]");
	
    
    if ( $rh_paras->{parameter2} =~ /^(ldap):/ ) {
	my $prot = $1;
	
	if ( ! $self->{valueldapconnection} ) {
	    $self->{valueldapserverdef} = 
		$libldap->defineServerFromURI($rh_paras->{parameter2},
					      'accountdatasource');
	    $logger->debug("setting up connection to accountdata source");
	    $self->{valueldapconnection} = 
		$libldap->connectServer($self->{valueldapserverdef});
	}
	
	if ( $prot eq 'ldap' ) {
	    $searchattr = $data->getTokenValue($rh_paras->{parameter2},
					     "searchattr=", ";");
#	    $mappingattr = $data->getTokenValue($rh_paras->{parameter2},
#						"mapping=", ";");


	} else {
	    $logger->error("Wrong protocol in uri! Giving up conversion");
	    return $string;
	}
    } 
    my $searchtype = 'exact';
    
    my $newvalue = $libldap->searchValueFromLDAP($attrname, $searchval, 
						 $searchattr,
						 $rh_paras->{parameter2},
						 $searchtype, 
						 'acountdatasource',
						 $self->{valueldapconnection},
						 $self->{valueldapserverdef},
						 $is_multi
						 );

#    print "newvalue befor: $newvalue\n";
    if ( $newvalue) {
	if ( $is_iso_to_utf ) {
	    $newvalue = $self->convert('iso-to-utf', $newvalue);
	}
	if ( $function ) {
	    $newvalue = $self->convert($function, $newvalue);
	}

    }
#    print "newvalue after: $newvalue\n";
#    exit();
    return($newvalue);
}

sub connectLdapServer {

 my $self = shift;

 my ($prot, $host, $port, $version, $binddn, $pw, $basedn, $scope, $timelimit) = @_;

 my $ldap = DAASIlib::LDAP->new();

 my $def = $ldap->defineServer($prot, $host, $port, $version, $binddn, $pw, $basedn, $scope, undef, undef, undef, $timelimit);

 $self->{ 'ldapconnection' } = $ldap->connectServer( $def );

 return ($self->{ 'ldapconnection' });

}

##################################################################
# Name: percent_encode
#
# Description:
# Encodes the input as percent
#
# Input:    $string: the string to encode
#
# Output:   the encoded string
#
# Created: TZ 2017-11-13
# Changes:
#
######################################################################
sub percent_encode {
    my $self = shift;
    my ($string) = @_;
    $logger->debug("url_encode(): string to encode: $string");
    $string = uri_escape($string);
    $logger->debug("url_encode(): encoded string: $string");
    return $string;
}



1;

# some old stuff
#		my $first = substr($1,0,2);
#		my $second = substr($1,2,2);
#		my $utf = chr(hex($first)).chr(hex($second));
#		my $utf16 = chr(hex($1));
#		my $utf8 = decode("UTF-32", $utf16);
#		my $utf8 = $self->convert("UTF-16-to-UTF-8", $utf16);
#		if ($utf) {
#		    $self->{data}{codes}{$3}=$utf;
#		    print "|$3| |$2| |$utf8| ($first $second)\n";
#		}
#		else {
#		    print "could not convert $3 ($1)\n";
#		}

#		my $utfchar = chr($1);
#		my $utfchar = chr("\x{$1}");
#		my $coderef = "0x$1";
#		$coderef = $coderef+0;
#		my $utfchar = pack("U0U*",$coderef);
#		print "UTF of $1: $utfchar\n";
#		last;

#		$self->{data}{codes}{$3}="\x{$1}";
#		$self->{data}{codes}{$3}=chr("\x{$1}");
#		$self->{data}{codes}{$3}=chr($1);
#		$self->{data}{codes}{$3}=chr("0x$1");
#		$self->{data}{codes}{$3}=hex($1);
#		$self->{data}{codes}{$3}=chr(hex($1));
#		$self->{data}{codes}{$3}=$1;

#    use Data::Dumper;
#    print Dumper($self->{data}{codes});
#    foreach my $code ( keys %{$self->{data}{codes}} ) {
#	print "code: $code\n";
#    }



__END__
# Below is stub documentation for your module. You'd better edit it!

=head1 NAME

DAASIlib::Convert - Perl extension for blah blah blah

=head1 SYNOPSIS

  use DAASIlib::Convert;
  blah blah blah

=head1 DESCRIPTION

Stub documentation for DAASIlib::Convert, created by h2xs. It looks like the
author of the extension was negligent enough to leave the stub
unedited.

Blah blah blah.

=head2 EXPORT

None by default.



=head1 SEE ALSO

Mention other useful documentation such as the documentation of
related modules or operating system documentation (such as man pages
in UNIX), or any relevant external documentation such as RFCs or
standards.

If you have a mailing list set up for your module, mention it here.

If you have a web site set up for your module, mention it here.

=head1 AUTHOR

Peter Gietz, E<lt>zrngi01@suse.deE<gt>

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2005 by Peter Gietz

This library is free software; you can redistribute it and/or modify
it under the same terms as Perl itself, either Perl version 5.8.1 or,
at your option, any later version of Perl 5 you may have available.


=cut
