#######################################################################
# Copyright: DAASI International GmbH 2018. All rights reserved.
#
# This is Open Source Software
# License: Artistic License 2.0
#
# Author: Peter Gietz, DAASI International GmbH, www.daasi.de
# For questions please mail to peter.gietz@daasi.de
#######################################################################
#!/usr/bin/perl -w


package DAASIlib::DBI;

use strict;
use warnings;
use vars qw($VERSION);

our $VERSION = "0.12";

use DBI;

use Log::Log4perl qw(:levels);
my $logger = Log::Log4perl->get_logger("main"); 
use DAASIlib::Data;
my $data = new DAASIlib::Data;



sub new {
    my $class;
    my $self;

    my ( $this , $param ) = @_;
    $class = ref($this) || $this;

    $self={
	data => {},
    };

    bless($self , $class );

    return $self;

}

sub init {
    my $self = shift;

    my ($is_verbose) = @_;

    $self->{data}{is_verbose} = $is_verbose;

}

sub defineSQLServerFromURI {
    my $self = shift;
    my ( $uri, $sql_statement, $sql_filename, $pwtoken ) = @_;

    ## URI format:  
    #print ("## dbi:<dbd>:host=<host>;port=<port>;database=<database>;user=<user>\n");
    if ( ! ( $uri =~ s{^dbi:(.*)}{$1}i ) ) {
	$logger->logdie("inputuri has to start with 'dbi:'");
    } 

    my ( $dbd, $params ) = split /:/, $uri;
    my ( $host, $port, $database, $user, $passwd, $is_printerror, 
	 $is_autocommit );
    my %dbi_info;
    
    $dbi_info{dbd}=$dbd;

    if ( $sql_statement ) {
	$dbi_info{statement}=$sql_statement;
    } elsif ( $sql_filename ) {
	$dbi_info{statement}=$self->loadSQLStatement( $sql_filename );
    } else {
	$logger->debug("No SQL statement found in database definition");
    }

    foreach my $param ( split /;/, $params ) {
	if ( $param =~ /^host=(.*)/ ) { $dbi_info{host}=$1; }
	elsif ( $param =~ /^port=(.*)/ ) { $dbi_info{port}=$1; }
	elsif ( $param =~ /^driver=(.*)/ ) { $dbi_info{driver}=$1; }
	elsif ( $param =~ /^server=(.*)/ ) { $dbi_info{server}=$1; }
	elsif ( $param =~ /^charset=(.*)/ ) { $dbi_info{charset}=$1; }
	elsif ( $param =~ /^database=(.*)/ ) { $dbi_info{database}=$1; }
	elsif ( $param =~ /^user=(.*)/ ) { $dbi_info{user}=$1; }
	elsif ( $param =~ /^table=(.*)/ ) { $dbi_info{table}=$1; }
	elsif ( $param =~ /^uidattribute=(.*)/ ) { $dbi_info{uidattribute}=$1; }
	elsif ( $param =~ /^mailattribute=(.*)/ ) { $dbi_info{mailattribute}=$1; }
	elsif ( $param =~ /^mapping=(.*)/ ) { $dbi_info{mapping}=$1; }
	elsif ( $param =~ /^printerror=(.*)/ ) { $dbi_info{is_printerror}=$1; }
	elsif ( $param =~ /^autocommit=(.*)/ ) { $dbi_info{is_autocommit}=$1; }
	elsif ( $param =~ /^do=(.*)/ ) { $dbi_info{'do'}=$1; }
	else { $logger->error("found unknown parameter in dbi-uri"); }
    }
	

    if ( ! $dbi_info{is_printerror} ) { $dbi_info{is_printerror}=0;}
    if ( ! $dbi_info{is_autocommit} ) { $dbi_info{is_autocommit}=0;}
	
    if ( $dbi_info{user} ) {
	$dbi_info{passwd} = $data->getSecret($pwtoken); 
    }

#print "dbi:$dbi_info{dbd}:$dbi_info{database}\@ $dbi_info{host}:$dbi_info{port}/$dbi_info{user}?$dbi_info{passwd}?$dbi_info{is_autocomit},$dbi_info{is_printerror}\n";

#    use Data::Dumper;
#    print Dumper(\%dbi_info);
    
    return (\%dbi_info);
} 


sub getFieldnamesFromSQLString {
    my $self = shift;
    my ($string, $filename) = @_;
    my @fieldnames = ();;

    if ( ! $string ) {
	$string =  $self->loadSQLStatement($filename);
    }

    if ( $string =~ s/SELECT\s?(.+)\s?FROM.+/$1/ ) {
	$string =~ s/\n//g;
	$string =~ s/\s//g;
	if ( $string ne '*' ) {
	    @fieldnames = split /,/, $string;
	}
    }

    return( @fieldnames );

}




sub loadSQLStatement {
    my $self = shift;
    my ( $sql_filename ) = @_;
    my $statement = '';

    if ( $sql_filename ) {
	$logger->debug("loading SQL statement from file $sql_filename");

	open SQLFH, "<$sql_filename" || 
	    $logger->logdie ("couldn't open file  $sql_filename for read");
	while (<SQLFH>) {
	    chomp;
	    $statement .= "$_ ";
	}
    }

    return($statement);
}


sub createCSVFileFromSQL {
    my $self = shift;
    my ($filename, $rh_sqlserver, $sep, $quote) = @_;

    my $dsn = $self->createDsn( $rh_sqlserver, $sep, $quote);

    $logger->info("Connecting to: $dsn");
    print "   * Connecting to DSN: [$dsn]\n";

    my $dbh = $self->connect($dsn, $rh_sqlserver);

    if ( $rh_sqlserver->{'do'} ) {
	$dbh->do($rh_sqlserver->{'do'});
    }
    my $formatedSQLstatement = 
	$self->formatSQLStatement($$rh_sqlserver{statement});
    $logger->info("preparing statement: $formatedSQLstatement");
#    $logger->info("preparing statement: $$rh_sqlserver{statement}");

# print ("preparing statement: $$rh_sqlserver{statement}\n");
    
    my $sth = $self->prepare($dbh, $rh_sqlserver); 

    open (FILE, ">$filename") ||
	$logger->logdie("Couldn't open CSV intermediate  file  $filename");

    my $rowscount = 0;

    $self->sqlExecute($sth, $dbh, 1);

    my $conf = new DAASIlib::CONF;
    if ($conf->is_debug ){
	$self->sqlDump($sth, $filename);
	$self->sqlExecute($sth, $dbh, 1);
    }

    while (my @row =  $sth->fetchrow_array() ) {
	my $fieldcount = 0;
	foreach my $field (@row) {
	    if ($fieldcount) {
		print FILE $sep;
	    }
	    if (defined($field)) {
		if ( $field =~ /$quote/ ) {
		    $field =~ s/$quote/\\$quote/g;
		}
		if ( $field =~ /$sep/ ) {
		    $field = $quote.$field.$quote;
		}
		print FILE "$field";
	    } else {
#		print "";
	    }
	    $fieldcount++;
	}

	$logger->warn("Problem in fetchrow_array(): ", $sth->errstr()) 
	    if $sth->errstr();

	print FILE "\n";

	$rowscount++;
    }

    if ( $rowscount) {
	$logger->info("wrote $rowscount rows to CSV file $filename");
    } else {
	$logger->warn("SQL statement with no results");
    }

    close(FILE);

    $self->disconnect($dbh);

    return($filename);
}

sub formatSQLStatement {
    my $self = shift;
    my ($sqlstatement) = @_;

    my $formattedStatement;

    my @newlinetokens = ( 'FROM', 'WHERE', 'AND', 'ORDERED');

    $sqlstatement =~ s/,\s*/,\n\t/g;

    foreach (@newlinetokens) {
	$sqlstatement =~ s/$_/\n$_/g; 
    }

    return ($sqlstatement);
}


sub searchValueFromSql {
    my $self = shift;
    my ( $attr, $mapvalue, $mapattr, $uri, $searchtype, $pwtoken ) = @_;

    my $cmp = "";
    # Is the following realy a good idea? by now I think so.
    if ($searchtype eq "str" ) {
	$cmp = "LIKE";
	$mapvalue = "'$mapvalue'";
    } elsif ($searchtype eq "sub" ) {
	$cmp = "LIKE";
	$mapvalue = "'%$mapvalue%'";
    } elsif ($searchtype eq "presub" ) {
	$cmp = "LIKE";
	$mapvalue = "'%$mapvalue'";
    } elsif ($searchtype eq "postsub" ) {
	$cmp = "LIKE";
	$mapvalue = "'$mapvalue%'";
    } elsif ( $searchtype eq "num" ) {
	$cmp = "=";
    } else {
	$logger->error("unknown searchtype $searchtype specified");
    }


    my $table = $data->getTokenValue( $uri, "table=", ";");
    my $sqlstatement = 
	"SELECT $attr FROM $table WHERE $mapattr $cmp $mapvalue";
 
    my $rh_sqlserver = $self->defineSQLServerFromURI($uri, $sqlstatement, 
						     "", $pwtoken);
    my $dsn = $self->createDsn( $rh_sqlserver, ',', '"');
    $logger->info("Connecting to: $dsn");
#    print "   * Connecting to DSN: $dsn\n";
    my $dbh = $self->connect($dsn, $rh_sqlserver);
    $logger->info("preparing statement: $$rh_sqlserver{statement}");
#    print "   * Preparing statement: $$rh_sqlserver{statement}\n";
    my $sth = $self->prepare($dbh, $rh_sqlserver); 
    $self->sqlExecute($sth, $dbh, 1);

    ## I only expect one value here:
    my ($value) = $sth->fetchrow_array();

    $sth->finish();
    $self->disconnect($dbh);

    return ($value);

}


sub performSearch {
    my $self = shift;
    my ( $uri, $sqlstatement, $pwtoken ) = @_;

    my $max = 100;

    my $rh_sqlserver = $self->defineSQLServerFromURI($uri, $sqlstatement, 
						     "", $pwtoken);
    my $dsn = $self->createDsn( $rh_sqlserver, ',', '"');
                                               #$sep, $quote);
    $logger->info("Connecting to: $dsn");
    print "   * Connecting to DSN: $dsn\n" if $self->{data}{is_verbose};
    my $dbh = $self->connect($dsn, $rh_sqlserver);
#    print "pwtoken: |$pwtoken| uri: |$uri| statement: |$sqlstatement|\n";
    $logger->debug("preparing statement: $$rh_sqlserver{statement}");
    my $sth = $self->prepare($dbh, $rh_sqlserver); 
    $self->sqlExecute($sth, $dbh, 1);

    my $i = 0;
    my @result;
    my $row_ref;
    while ( $row_ref = $sth->fetchrow_arrayref() ) {	
	push @result, [ @{$row_ref} ];
	$i++;
	if ( $i > $max ) {
	    print "  * more than max rows retrieved, thus stopping retrieving data here\n" if $self->{data}{is_verbose};
	    last;
	}
    }

    $sth->finish();
    $self->disconnect($dbh);

    return (\@result);

}


sub loadValuesFromSql {
    my $self = shift;
    my ( $ra_list, $uri, $attr, $pwtoken ) = @_;

    my $table = $data->getTokenValue($uri, "table=", ";");
    
    my $field;
    if ( $attr) {
	$field = $attr;
    } else {
	my $field = $data->getTokenValue($uri, "attribute=", ";");
    }


    my $rh_sqlserver = $self->defineSQLServerFromURI($uri, "SELECT $field FROM $table", '', $pwtoken);
    my $dsn = $self->createDsn( $rh_sqlserver, ',', '"');
                                               #$sep, $quote);
    $logger->info("Connecting to: $dsn");
    print "   * Connecting to DSN: $dsn\n";
    my $dbh = $self->connect($dsn, $rh_sqlserver);
    $logger->info("preparing statement: $$rh_sqlserver{statement}");
    my $sth = $self->prepare($dbh, $rh_sqlserver); 
    $self->sqlExecute($sth, $dbh, 1);

    my $entry_count = 0;

    while (my @values =  $sth->fetchrow_array() ) {
	if ( $values[0] ) {
	    $logger->debug ("adding values to unique list [@values]");
	    push @{$ra_list}, @values;
	    $entry_count ++;
	}
    }
     
    if ( ! $entry_count) {
	$logger->warn("No entries found");
    }

    $self->disconnect($dbh);
}




sub connect {
    my $self = shift;
    my ( $dsn, $rh_sqlserver ) = @_;
#    print "is_auto: $$rh_sqlserver{is_autocommit}\n";
    my $dbh = DBI->connect($dsn, $$rh_sqlserver{user}, $$rh_sqlserver{passwd},{
	PrintError => $$rh_sqlserver{is_printerror},
	AutoCommit => $$rh_sqlserver{is_autocommit} })
	|| $logger->logdie("Error while connecting SQL database $dsn user=$$rh_sqlserver{user}: $DBI::errstr");
    return ($dbh);

}


sub prepare {
    my $self = shift;
    my ($dbh, $rh_sqlserver) = @_;

    my $sth = $dbh->prepare($$rh_sqlserver{statement}) 
	|| $logger->logdie("Error while preparing SQL statement: ".
			   $dbh->errstr() );

    return($sth);

}

sub sqlDump {
    my $self = shift;
    my ($sth, $filename) = @_;

    open (DUMPFILE, ">${filename}.dump") ||
	$logger->logdie("Couldn't open CSV debug dump file  ${filename}.dump");
    my $rows = $sth->dump_results(200, "\n", ',', \*DUMPFILE); 
    close(DUMPFILE);
    $logger->info("Dumped SQL file ($rows rows) because debug flag was set");
}

sub disconnect {
    my $self = shift;
    my ($dbh) = @_;

    $dbh->disconnect() || 
	$logger->warn("Failed to discoonnect: ", $dbh->errstr() );


}


sub sqlExecute {
    my $self = shift;
    my ($sth, $dbh, $is_die) = @_;

    if ( ! $sth->execute() ) { 
	my $errstr = $sth->errstr();
	$logger->error("Error while executing SQL statement: ".
			$errstr );

	if ( $is_die ) {
	    $dbh->disconnect() || 
		$logger->warn("Failed to discoonnect: ", $dbh->errstr() );
	    $logger->logdie("Exiting: SQL error: $errstr");
	}
    }

}



sub createDsn {
    my $self = shift;
    my ($rh_sqlserver) = @_;
    my $dsn = "DBI:$$rh_sqlserver{dbd}:";

    if ($$rh_sqlserver{server}) {
	$dsn.="server=$$rh_sqlserver{server};";
    } else {

	if ($$rh_sqlserver{host}) {
	    $dsn.="host=$$rh_sqlserver{host};";
	}

	if ($$rh_sqlserver{port}) {
	    $dsn.="port=$$rh_sqlserver{port};";
	}

	if ($$rh_sqlserver{driver}) {
	    $dsn.="driver=$$rh_sqlserver{driver};";
	}

    }


    if ($$rh_sqlserver{database}) {
	$dsn.="database=$$rh_sqlserver{database};";
    }


    if ($$rh_sqlserver{charset}) {
	$dsn.="charset=$$rh_sqlserver{charset};";
    }
    
    return($dsn);
}

sub enhanceSQLStatement {

    my $self = shift;
    my ($sqlstatement, $newcondition) = @_;


    my ($select, $condition, $rest);
    if ( $sqlstatement =~ /WHERE/ ) {
	$newcondition= " AND $newcondition";
	if ( $sqlstatement =~ /^(.+)WHERE(.+)ORDER(.+)/ ) {
	    $select = $1; 
	    $condition = $2;
	    $rest = "ORDER$3";
	}
	elsif ( $sqlstatement =~ /^(.+)WHERE(.+)/ ) {
	    $select = $1; 
	    $condition = $2;
	    $rest = q();		
	}
	$sqlstatement = "${select}WHERE$condition$newcondition$rest"; 
    }
    else {
	$sqlstatement = "${sqlstatement} WHERE $newcondition";
    }

    return ($sqlstatement);

}

1;
