#######################################################################
# Copyright: DAASI International GmbH 2018. All rights reserved.
#
# This is Open Source Software
# License: Artistic License 2.0
#
# Author: Peter Gietz, DAASI International GmbH, www.daasi.de
# For questions please mail to peter.gietz@daasi.de
#######################################################################
#!/usr/bin/perl -w


package DAASIlib::Data;

use strict;
use warnings;
use vars qw($VERSION);


## not needed here: 
## use DBI;
use Log::Log4perl qw(:levels);
use File::Basename;
#use Sys::Hostname;
use Socket;
#use Net::Nslookup;
use Encode qw(encode decode);

my $logger = Log::Log4perl->get_logger(''); 

our $VERSION = "0.10";

our $PROGNAME;
our $SECRETFILE;

sub new {
    my $class;
    my $self;

    my ( $this , $param ) = @_;
    $class = ref($this) || $this;

    $self={
	data => {},
    };

    bless($self , $class );

    return $self;

}


sub init {
    my $self = shift;
    my ($progname, $pwfile) = @_;

#    print "Data.pm: Setting progname $progname\n";
    $PROGNAME = $progname;
    $SECRETFILE = $pwfile

}

sub progname {
    return $PROGNAME;
}

sub test {
    print "########### TEST\n";
}

sub getProgramFiles {
    my $self = shift;
    my ( $call, $addtoetc, $etcpath ) = @_;

    my ($progname, $progpath, $progsuffix, $etcdir, $sysconfig);

    ($progname, $progpath, $progsuffix) = fileparse($call, ".pl");

    if ( $progpath =~ /^\.[^.]/ ) {
	my $longpath = `pwd`;
	chomp($longpath);
	$progpath =~ s/^./$longpath/;
    }
    if ( $addtoetc ) {
	$etcdir = "${progpath}${addtoetc}etc/";
	if ( ! -e "${etcdir}${progname}.sys" ) {
	    $etcdir = "${progpath}${addtoetc}${addtoetc}etc/";
	    if ( ! -e "${etcdir}${progname}.sys" ) {
		print "sysfile not found\n";
	    }
	}

    }
    elsif ($etcpath) {
	$etcdir = $etcpath;
    } 
    else {
	$etcdir = "${progpath}etc/";
    }

    $sysconfig = "${etcdir}${progname}.sys";

    return ($progname, $progpath, $etcdir, $sysconfig);

}



sub getTokenValue {
    my $self = shift;
    my ( $string, $token, $endvaluetoken ) = @_;

    if ( $string !~ /$endvaluetoken$/ ) { 
	$string.=$endvaluetoken;
    }

    my $value="";

    if ( $string  =~ /$token(.*?)$endvaluetoken/ ) {
	$value = $1;
    } else {
	$logger->warn("Token $token missing in dbi:-parameters");
    }

    return($value);
}

sub getTokenValueFromFile {
    my $self = shift;
    my ( $token, $filename ) = @_;

    my $value;
    my $is_open=0;

    if ( ! -e $filename ) {
	$logger->error("File $filename does not exist!");
    } else {
	$is_open = open ( FH, "<$filename") || 
	    $logger->error("Couldn't open file $filename!");
    }

#    print "is_open: [$is_open]\n";
    if (  $is_open ) {
	while ( <FH> ) {
	    if ( $_ =~ /$token\s*(.*)/ ) {
		$value = $1;
		last;
	    }
	}
	close (FH);
    }

    if (! $value ) {
	$logger->error("found empty value of token $token in file $filename");
    } else {
	$logger->debug("found value of token $token in file $filename");
    }

    return($value);
}




sub getSecret {
    my $self = shift;
    my ( $param ) = @_;

    $logger->debug("getSecret $param from $SECRETFILE");
#    my $secretfile = "$PROGNAME.secret";
    my $secret;

    if ( ! -e $SECRETFILE ) {
	$logger->logdie("File $SECRETFILE does not exist!");
    }
    open ( FH, "< $SECRETFILE") || 
	$logger->logdie("Couldn't open secret file $SECRETFILE!");
 
    while ( <FH> ) {
	if ( $_ =~ /$param\s+(.*)/ ) {
	    $secret = $1;
	    last;
	}
    }

    close (FH);

    if (! $secret ) {
	$logger->error("found empty or no value for $param in secret file $SECRETFILE");
#	print "may be token $param is missing in file $SECRETFILE\n";
    } else {
	$logger->debug("found pw of $param in secret file $SECRETFILE");
    }

    return($secret);
}

### CHANGES
#   MH 2017-10-20 did not handle empty values, i.e. insuitable for CSV import
#                 now always returns 0 or 1, and not some unforseeable value
sub areTheyDifferent {
    my $self = shift;
    my ($ra_newvalues, $ra_oldvalues, $is_casesensitive) = @_;

#   my @diff = ();
    my $aredifferent = 0;
    my %count;
    my $e;


    if ( $is_casesensitive) {
	foreach $e ( @{$ra_newvalues}, @{$ra_oldvalues}) {
	    $count{$e}++;
	}
    }
    else {
	foreach $e ( @{$ra_newvalues}, @{$ra_oldvalues}) {
	    $count{lc($e)}++;
	}
    }

    foreach $e ( keys %count) {
	if ( $count{$e} != 2 ) {
	    $logger->debug("found difference in value $e which occurs $count{$e} instead of 2");
#           push @diff, $e;
	    $aredifferent = 1;
	    last;
	}
    }

    # return ($diff[0] ); 
    return $aredifferent; 
}

sub arrayToLower {
    my $self = shift;
    my ($ra_values) = @_;

    foreach (@{$ra_values}) {
	$_ = lc($_);
    }

    return ($ra_values);
}

sub findNewValues {
    my $self = shift;
    my ($ra_newvalues, $ra_oldvalues, $is_caseignore) = @_;

    # only give back those values in newvalues that are not in oldvalues
 #   print "findNewValues got new: [@{$ra_newvalues}], old: [@{$ra_oldvalues}]\n";
    my %seen;

    if ( $is_caseignore ) {
	@seen{@{$self->arrayToLower($ra_newvalues)}} = ();
	delete @seen{@{$self->arrayToLower($ra_oldvalues)}};
    } 
    else {
	@seen{@{$ra_newvalues}} = ();
	delete @seen{@{$ra_oldvalues}};
    }
    my @newonly = keys %seen;
#    print "findNewValues returns: [".join ('|', @newonly)."]\n";
    return (@newonly);
}

sub findIntersection {
    my $self = shift;
    my ($ra_newvalues, $ra_oldvalues, $is_caseignore ) = @_;

    # only give back those array elements that are both in newvalues and in oldvalues
    my @a;
    my @b;

    @a = $is_caseignore ? @{$self->arrayToLower($ra_newvalues)} 
				: @{$ra_newvalues} ;
    @b = $is_caseignore ? @{$self->arrayToLower($ra_oldvalues)} 
				: @{$ra_oldvalues} ;

    @a = $self->makeUnique(@a);
    @b = $self->makeUnique(@b);
    my %union=();
    my %isect = ();
    my @isect=();
    my $e;

    foreach $e (@a) { $union{$e} = 1; }
    foreach $e (@b) {
	if ( $union{$e} ) {  $isect{$e} = 1; }
	$union{$e} = 1;
    }

    @isect = keys %isect;

#    print "isect: [@isect]\n";
    return (@isect);
}



sub makeUnique {
    my $self = shift;
    my ( @nonunique ) = @_;

    my %union=();

    foreach my $e (@nonunique) {
	$union{$e} = 1; 
    }

    my @unique = keys %union;

    return(@unique);
}

sub removeElements {
    my $self = shift;
    my ( $ra_unclean, $pattern ) = @_;
    my %union;
#    print "Pattern: /$pattern/\n";
    if ( $pattern =~ /^\/(.*)\/$/) { $pattern = $1; }

     foreach my $e (@{$ra_unclean}) {
	 if ( $e !~ /$pattern/ ) {
	     $union{$e} = 1; 
	 }
    }
    return ( keys %union);

}


sub getmax {
    my $self = shift;
    my @values = @_;
    
    my $max = shift @values;
    
    foreach ( @values ) {
	if ( $_ > $max ) {
	    $max = $_;
	}
    } 
    return $max;
}


sub maxLen {
    my $self = shift;
    my ($ra_strings, $maxlen) = @_;

#    use Data::Dumper;
#    print Dumper($ra_strings);
#    exit();
    my @array = @{$ra_strings};

    if ( $maxlen) {
	$maxlen = length(shift(@array));
    }
    else { $maxlen = 0; }

    foreach (@array) {
	$maxlen = length($_) > $maxlen ? length($_) : $maxlen;
    }
    return ($maxlen);
}


sub hideMail {
    my $self = shift;
    my ($mail, $img) = @_;

    my @parts = split /\@/, $mail;
    my $hidden_mail = $parts[0].'<img src="' . 
	$img . 
	'" border="0" style="position:relative;top:3px">'.$parts[1];
    
    return ($hidden_mail);
}


sub fileTimeStamp {

#Did this ever work without  my $self = shift;?? Not sure where it is used
# thus created a new method below
    my ( $sek, $min, $hour, $day, $mon, $year ) = localtime();
    my $filetimestamp = sprintf("%4d%02d%02d-%02d%02d%02d", $year+1900, 
				$mon+1, $day, $hour, $min, $sek);	    

    return($filetimestamp);
}

sub getFileTimeStamp {
    my $self = shift;
    my ( $sek, $min, $hour, $day, $mon, $year ) = localtime();
    my $filetimestamp = sprintf("%4d%02d%02d-%02d%02d%02d", $year+1900, 
				$mon+1, $day, $hour, $min, $sek);	    

    return($filetimestamp);
}


sub containsarrayelement {
    my $self = shift;
    my ( $string, $arrptr ) = @_;

    my $contains = 0;

    foreach ( @{$arrptr} ) {
	if ( $string =~ /$_/ ) {
	    $contains = 1;
	}
    }

    return $contains;
}


sub incrementfile {
    my $self = shift;
    my ($filename, $len) = @_;
    # stolen from perl cook book:
    use Fcntl qw(:DEFAULT :flock);

    sysopen(FH, $filename, O_RDWR|O_CREAT) 
	or die "Kann $filename nicht oeffnen: $!";

    flock(FH, LOCK_EX)
	or die "kein Schreiblock auf $filename: $!";

    my $number = <FH> || 0;

    seek(FH,0,0)
	or die "kann $filename nicht neu positionieren: $!";

    truncate (FH,0)
	or die "kann $filename nicht neu positionieren: $!";

    $number++;
    if ($len) {
	$number = sprintf("%0${len}d",$number);
    }

    print FH $number."\n"
	or die "kann $filename nicht schreiben: $!";

    close (FH)
	or die "kann $filename nicht schliessen: $!";

    return $number;

}


sub getpid {
    my $self = shift;
    my ( $pidfile ) = @_;
     open FH, "<$pidfile" or print  "   couldn't open pid file $pidfile\n";
    $/ = "";
    my $pid = <FH>;
    chomp($pid);

    $/="\n";
    return $pid;
}

sub gethost {
    my $self = shift;

#    my $host = hostname;
    my $ip = $self->getip();
    my $iaddr = inet_aton($ip);
    my $host = gethostbyaddr($iaddr, AF_INET);
#    my $host = nslookup(host => $ip, server => '192.168.100.16');
    return $host;
}

sub getip {
    my $self = shift;
    my $ip;

    my $interface;
    my %IPs;

    foreach ( qx{ (LC_ALL=C /sbin/ifconfig -a 2>&1) } ) {
        $interface = $1 if /^(\S+?):?\s/;
        next unless defined $interface;
        $IPs{$interface}->{STATE}=uc($1) if /\b(up|down)\b/i;
        $IPs{$interface}->{IP}=$1 if /inet\D+(\d+\.\d+\.\d+\.\d+)/i;
    }
    
    if ( $IPs{'eth0'} ) {
	$ip = $IPs{'eth0'}{IP};
    }
    else {
	foreach ( keys %IPs ) {
	    if ( $_ ne 'lo' ) {
		$ip = $IPs{$_}{IP};
	    }
	}
    }

    if ( ! $ip ) {
	$ip = $IPs{'lo'}{IP};
    }

    if ( ! $ip ) {
	$ip = '127.0.0.1';
    }

    return $ip;
}

sub getRandNumString {
    my $self = shift;
    my ($len) = @_;
    return $self->getRandString($len,"48","57");
}

sub getRandAlphaAllString {
    my $self = shift;
    my ($len) = @_;
    return $self->getRandString($len,"65;97","90;122");
}

sub getRandAlphaNumAllString {
    my $self = shift;
    my ($len) = @_;
    return $self->getRandString($len,"48;65;97;48","57;90;122;57");
}


sub getRandAlphaCapitalString {
    my $self = shift;
    my ($len) = @_;
    return $self->getRandString($len,"65","90");
}

sub getRandAlphaNumCapitalString {
    my $self = shift;
    my ($len) = @_;
    return $self->getRandString($len,"48;65;48","57;90;57");
}

sub getRandAlphaSmallString {
    my $self = shift;
    my ($len) = @_;
    return $self->getRandString($len,"97","122");
}

sub getRandAlphaNumSmallString {
    my $self = shift;
    my ($len) = @_;
    return $self->getRandString($len,"48;97;48","57;122;57");
}

sub getRandHexString {
    my $self = shift;
    my ($len) = @_;
    return $self->getRandString($len,"48;65","57;70");
}

sub getRandString {
    my $self = shift;
    my ( $len, $startdefs, $enddefs ) = @_;

    my $string='';
    
    if ( $startdefs && $enddefs ) {
	my $allowed='';
	my @starts = split / ?; ?/, $startdefs;
	my @ends = split / ?; ?/, $enddefs;
	my $max = $#starts; 
##	print "Max: $max\n";
	for ( my $n = 0; $n <= $max; $n++ ) {
	    for (my $c = $starts[$n]; $c <= $ends[$n]; $c++ ) {
		my $char = chr(int($c));
		$allowed.=$char;
	    }
	}
	my $allowedlen = length($allowed);
##	print "Allowed: $allowed\n";
##	print "len: $allowedlen\n";

	for (my $i = 0; $i < $len; $i++) {
	    my $x = int(rand($allowedlen));
	    my $char;
	    $char = substr($allowed,$x,1);
	    $string.=$char;
	}
    }
    else {
	for (my $i = 0; $i < $len; $i++) {
	    $string.=$self->getrandchar();
	}
    }

    return $string;

}


sub getrandchar {
    my $self = shift;

    my $char ='';
    my $x = int(rand(10));

    if ( $x > 4 ) {
	$char = chr(int(rand(91-33)+33));
    }
    else {
	$char = chr(int(rand(126-97)+97));
    }

    return $char;
}


##################################################################
# Name: utf8substr
#
# Description:
# Works just like perl func substr, but also works with utf8 strings 
# "Extracts a substring out of EXPR and returns it.  First character is at 
# offset 0, or whatever you've set $[ to (but don't do that).  If OFFSET 
# is negative (or more precisely, less than $[), starts that far from the 
# end of the string.  If LENGTH is omitted, returns everything to the end 
# of the string.  If LENGTH is negative, leaves that many characters
# off the end of the string."
#
# Input: 
# 1.) $expr: string that may contain utf8
# 2.) $offset: offset where the substring starts
# 3.) $length: (Optional) length of the substrstring 
# 4.) $replacement: (Optional) string. This allows you to replace parts of 
#     the EXPR and return what was there before in one operation
# 
# Output:
# substring
#
# Created: PG 2013-07-12
# Changes: 
# 
######################################################################
sub utf8substr {
    my $self = shift;
    my ( $expr, $offset, $length, $replace ) = @_;

    my $bytestr = decode('utf-8',$expr);
    my $bytesubstr = substr($bytestr,$offset, $length, $replace);
    my $substr = encode('utf-8',$bytesubstr);

    return $substr;
}


##################################################################
# Name: utf8length
#
# Description:
# Works just like perl func length, but also works with utf8 strings 
# "Returns the length in characters of the value of EXPR"
#
# Input: 
# 1.) $expr: string that may contain utf8
# 
# Output:
# $length length of the string
#
# Created: PG 2013-07-12
# Changes: 
# 
######################################################################
sub utf8length {
    my $self = shift;
    my ( $expr ) = @_;

    my $bytestr = decode('utf-8',$expr);
    my $length = length($bytestr);

    return $length;
}



##################################
### here come some File oriented methods may be they should go into
### a DAASIlib::File.pm



1;
