#######################################################################
# Copyright: DAASI International GmbH 2018. All rights reserved.
#
# This is Open Source Software
# License: Artistic License 2.0
#
# Author: Peter Gietz, DAASI International GmbH, www.daasi.de
# For questions please mail to peter.gietz@daasi.de
#######################################################################
#!/usr/bin/perl -w


package DAASIlib::LDAP;

use strict;
use warnings;
use vars qw($VERSION);
our $VERSION = "0.33";

use Log::Log4perl qw(:levels);
my $logger = Log::Log4perl->get_logger('');
use DAASIlib::Data;
my $data = new DAASIlib::Data;
use DAASIlib::Entry;


use Net::LDAP qw(LDAP_SUCCESS LDAP_TIMELIMIT_EXCEEDED LDAP_SIZELIMIT_EXCEEDED
		 LDAP_NO_SUCH_OBJECT LDAP_ALREADY_EXISTS
                 LDAP_NAMING_VIOLATION );
use Net::LDAP::LDIF;
use Net::LDAP::Entry;
use Net::LDAP::Util;
use Net::LDAP::Schema;
use Net::LDAP::Control;
use Net::LDAP::Control::ProxyAuth;

use Net::LDAP::Filter;
use Net::LDAP::FilterMatch;
use Net::LDAP::Extension::SetPassword;


#use Net::LDAP::Constant qw();
use MIME::Base64;

sub new {
    my $class;
    my $self;

    my ( $this , $param ) = @_;
    $class = ref($this) || $this;

    $self={
	data => {},
    };

    bless($self , $class );

    return $self;

}


sub initReplogFile {
    my $self = shift;
    my ($filename, $replogfile, $is_removefile) = @_;

    use File::Flock;
    use File::Temp qw/ tempfile tempdir /;

    my $fh = new File::Temp(TEMPLATE => 'dbconnector.XXXXX',
			    SUFFIX => '.tmp',
			    UNLINK => 0,
			    );
    if ( ! $filename ) {
	$filename = $replogfile;
	if ( ! $filename) {
	    ### todo früher in CONF abfangen
	    $logger->logdie("Replication logfile specification is missing. You have to either specify it with inputfile (-i) or as part of inputuri (-u)");
	}

    }
    open (my $fpin, "<$filename") ||
	$logger->logdie("Couldn't open input file  $filename");
    lock($filename);


    {
	my $oldFN = select( $fpin);
	local ( $/, $\ );
	undef $/; undef $\;
	select( $oldFN );

        my $data = <$fpin>;
	print $fh $data;
    }
    close ($fpin);

#    if ( $filename !~ /.tmp$/ ) {
    if ( $is_removefile ) {
	$logger->debug("erasing $filename");
	if ( !open( $fpin, "> $filename" )) {
	    print STDOUT "couldn't erase replog file <$filename>: $!";
	}
	close ($fpin);
    }
    $logger->debug("unlocking $filename");
    unlock($filename);

    return ($fh)

}

sub getReplogInfo {
    my $self = shift;
    my ($ldapuri) = @_;

    my %reploginfo;
    my $slapdconf = $self->getConfigFromUri($ldapuri);
    my $counter = 0;
    my $fh;
    my $rh_replica;

    if ( ! $slapdconf ) {
	$logger->error("getReplogInfo: no Slapd config file specification in LDAP URI");
	return undef;
    } else {
	open( $fh, "<$slapdconf" ) ||
	    $logger->logdie("Couldn't open Slapd.conf file  $slapdconf");

	while ( my $line = <$fh> ) {
	    if ( $line =~ /^\s*replica\s+(.*)/ ) {
#		print "Found replica statement in configfile\n";
		my $hostport;
		($hostport, $rh_replica) = $self->getReplicaInfo( $1, $fh );
		if ( $hostport ) {
		    if ( ! $rh_replica ) {
			$logger->error("Missing replica info for host
                                        $hostport");
		    }
		    else {
			$logger->info ("Found replicastatement for host
                                        $hostport");
			$reploginfo{replica}{$hostport} = $rh_replica;
		    }
		}
		else {
		    $logger->error("Missing hostport in replica info");
		}

	    }
	    elsif ($line =~ /^\s*replogfile\s+(.*)/) {
		$reploginfo{replogfile} = $1;
	    }
	}
	close $fh;
#    print Dumper(%reploginfo);

    }


    return(\%reploginfo);

}


sub getReplicaInfo {
    my $self = shift;
    my ($string, $fh) = @_;

 ### From slurp config.c:
 #* Parse a "replica" line from the config file.  replica lines should be
 #* in the following format:
 #* replica    host=<hostname:portnumber> binddn=<binddn>
 #*            bindmethod="simple" credentials=<creds>
 #*
 #* where:
 #* <hostname:portnumber> describes the host name and port number where the
 #* replica is running,
 #*
 #* <binddn> is the DN to bind to the replica slapd as,
 #*
 #* bindmethod is "simple", and
 #*
 #* <creds> are the credentials (e.g. password) for binddn.  <creds> are
 #* only used for bindmethod=simple.
 #*
 #* The "replica" config file line may be split across multiple lines.  If
 #* a line begins with whitespace, it is considered a continuation of the
 #* previous line.

    my $line = " ";
    my $cursor = tell($fh);
    my $replicastr = $string;
    my %replica;

    while ( $line = <$fh> ) {
	if ( $line =~ /^\s/ ) {
	    $replicastr .= $line;
	    $cursor = tell($fh);
	} else {
	    seek ($fh, $cursor, 0);
	    last;
	}
    }
#    print "replicastring: [$replicastr]\n";
    $replicastr =~ s/=\"([^ ]+) ([^ ]+)\"/=\"$1%20$2\"/;
#    print "now:           [$replicastr]\n";

    my @replicalines = split /\s+/,$replicastr;
    my $hostport;
    my $schema;

    foreach (@replicalines) {
#	print "analysing line: [$_]\n";
	my ( $attr, $val) = split /=/, $_, 2;
	$val =~ s/^\"//;
	$val =~ s/\"$//;
	$val =~ s/%20/ /;
	if ( $attr eq 'host' ) {
	    $hostport = $val;
	    my ($host, $port ) = split /:/, $val, 2;
	    $replica{host} = $host;
	    $replica{port} = $port;
	}
	elsif ( $attr eq 'uri' ) {
	    ( $schema, $hostport) = split m{://}, $val, 2;
	    my ($host, $port ) = split /:/, $hostport, 2;
	    $replica{schema} = $schema;
	    $replica{host} = $host;
	    $replica{port} = $port;
	}
	else {
#	    print "$attr = $val\n";
	    $replica{$attr} = $val;
	}
    }

#    $line = readline($fh);
#    print "next line is [$line]\n";
#    print "returning hostport: [$hostport]\n";
    return ($hostport, \%replica);
}


sub getReplogFileFromConfigFromUri {
    my $self = shift;
    my ($ldapuri) = @_;
    my $configfilename = $self->getConfigFromUri($ldapuri);

    print "found configfile: $configfilename\n";



#    print "found replogfile: $replogfilename\n";

    return ($configfilename);

}

sub getConfigFromUri {
    my $self = shift;
    my ($ldapuri) = @_;

    if ( $ldapuri) {
#	print "getconfig from uri: $ldapuri\n";
	my $rh_ldapserver = $self->defineServerFromURI($ldapuri, 'notoken');
	return $rh_ldapserver->{configfile};
    }
    else {
#	print "no URI in getConfigFromURI!!!\n";
	return "";
    }

}


sub initLdifFile {
    my $self = shift;

    my ($ldif_file, $mode) = @_;

    my $ldif = Net::LDAP::LDIF->new( $ldif_file, $mode, onerror => 'undef' );

    die "Couldn't init LDIF file for mode $mode" unless $ldif;

    return ($ldif);
}



sub loadValuesFromLdap {
    my $self = shift;
    my ( $ra_list, $uri, $attr, $token, $filter, $control ) = @_;

    $logger->debug("LDAP::loadValuesFromLdap(): loading values $attr from Server $uri");
#    print "loading values $attr from Server $uri\n";
    if ( $filter) {
	$uri = $self->insertFilterInUri($uri, $filter);
	$logger->debug("LDAP::loadValuesFromLdap(): added filter to uri: $uri");
#	print "added filter to uri: $uri\n";
    }

    my $rh_ldapserver =
	$self->defineServerFromURI($uri, $token, 0);

    $rh_ldapserver->{attribs} = [ $attr ];
    $rh_ldapserver->{control} = $control;

    my $ldap = $self->connectServer($rh_ldapserver, 1);
    if ( ! $ldap ) {
	$logger->logdie("LDAP::loadValuesFromLdap(): Could not connect to "
			."Server $$rh_ldapserver{host}:"
			."$$rh_ldapserver{port}!");
    }
    else {
	my $mesg = $self->doSearch($ldap, $rh_ldapserver);
	if ($mesg->entries) {
	    $logger->debug("Found ".$mesg->entries." entries");
	    my $entry_found;
	    foreach $entry_found ($mesg->all_entries) {
		my @values = $entry_found->get_value( $attr );
		if ( $values[0] ) {
#		    $logger->debug ("LDAP::loadValuesFromLdap(): adding values to unique list [@values]");
		    push @{$ra_list}, @values;
		}
	    }
	} else {
	    $logger->warn("LDAP::loadValuesFromLdap(): No entries found");
	}

	$ldap->unbind;
    }

}

sub getValueFromUri {
    my $self = shift;
    my ( $uri, $valuetype) = @_;

    my $value = "";

    if ( $valuetype eq "attributes" ) {
	if ( $uri  =~ /^[^?]*?\?([^?]+?)\?.*$/ ) {
	    $value = $1;
	} else {
	    $logger->warn("URI part $valuetype missing in LDAP URI");
	}
    } else {
	$logger->logdie("LDAP: getValueFromUri called with unknown valuetype");
    }

    return($value);
}

sub defineServerFromReplicaHash {
    my $self = shift;
    my ( $rh_server ) = @_;
    my $rh_ldapdef = $self->defineServer( $rh_server->{schema},
					  $rh_server->{host},
					  $rh_server->{port},
					  undef,
					  $rh_server->{binddn},
					  $rh_server->{credentials},
					  );
	### todo bindmethod and other replica parameters

    return ($rh_ldapdef);
}


# Changes: TZ 2017-06-01 Add parameter lang. When set, attributes of that language
#                        are returned.
sub defineServerFromURI {
    my $self = shift;
    my ( $uri, $pwtoken,
	 $sizelimit, $timelimit, $deref, $is_typesonly,
	 $rootpw, $control, $lang ) = @_;

## URI format:  ldap://<host>[:<port>]/<basedn>?<attributes>?<scope>?<filter>?<extension>

    if ( ! $pwtoken ) {
	$logger->error("LDAP::defineServerFromURI(): Call of defineServerFromURI with no token!");
    }
    else {
	if ( ! $uri ) {
	    $logger->error("LDAP::defineServerFromURI(): Call of defineServerFromURI with no uri! (token: $pwtoken) ");
	}
	else {
	    $logger->debug("LDAP::defineServerFromURI(): Defining LDAP server from [$uri] ($pwtoken)");
	}
    }
    my $schema;

#    if ( ! ( $uri =~ s{^(ldaps?)://(.*)}{$2}) ) {
    if ( ! $uri || ! ( $uri =~ m{^(ldaps?)://(.*)}) ) {
	my $str = $uri ? $uri : "[no uri]";
	$logger->logdie("LDAP::defineServerFromURI(): uri has to start with 'ldap://' I got: $str");
    }
    ($schema, $uri) = split m{://}, $uri;

#print "schema: [$schema], rest of uri: [$uri]\n";

    my ( $hostport, $params ) = split /\//, $uri, 2;
    my ( $host, $port ) = split /:/, $hostport;

    my ($basedn,$attribs,$scope,$filter,$extensionstring);
    my $bindname = '';
    my $credentials = '';
    my $configfile;
    my $searchattr;

    my @attributes = undef;

    if ( $params ) {
	($basedn,$attribs,$scope,$filter,$extensionstring) =
	    split /\?/, $params;

	if ($attribs) {
	    @attributes= split /,/, $attribs;
	}


	if ( $basedn ) {
	    $basedn =~ s/%2c/,/g;
	}
    }


#print "Bindname is: $bindname\n";
#print "ldap://$host:$port/$basedn?$attribs?$scope?$filter?$extension\n";

### defaults:
#    if ( ! $port ) { $port = 389; }
#    if ( ! $scope ) { $scope = "sub"; }
#    if ( ! $filter ) { $filter = "(objectclass=*)"; }
#    if ( ! $sizelimit ) { $sizelimit = 0; }
#    if ( ! $timelimit ) { $timelimit = 389; }
#    if ( ! $deref ) { $deref = 0; }
#    if ( ! $is_typesonly ) { $is_typesonly = 0; }
    if ( $control ) {
	$logger->debug("LDAP::defineServerFromURI(): called with a control");
    }

    my $rh_ldapdef = $self->defineServer( $schema, $host, $port, 3,
					  $bindname, $credentials,
					  $basedn, $scope, \@attributes,
					  $filter, $sizelimit, $timelimit,
					  $deref, $is_typesonly, undef,
					  $configfile, $searchattr, undef,
					  undef, undef, $control, $lang);


    if ( $extensionstring ) {
	$self->defineServerExtensions($rh_ldapdef, $extensionstring,
				      $pwtoken, $rootpw);
    }


    return ($rh_ldapdef);
}


sub defineServerExtensions {
    my $self = shift;
    my ($rh_ldapdef, $extensionstring, $pwtoken, $rootpw) = @_;
#	    print "\nextensionstring: [$extensionstring]\n";

    my @extensions = split /;/, $extensionstring;

    my @allowedextensions = qw/ bindname searchattr userbase userattribute
	rolebase roleattribute rolememberattribute tls userfilter rolefilter
	config rootdn/;

    foreach my $extension ( @extensions ) {
#	print "Checking extension: $extension\n";
	my $is_found = 0;
	foreach my $allowed ( @allowedextensions ) {
#	    print "    Checking allowed: $allowed\n";

	    if ( $extension =~ /^$allowed=(.*)$/ ) {
		my $value = $1;
		$value =~ s/%2c/,/g;
		$rh_ldapdef->{$allowed} = $value;
		$is_found = 1;
#		print "found allowed extesnion: $allowed\n";
	    }
	}

	if ( ! $is_found ) {
	    $logger->logdie("Wrong extension [$extension]. The only allowed extensions in LDAP URI are: @allowedextensions ");
	}

    }


    if ( $rh_ldapdef->{bindname} && $pwtoken ne 'notoken' ) {
	if (! $pwtoken ) {
	    $logger->error("Cannot retrieve Password for bindname from passwordfile");
	}
	else {
	    $rh_ldapdef->{credentials} = $data->getSecret($pwtoken);
	}
    }

    if ( $rh_ldapdef->{rootdn} && $rootpw ) {
	$rh_ldapdef->{rootpw} = $rootpw;
    }

    return ($rh_ldapdef);

}


sub defineServerTLS {
    my $self = shift;
    my ($rh_ldapdef, $is_tls, $tls_cafile, $tls_verify,
	$binddn, $bindpw, $tls_cypher ) = @_;
    $logger->debug("defineServerTLS: starting");


    $rh_ldapdef->{is_tls} = $is_tls ? $is_tls : 0;
    $rh_ldapdef->{tls_cafile} = $tls_cafile ? $tls_cafile : '';
    $rh_ldapdef->{tls_verify} = $tls_verify ? $tls_verify : 'none';
    $rh_ldapdef->{tls_cypher} = $tls_cypher ? $tls_cypher : '';
    $rh_ldapdef->{binddn} = $binddn ? $binddn : '';
    $rh_ldapdef->{credentials} = $bindpw ? $bindpw : '';


    return ($rh_ldapdef);

}


# Changes: TZ 2017-06-01 Add parameter lang. When set, attributes of that language
#                        are returned.
sub defineServer {
    my $self = shift;
    my ( $schema, $host, $port, $version, $binddn, $credentials, $basedn,
	 $scope, $ra_attribs, $filter, $sizelimit, $timelimit, $deref,
	 $is_typesonly, $rf_callback, $configfile, $searchattr, $is_tls,
	 $tls_cafile, $tls_verify, $control, $lang ) = @_;

##    $logger->debug("LDAP::defineServer(): called with sizelimit: $sizelimit");
    my %ldapdef;
    $ldapdef{schema} = $schema ? $schema : 'ldap';
    $ldapdef{host} = $host ? $host : 'localhost';
    $ldapdef{port} = $port ? $port : 389;
    $ldapdef{version} = $version ? $version : 3;
    $ldapdef{binddn} = $binddn;
    $ldapdef{credentials} = $credentials;
    $ldapdef{basedn} = $basedn;
    $ldapdef{scope} = $scope ? $scope : 'sub';
    $ldapdef{attribs} = $$ra_attribs[0] ? $ra_attribs : ['*'];
    $ldapdef{filter} = $filter ? $filter : '(objectclass=*)';
    $ldapdef{sizelimit} = $sizelimit ? $sizelimit : undef;
    $ldapdef{timelimit} = $timelimit ? $timelimit : undef;
    $ldapdef{deref} = $deref ? $deref : 'find';
    $ldapdef{is_typesonly} = $is_typesonly ? $is_typesonly : 0;
    $ldapdef{callback} = $rf_callback ? $rf_callback : '';
    $ldapdef{configfile} = $configfile ? $configfile : '';
    $ldapdef{searchattr} = $searchattr ? $searchattr : '';
    $ldapdef{is_tls} = $is_tls ? $is_tls : 0;
    $ldapdef{tls_cafile} = $tls_cafile ? $tls_cafile : '';
    $ldapdef{tls_verify} = $tls_verify ? $tls_verify : 'none';
    $ldapdef{control} = $control ? $control : '';
    $ldapdef{lang} = $lang ? $lang : '';

    return ( \%ldapdef);
}

sub connectServer {
    my $self = shift;
    my ( $rh_ldapdef, $dontforcebind, $isrootbind ) = @_;

    my $uri = "$$rh_ldapdef{schema}://$$rh_ldapdef{host}:$$rh_ldapdef{port}";

    $logger->debug("LDAP::connectServer(): Connecting to $uri");

    my $ldap = Net::LDAP->new( $uri,
			       version => $$rh_ldapdef{version},
			       timeout => $$rh_ldapdef{timelimit} );

    if ( !$ldap ) {
	$logger->error("LDAP::connectServer(): ERROR while creating ".
		       "connection to ldap server ".
		       "$$rh_ldapdef{host}:$$rh_ldapdef{port}");
	return(undef);
    }

    if ( $rh_ldapdef->{is_tls} ) {
	$logger->debug("LDAP::connectServer(): starting TLS ");
	$logger->debug("LDAP::connectServer(): verify: $rh_ldapdef->{tls_verify}")
	    if  $rh_ldapdef->{tls_verify};
	$logger->debug("LDAP::connectServer(): cafile: $rh_ldapdef->{tls_cafile}")
	    if $rh_ldapdef->{tls_cafile};
	$logger->debug("LDAP::connectServer(): cypher: $rh_ldapdef->{tls_cypher}")
	    if $rh_ldapdef->{tls_cypher};

	my $tlsmesg = $ldap->start_tls ( verify => $rh_ldapdef->{tls_verify},
					 cafile => $rh_ldapdef->{tls_cafile},
					 ciphers => $rh_ldapdef->{tls_cypher});
	if ( $tlsmesg->code != Net::LDAP::LDAP_SUCCESS ) {
            $logger->error ("LDAP::connectServer(): ERROR in ldap->start_tls: " . $tlsmesg->error );
        }
	else {
	    $logger->debug ("LDAP::connectServer(): Start_TLS operation succeeded");
	    $logger->debug ("LDAP::connectServer(): tls cipher: " .
			     $ldap->cipher . "\n" );
	    $logger->debug ("LDAP::connectServer(): tls certificate: " .
			     $ldap->certificate . "\n" );

	}
    }

    my ($binddn, $credentials, $isbind);


    if ( $isrootbind ) {
	$binddn = $$rh_ldapdef{rootdn};
	$credentials = $$rh_ldapdef{rootpw};
	$isbind=1;
    }
    elsif ( $$rh_ldapdef{binddn} || $$rh_ldapdef{bindname} ) {

	$binddn = $$rh_ldapdef{binddn} ? $$rh_ldapdef{binddn} : $$rh_ldapdef{bindname};
	$credentials = $$rh_ldapdef{credentials};
	$isbind=1;
    }
    else {
#	use Data::Dumper;
#	$logger->debug("no binddn found in rh_ldap: ".Dumper($rh_ldapdef));
    }

    if ( $isbind ) {
	$logger->debug("LDAP::connectServer(): Binding as $binddn");

	my @args = (
	    password => $credentials,
	    version => $$rh_ldapdef{version},
	);
	### we do NOT need proxyauth control in bind!!

        my $mesg = $ldap->bind( $binddn, @args );

        unless ( $mesg->code == Net::LDAP::LDAP_SUCCESS ) {
	    $logger->error("LDAP::connectServer(): Error in LDAP-Bind: ".$mesg->error);
	    if (! $dontforcebind) {
		$logger->error("LDAP::connectServer(): Returning undef instead of ldap connection");
		$ldap = undef;
	    }
        }
    }

    return $ldap;
} # connectServer

# Changes: TZ 2017-05-31 Add parameter lang. When set, attributes of that language
#                        are returned.
sub doSearch {
    my $self = shift;
    my ( $ldap, $rh_ldapdef ) = @_;

#    use Data::Dumper;
#    print Dumper( $rh_ldapdef );

    my $basedn = "$$rh_ldapdef{basedn}";
    $basedn =~ s/([^\\])\+/$1\\+/g;

    $logger->debug("LDAP::doSearch(): Searching from base: $basedn")
	if $basedn;
    $logger->debug("LDAP::doSearch(): scope $$rh_ldapdef{scope}")
	if $$rh_ldapdef{scope};
    $logger->debug("LDAP::doSearch(): filter: $$rh_ldapdef{filter} ")
	if $$rh_ldapdef{filter};

    if ( $$rh_ldapdef{attribs}) {
	$logger->debug("LDAP::doSearch(): retrieve attributes: @{$$rh_ldapdef{attribs}} ");
    } else {
	$logger->debug("LDAP::doSearch(): retrieve all attributes");
    }
    if ( ! $$rh_ldapdef{filter}) {
	$$rh_ldapdef{filter}="(objectclass=*)";
    }

    my $mesg;
    if ( ! $ldap ) {
	$logger->error("LDAP::doSearch(): called without ldap connection");
    }
    else {
	my @args = (
	    base      => $basedn,
	    scope     => $$rh_ldapdef{scope},
	    filter    => $$rh_ldapdef{filter},
	    attrs     => $$rh_ldapdef{attribs},
	    deref     => $$rh_ldapdef{deref},
	    sizelimit => $$rh_ldapdef{sizelimit},
	    timelimit => $$rh_ldapdef{timelimit},
	    typesonly => $$rh_ldapdef{is_typesonly},
	    callback  => $$rh_ldapdef{callback},
	    );
	if ( $$rh_ldapdef{control} ) {
	    $logger->debug("LDAP::doSearch(): adding control $$rh_ldapdef{control} to args");
	    push @args, control => [ $$rh_ldapdef{control} ];
	}

	$mesg = $self->ldap_search( $ldap, $$rh_ldapdef{lang}, @args );

	if ( $mesg->code == Net::LDAP::LDAP_SUCCESS ) {
	    $logger->debug("search performed successfully found ".
			  $mesg->count()." entries"  );
	} else {
	    $self->logLDAPError($mesg, "search");
	}
    }

    return ($mesg);
}

##################################################################
# Name: ldap_search
#
# Description: Calls the perl LDAP search method and replaces
#              the standard attributes by their lang-specific
#              attributes if $lang is defined.
#              This method MUST be used insted of calling
#              the perl LDAP search method, otherwise lang-specific
#              attributes are lost!!!
#
# Input: $ldap: Net::LDAP::Message object
#        $lang: the language identifier such as "en" or "de_DE"
#        @args: LDAP search method arguments
#
# Output: $mesg: Net::LDAP::Message object
#
#
# Created: TZ 2017-06-01
# Changes:
#
#
######################################################################
sub ldap_search {
    my $self = shift;
    my ( $ldap, $lang, @args ) = @_;

    my $mesg = $ldap->search( @args );

    if ( $mesg->code == Net::LDAP::LDAP_SUCCESS ) {
        $mesg = $self->keep_only_lang_attributes($mesg, $lang) if $lang;
    }
    return $mesg;
}

##################################################################
# Name: keep_only_lang_attributes
#
# Description: If lang is not present $mesg is returned as is.
#              Otherwise enries are cloned as DAASIlib::Entry
#              objects that return language specific attribute
#              values matching lang - if they exist - in their
#              get_value method. If for an attribute
#              no lang specific value is found, the standard attribute
#              values without attribute options are returned.
#
# Input: $mesg: Net::LDAP::Message object
#        $lang: the language identifier such as "en" or "de_DE"
#
# Output: $mesg
#
#
# Created: TZ 2017-06-01
# Changes:
#
#
######################################################################
sub keep_only_lang_attributes {
    my $self = shift;
    my ( $mesg, $lang ) = @_;

    my @entries = $mesg->entries();
    my @newentries = ();

    if ($lang) {
        $logger->debug("keep_only_lang_attributes(): Lang is: $lang");

#        use Data::Dumper;
#        $logger->debug("keep_only_lang_attributes(): search result before: ".Dumper($mesg));

        foreach my $entry ( @entries ) {
            my $newentry = DAASIlib::Entry->new($lang);
            $newentry->copy($entry);
            push @newentries, $newentry;
        }
        $mesg->{entries} = \@newentries;

#        $logger->debug("keep_only_lang_attributes(): search result after: ".Dumper($mesg));
    }
    return $mesg;
}


sub logLDAPError {
    my $self = shift;
    my ( $mesg, $operation, $onscreen ) = @_;


    if ( ! $operation) {
	$operation = "operation";
    }

    if ( $mesg->code == LDAP_TIMELIMIT_EXCEEDED ||
	 $mesg->code ==  LDAP_SIZELIMIT_EXCEEDED ||
	 $mesg->code ==  LDAP_NO_SUCH_OBJECT ||
	 $mesg->error eq "No attributes to update" ) {

	$logger->warn("LDAP $operation errorcode ".$mesg->code.": ".$mesg->error);
    }
    elsif ( $mesg->code == LDAP_SUCCESS ) {
	$logger->debug("Result in LDAP $operation: ".$mesg->error
		       ." (".$mesg->code.")");
    }
    else {
	$logger->error("Error ".$mesg->code." in LDAP $operation: ".$mesg->error);
    }
    if ( $onscreen) {
	print "   Result in LDAP $operation: ".$mesg->error." (".$mesg->code.")\n";
    }
}

sub updateServerFromLDIF{
    my $self = shift;
    my ( $filename, $rh_ldapdef, $rewrite ) = @_;

#    use Data::Dumper;
#    print Dumper($rh_ldapdef);


    my $ldif = $self->initLdifFile($filename,"r");
    my $ldiferrorfile = "$filename.errors.txt";
    my $ldiferror = $self->initLdifFile($ldiferrorfile,"w");
    my $ldiferrorhandle = $ldiferror->handle();
    my $ldap = $self->connectServer($rh_ldapdef);

    my $mesg;
    my $entry;
    my @attribs =();
    while( not $ldif->eof ( ) ) {
	$entry = $ldif->read_entry ( );
	my $dn = $entry->dn();
	if ( $ldif->error ( ) ) {
	    $logger->error("In LDIF file: ", $ldif->error ( ) );
	    $logger->error("Lines:\n", $ldif->error_lines ( ) );
	} else {
	    $logger->debug("Updating entry ".$entry->dn());
	    $mesg = $entry->update($ldap);
	    unless ( $mesg->code == Net::LDAP::LDAP_SUCCESS ) {
		if ( $rewrite && $mesg->code == LDAP_ALREADY_EXISTS ) {
		    $logger->warn("entry ($dn)already exists, will rewrite ($rewrite)");
		    if ( $rewrite =~ /^ok_except_(.*)$/ ) {
			@attribs = split /_/, $1;
		    }
		    my $oldentry = $self->sessionReadEntry($entry->dn(), $ldap,
							   $rh_ldapdef);
		    if (! $oldentry ) {
			$logger->logdie("Expected entry ($dn) does not exist!");
		    }

		    $mesg = $self->updateOldEntry($oldentry, $entry, $ldap,
						  \@attribs);
		    if ( $mesg && $mesg->code != Net::LDAP::LDAP_SUCCESS ) {
			$logger->error("Error in LDAP update (2nd. try) in entry $dn: ".$mesg->error);
		    }
		}
		else {
		    $logger->error("Error in LDAP update of entry $dn: ".$mesg->error);
		    $logger->info("Writing not updated entry to $ldiferrorfile");
		    print $ldiferrorhandle
			"\n# Error in LDAP update: ".$mesg->error;
		    $ldiferror->write_entry($entry);
		}
	    }
	}
    }
    $ldap->unbind if $ldap;
    $ldif->done ( );
    $ldiferror->done ( );
    return ($mesg);
}

sub createDnFromObject {
    my $self = shift;
    my ( $rh_object, $be_stupid ) = @_;
    my $dn = "";

    my $rdnattr = $rh_object->{classinfo}{rdnattr};
    my $rdnattr_lc = lc($rdnattr);

    my $basedn = $rh_object->{classinfo}{basedn};

#    print "basedn: $basedn\n";
    while ( $basedn =~ /(.*?)#\((.+?)\)(.*)/ ) {
	my $before = $1;
	my $attr = $2;
	$logger->debug("adding val from attribute $attr");
	my $val = $rh_object->{attributes}{$attr}[0];
	if ( $val) {
	    $logger->debug("found val $val from attribute $attr");
	}
	else {
	    $logger->error("Couln't find any value in Attribute $attr");
	}
	my $after = $3;
	$basedn = "$before$val$after";
    }


    if ( $rh_object->{attributes}{$rdnattr}[0] ) {
	$dn = "$rdnattr=".
	    "$rh_object->{attributes}{$rdnattr}[0],".
	    "$basedn";
    }
    elsif (  $rh_object->{attributes}{$rdnattr_lc}[0] ) {
	$dn = "$rdnattr=".
	    "$rh_object->{attributes}{$rdnattr_lc}[0],".
	    "$basedn";
    }
    else {
	if ($be_stupid) {
	    $dn = "$rdnattr=XXX,$basedn";
	}
	else {
	    $logger->error("Couldn't create DN: RDN attribute $rdnattr has no value");

	    use Data::Dumper;
	    $logger->debug(Dumper($rh_object->{attributes}));
	}
    }

    $dn =~ s/([^\\])\+/$1\\+/g;
    
    $logger->debug("Returning dn '$dn");
    return ($dn);
}


## Not really used in dbconnector project:
sub createDiffEntry {
    my $self = shift;
    my ( $oldentry, $newentry ) = @_;

    if ( ! $newentry ) {
	print "createDiffEntry called with no newentry\n";
	return;
    }
    if ( ! $oldentry ) {
	print "createDiffEntry called with no newentry\n";
	return;
    }

    my $diffentry = $oldentry->clone();


#    my $is_change;
#    my $is_newvalue;

    foreach my $attr ( $newentry->attributes ) {
	my @values = $newentry->get_value( $attr );
	if ( $values[0] ) {
	    my @oldvalues = $oldentry->get_value( $attr );
	    my @newvalues = $data->findNewValues(\@values, \@oldvalues );
#	    print "testing whether newvalues [@values] of attribute $attr do not exist in old values [@oldvalues]\n";
	    if ( @newvalues ) {
		$diffentry->add($attr => @newvalues );
	    }
	}
    }

    return ($diffentry);
}

# An alternative from ldifdiff.pl: ($Id: ldifdiff.pl,v 3.7 2005/03/15 14:22:45 subbarao Exp $)
# Generate LDIF to update $target with information in $source.
# Optionally restrict the set of attributes to consider.
sub updateFromEntry
{
    my $self = shift;
    my ($source, $target, @attrs) = @_;
    my ($attr, $val, $ldifstr);

    my @sourceattrs;
    my (%ciscmp, %dnattrs, %sharedattrs);
    my $keyattr;
#    GetOptions('a|sourceattrs=s' => sub { @sourceattrs = split(/,/, $_[1]) },
#	       'c|ciscmp=s' => sub { my @a = split(/,/,lc $_[1]); @ciscmp{@a} = (1) x @a },
#	       'dnattrs=s' => sub { my @a = split(/,/,lc $_[1]); @dnattrs{@a} = (1) x @a },
#	       'k|keyattr=s' => \$keyattr,
#	       'sharedattrs=s' => sub {my @a=split(/,/,lc $_[1]);@sharedattrs{@a}=(1) x @a}
#	       );
    %ciscmp = (objectclass => 1, manager => 1, member => 1, owner => 1,
	       uniquemember => 1)
	unless keys %ciscmp;
    %dnattrs = (manager => 1, member => 1, owner => 1, uniquemember => 1)
	unless keys %dnattrs;
    %sharedattrs = (objectclass => 1)
	unless keys %sharedattrs;



    unless (@attrs) {
	# add all source entry attributes
	@attrs = $source->attributes;
	# add any other attributes we haven't seen from the target entry
	foreach my $tattr ($target->attributes) {
	    push(@attrs, $tattr) unless grep(/^$tattr$/i, @attrs);
	}
    }

    $target->{changetype} = 'modify';

    foreach $attr (@attrs) {
	my $lcattr = lc $attr;
	next if $lcattr eq 'dn'; # Can't handle modrdn here

	# Build lists of unique values in the source and target, to
	# speed up comparisons.
	my @sourcevals = $source->get_value($attr);
	my @targetvals = $target->get_value($attr);
	my (%sourceuniqvals, %targetuniqvals);
	foreach (@sourcevals) {
	    my ($origval, $val) = ($_, $_);
	    $val = lc $val if $ciscmp{$lcattr};
	    # Get rid of spaces after non-escaped commas in DN attrs
	    $val =~ s/(?<!\\),\s+/,/g if $dnattrs{$lcattr};
	    $sourceuniqvals{$val} = $origval;
	}
	foreach (@targetvals) {
	    my ($origval, $val) = ($_, $_);
	    $val = lc $val if $ciscmp{$lcattr};
	    # Get rid of spaces after non-escaped commas in DN attrs
	    $val =~ s/(?<!\\),\s+/,/g if $dnattrs{$lcattr};
	    $targetuniqvals{$val} = $origval;
	}
	foreach my $val (keys %sourceuniqvals) {
	    if (exists $targetuniqvals{$val}) {
		delete $sourceuniqvals{$val};
		delete $targetuniqvals{$val};
	    }
	}

	# Move on if there are no differences
	next unless keys(%sourceuniqvals) || keys(%targetuniqvals);

	# Make changes as appropriate
	if ($sharedattrs{$lcattr}) {
	    # For 'shared' attributes (e.g. objectclass) where $source may not
	    # be a sole authoritative source, we issue separate delete and
	    # add modifications instead of a single replace.
	    $target->delete($attr => [ values(%targetuniqvals) ])
		if keys(%targetuniqvals);
	    $target->add($attr => [ values(%sourceuniqvals) ])
		if keys(%sourceuniqvals);
	}
	else {
	    # Issue a replace or delete as needed
	    if (@sourcevals) { $target->replace($attr => [ @sourcevals ]) }
	    else { $target->delete($attr) }
	}
    }


#    my $is_changed = 1;
    # Get rid of the "changetype: modify" if there were no changes
    if ( !  @{$target->{changes}} ) {
	delete($target->{changetype});   # unless @{$target->{changes}};
#	$is_changed = 0;
    }

#    return ( $is_changed );
}



sub createEntryFromObject {
    my $self = shift;
    my ( $rh_object, $excludepattern, $be_stupid ) = @_;

    my $entry = undef;

    my $dn = $self->createDnFromObject($rh_object, $be_stupid);

    if ( $dn ) {
	my @attributenames;
	@attributenames = $data->makeUnique(@{$rh_object->{attributenames}});

	if ($excludepattern) {
	    @attributenames = $data->removeElements(\@attributenames, $excludepattern);
	}

	$entry = Net::LDAP::Entry->new;
	$entry->dn($dn);
	$entry->add( "objectClass" => $rh_object->{classes} );

	foreach my $field ( @attributenames ) {
	    if ( @{$rh_object->{attributes}{$field}}[0] ) {
		$entry->add ( $field => $rh_object->{attributes}{$field});
	    }
	}

    }
    else {
	$logger->error("Couldn't create DN in createEntryFromObject");
	print "Couldn't create DN in createEntryFromObject\n";
    }

    return ($entry);
}

sub printEntry {
    my $self = shift;
    my ( $entry, $fp, $is_nodn) = @_;

    if ( ! $is_nodn ) {
	print $fp "dn: ", $entry->dn(), "\n";
    }
    foreach my $attr ( $entry->attributes ) {
	foreach my $value ( $entry->get_value( $attr ) ) {
	    if ( $value =~ /[^\x20-\x7f]/ ) {
# use MIME::Base64;
#
# $encoded = encode_base64('Aladdin:open sesame');
# $decoded = decode_base64($encoded);
		my $encvalue = encode_base64($value);

		if ( $encvalue  ) {
		    print $fp "#encoded: $attr: $value\n";
		    print $fp "${attr}:: $encvalue";
		}
		else {
		    print "something is strange with value $value\n"
		}

	    }
	    else {
		print $fp "${attr}: $value\n";
	    }
	}
    }
    print $fp "\n";
}

sub stringifyEntry {
    my $self = shift;
    my ( $entry, $prefix) = @_;

    my $entrystring;

    my $dn = $entry->dn();

    $entrystring = $self->stringify("dn: $dn\n", $prefix);

    foreach my $attr ( $entry->attributes ) {
	foreach my $value ( $entry->get_value( $attr ) ) {
	    if ( $value =~ /[^\x20-\x7f]/ ) {
# use MIME::Base64;
#
# $encoded = encode_base64('Aladdin:open sesame');
# $decoded = decode_base64($encoded);
		my $encvalue = encode_base64($value);

		$entrystring .=
		    $self->stringify("#encoded: $attr: $value\n",
				     $prefix);
		$entrystring .=
		    $self->stringify("${attr}:: $encvalue",$prefix);
	    }
	    else {
		$entrystring .=
		    $self->stringify("${attr}: $value\n",$prefix);
	    }
	}
    }
    $entrystring .= $self->stringify("\n",$prefix);

    return ($entrystring);
}

sub stringify {
    my $self = shift;
    my ( $string, $prefix) = @_;

    if (  $prefix ) {
	return("${prefix}$string");
    }
    else {
	return ($string);
    }

}


#### BEWARE: not control and thus no AuthzTo
sub readEntryFromUri {
    my $self = shift;
    my ( $dn, $uri, $uripwtoken ) = @_;

    my $rh_ldapserver =  $self->defineServerFromURI($uri, $uripwtoken, 0);

    $rh_ldapserver->{scope} = "base";
    $rh_ldapserver->{basedn} = $dn;
#    if ( $rh_ldapserver->{binddn} ) {
#	$rh_ldapserver->{credentials} = $data->getSecret($uripwtoken);
#    }

    my $ldap = $self->connectServer($rh_ldapserver);
#    print "base search with bas = $dn\n";

    my $entry = sessionReadEntry($dn,$ldap, $rh_ldapserver);

#    my $mesg = $self->doSearch($ldap, $rh_ldapserver);
#    my $entry = undef;
#    if ($mesg->code == Net::LDAP::LDAP_SUCCESS && $mesg->entries == 1) {
#	$entry = $mesg->pop_entry();
#    } else {
#	$logger->warn("Entry $dn not found");

#	print "Entry $dn not found\n";
#    }

    return $entry;
}


#	if ($rh_output_object->{operationalattributes}{is_modifyentry}) {

#	    print "OK now compare what we have got with the corresponding LDAP enntry\n";
#	    use Data::Dumper;
#	    print Dumper(%{$$rh_output_object});
#	    my $rdnattr = $rh_output_object->{classinfo}{rdnattr};
#
#	    if (  $is_ldap_output &&
#		  $rh_output_object->{attributes}{$rdnattr}[0] ) {
#
#		my $dn = "$rdnattr=".
#		    "$rh_output_object->{attributes}{$rdnattr}[0],".
#		    "$rh_output_object->{classinfo}->{basedn}";
#		print "searching DN: $dn at:\n";
#		print "          $conf->{data}{outputuri}\n";



sub findModificationsFromLDAP {
    my $self = shift;
    my ($newentry, $ldap, $rh_ldapserver,$ra_attributes ) = @_;

    my $entry;
    my $is_changes = 0;

#    my $oldentry = $self->sessionReadEntry($dn,$ldap, $rh_ldapserver);
    my $oldentry = $self->sessionSearchEntry($ldap, $rh_ldapserver);

#	use Data::Dumper;
#	print Dumper($newentry);


    if ( $oldentry ) {

	$logger->debug("Found old entry to modify\n");
#	print "findModificationsFromLDAP calling findModifications\n";
	($is_changes, $entry) = $self->findModifications("MOD", $oldentry, $newentry,
					  $ra_attributes);
#	print "back in findModificationsFromLDAP with entry",$entry->dn(),"\n";
#	print Dumper ($oldentry);
	$entry->changetype("modify");
    } else {
	$entry = $newentry;
	$is_changes = 1;
	$entry->changetype("add");
    }

#	print Dumper($entry);
#    exit();

    return($is_changes, $entry);
}

sub findModifications {
    my $self = shift;
    my ($mode, $oldentry, $newentry, $ra_attributes) = @_;

    my $diffentry= $oldentry;
    my $is_change;
    my $is_newvalue;

#	use Data::Dumper;
#	print Dumper($diffentry);

#    $logger->debug("findModifications called with attributes: ".
#		   "@{$ra_attributes}");
#    print "findModifications called with attributes: @{$ra_attributes}";
#    print Dumper($newentry);
#    exit();
    foreach my $attr ( $newentry->attributes ) {
	$is_change = 0;
	$is_newvalue = 0;
	my @values = $newentry->get_value( $attr );
	if ( $ra_attributes ) {
	    if ( grep { $attr eq $_} @{$ra_attributes} ) {
		if ( $values[0] ) {
		    $is_newvalue = 1;
		}
	    }
	} else {
	    if ( $values[0] ) {
		$is_newvalue = 1;
	    }
	}



	if ($is_newvalue ) {
	    my @oldvalues = $oldentry->get_value( $attr );

	    my @newvalues = $data->findNewValues(\@values, \@oldvalues );
#	    print "testing whether newvalues [@values] of attribute $attr do not exist in old values [@oldvalues]\n";
	    if ( @newvalues ) {
		$is_change = 1;
#		print "   the values really seem to be new!\n";
		$logger->debug("findmodifications found something to modify ($mode) in attr: $attr=@newvalues");

#	    print "Adding new values to old entry (mode $mode):\n";
#	    foreach (@values) {
#		print "   [$_]\n";
#	    }

		if ($mode eq "MOD") {
		    $diffentry->replace ( $attr => @newvalues );
		} elsif ($mode eq "ADD") {
		    $diffentry->add ( $attr => @newvalues );
		} elsif ($mode eq "DEL") {
#		    $diffentry->delete ( $attr => @newvalues );
#		    $diffentry->delete ( $attr );
		    # deletes will be handled later
		}
	    }
	}
#	print "Juhu\n";
    }
#    $diffentry->changetype("modify");
#    print Dumper($diffentry);
#    exit();
    return ($is_change, $diffentry);
}



sub updateEntry {
    my $self = shift;
    my ($entry, $ldap) = @_;

    my $mesg = undef;

#    print "updateEntry calling entry-update on entry ", $entry->dn(),"\n";
#    use Data::Dumper;
#    print Dumper ($entry);
#    print Dumper ($ldap);

    $mesg = $entry->update($ldap);
     # This is a temporary patch, until we know if there is something
     # wrong in module Net::LDAP::Entry (patched line 253 of version 0.20)
#     {
#       my %changes ;
#       foreach my $attr ($entry->attributes( nooptions => 1 )) {
#         my [at]values = $entry->get_value($attr) ;
#         $changes{$attr} = [at]values == 1? $values[0]: \@values ;
#       }
#       $msg = $ldap->modify($entry,replace => \%changes) ;
#     }

#    print "Hi there\n";
    if ( $mesg->code != LDAP_SUCCESS ) {
#	print "Error in Update: ".$mesg->code.":".$mesg->error."\n";
	$self->logLDAPError($mesg, "updateEntry");
    }
    return ( $mesg->code );

}

sub updateOldEntry {
    my $self = shift;
    my ($oldentry, $newentry, $ldap, $ra_attribs) = @_;

    my %nonoattribs;

    foreach my $e ( @{$ra_attribs}) {
	$logger->debug("attribute $e will be ignored");
	$nonoattribs{lc($e)}++;
    }

    my $is_change = 0;
    my $mesg;
    $oldentry->changetype('modify');
    foreach my $attr ( $newentry->attributes ) {
	if ( ! $nonoattribs{lc($attr)} ) {
	    my @vals = $newentry->get_value($attr);
	    my @oldvals = $oldentry->get_value($attr);
	    my $is_casesensitive =  lc($attr) eq 'objectclass' ? 0 : 1;

	    if ( $data->areTheyDifferent(\@vals,\@oldvals, $is_casesensitive) ) {
		$oldentry->replace( $attr => \@vals );
		$is_change++;
		$logger->debug("change in attr $attr: @vals (oldvals: @oldvals");
	    }
	}
    }

    if ( $is_change ) {
	$logger->debug("Found changes in $is_change values thus updating");
	$mesg = $oldentry->update($ldap);
    }
    else {
	$logger->warn("no changes found");
#	$mes->g = Net::LDAP::LDAP_SUCCESS;
	$mesg = undef;
    }

    return ($mesg);
}

sub authenticateUser {
    my $self = shift;
    my ( $userid, $passw, $rh_ldapserver, $uidattr, $noneed2bind ) = @_;

    my $result = undef;
    my $entry = undef;
#    my $errortext;

    if ( !$uidattr) { $uidattr = 'uid';}

    $rh_ldapserver->{filter} = "($uidattr=$userid)";
    $logger->debug("LDAP::authenticateUser(): starting connectServer");
    my $ldap= $self->connectServer($rh_ldapserver, 0);
    if ( ! $ldap ) {
	$logger->error("LDAP::authenticateUser(): failed");
	return undef;
    }

    $logger->debug("LDAP::authenticateUser(): starting doSearch");
    my $mesg = $self->doSearch($ldap, $rh_ldapserver);


    if ($mesg->code == Net::LDAP::LDAP_SUCCESS && $mesg->count == 1) {
	$logger->debug("LDAP::authenticateUser(): found exactly one entry");
	$entry = $mesg->pop_entry();
	if ( $noneed2bind ) {
	    $result =  $entry->dn;
	}
	else {
	    $logger->debug("LDAP::authenticateUser(): starting bind with ".
			  $entry->dn);
	    $mesg = $ldap->bind( $entry->dn,
				 password => $passw,
				 version => $rh_ldapserver->{version},
		);

	    if ($mesg->code == Net::LDAP::LDAP_SUCCESS) {
		$result =  $entry->dn;
	    }
	    else {
		$logger->warn("Bind with ".$entry->dn()." not successfull: "
			      .$mesg->code);
	    }
	}
    } else {

	$logger->warn("result of doSearch: ".$mesg->code.'/'.$mesg->error
		  .'/ count: '.$mesg->count);

#	$logger->warn("Entry ".$entry->dn." not found");
    }
    $ldap->unbind;

    return ($result);
}

sub sessionReadEntry {
    my $self = shift;
    my ( $dn, $ldap, $rh_ldapserver ) = @_;

    $rh_ldapserver->{scope} = "base";
    $rh_ldapserver->{basedn} = $dn;

    my $mesg = $self->doSearch($ldap, $rh_ldapserver);
    my $entry = undef;
    if ($mesg->code == Net::LDAP::LDAP_SUCCESS && $mesg->entries == 1) {
	$entry = $mesg->pop_entry();
    } else {
	$logger->warn("Entry $dn not found");
    }
    return $entry;
}

sub sessionSearchSingleEntry {
    my $self = shift;
    my ( $ldap, $rh_ldapserver ) = @_;

    my $filter = $rh_ldapserver->{filter};
#    print "xxx\n";
    my $mesg = $self->doSearch($ldap, $rh_ldapserver);
    my $entry = undef;
    if ($mesg->code == Net::LDAP::LDAP_SUCCESS && $mesg->entries == 1) {
#    print "xxxa\n";
	$logger->debug("sessionSearchEntry found one entry");
	$entry = $mesg->pop_entry();
    } else {
#    print "xxxb\n";

	$logger->warn("Single entry for filter $filter not found, but "
		      . $mesg->entries . "entries");
#print "Single entry for filter $filter not found\n";
    }

#    print "zzz\n";
#    exit();
    return $entry;
}


## bad name will be deleted soon:
sub sessionSearchEntry {
    my $self = shift;
    my ( $ldap, $rh_ldapserver ) = @_;

    my $filter = $rh_ldapserver->{filter};

    my $mesg = $self->doSearch($ldap, $rh_ldapserver);
    my $entry = undef;
    if ($mesg->code == Net::LDAP::LDAP_SUCCESS && $mesg->entries == 1) {
	$entry = $mesg->pop_entry();
    } else {
	$logger->warn("Single entry for filter $filter not found");
#print "Single entry for filter $filter not found\n";
    }
    return $entry;
}


sub createLDIFFileFromLDAP{
    my $self = shift;
    my ( $filename, $rh_ldapdef ) = @_;

    my $returnstr = "";
#    my $libldap = new DAASIlib::LDAP;
    my $ldap = $self->connectServer($rh_ldapdef);

    if ( ! $ldap ) {
	$logger->logdie("Could not connect to Server $$rh_ldapdef{host}:".
			"$$rh_ldapdef{port}!");
    }

    $logger->debug("connected to ".$self->createLdapUri($rh_ldapdef) );


#    $logger->debug("connected to ldap://$$rh_ldapdef{host}:$$rh_ldapdef{port}".
#		  "/$$rh_ldapdef{basedn}?".
#		  $$rh_ldapdef{attribs} ?
#		            join(',', $$rh_ldapdef{attribs}).'?' : '?'.
#		  $$rh_ldapdef{scope}   ? "$$rh_ldapdef{scope}?" : '?'.
#		  $$rh_ldapdef{filter}  ? "$$rh_ldapdef{filter}?" : '?'.
#		  $$rh_ldapdef{binddn}  ? "binddn=$$rh_ldapdef{binddn}" : '');
#    print "   * connected to $$rh_ldapdef{host}\n";

    my $mesg = $self->doSearch($ldap, $rh_ldapdef);

    my $ldif = $self->initLdifFile($filename,"w");

    if ($mesg->entries) {
	$logger->debug("Found ".$mesg->entries." entries");
	my $entry_found;
	foreach $entry_found ($mesg->all_entries) {
#	    print "Found entry ".$entry_found->dn."\n";
	    $ldif->write_entry( $entry_found );
	}
	$returnstr=$filename;
    } else {
	$logger->warn("No entries found");
    }

    $ldif->done();
    $ldap->unbind;
#    $logger->logdie("Exiting now cause code is missing yet");

    return($returnstr);
}


sub createLdapUri {
    my $self = shift;
    my ( $rh_ldapdef ) = @_;

    my $uri = "ldap://";

    $uri .= "$$rh_ldapdef{host}:$$rh_ldapdef{port}";
    $uri .= "/$$rh_ldapdef{basedn}?";
    $uri .=  @{$rh_ldapdef->{attribs}} && @{$rh_ldapdef->{attribs}}[0] ne '*' ?
	join(',', @{$rh_ldapdef->{attribs}}).'?' : '?';
    $uri .= $$rh_ldapdef{scope} &&  $$rh_ldapdef{scope} ne 'sub'  ?
	"$$rh_ldapdef{scope}?" : '?';
    $uri .= $$rh_ldapdef{filter} && $$rh_ldapdef{filter} ne '(objectclass=*)' ?
	"$$rh_ldapdef{filter}?" : '?';
    $uri .= $$rh_ldapdef{binddn}  ? "binddn=$$rh_ldapdef{binddn}" : '';

    return $uri;

}

sub readSchemafromServer {
    my $self = shift;
    my ( $prot, $host, $port ) = @_;
    my ($schema, $ldap);


    my $rh_ldapdef = $self->defineServer($prot, $host, $port);

    $ldap = $self->connectServer($rh_ldapdef);

    if ( ! $ldap ) {
	die "Could not connect to Server $host:$port!\n";
    }
#print "   * connected to $$rh_ldapdef{host}\n";


#    $ldap = Net::LDAP->new ( $host, port => $port );
    $schema = $ldap->schema ( );
#    print "loaded schema\n";
    if ( $schema ) {
	if ($schema->error() ) {
### TODO log this
	    print "Error after loading schema: ";
	    print "$schema->error()\n";
	} else {
#	    print "storing schema\n";
	    $self->{schema} = $schema;
	}
    } else {
### TODO log this
	print "Error couldn't load schema!\n";
    }
    my $mesg = $ldap->unbind;
    print "   * disconnecting from $$rh_ldapdef{host}\n";

#    print "returning schema\n";
    return ($schema);
}


sub writeEntry {
    my $self = shift;
    my ( $fh, $entry, $changetype ) = @_;

    print $fh ($entry->dn(), "\n");
    if ( $changetype ) {
	print $fh "changetype: $changetype\n";
    }

    foreach my $objectclass ( $entry->get_value('objectClass') ) {
	print $fh "objectClass: $objectclass\n";
    }

    foreach my $attr ( $entry->attributes() ) {
	if ( lc($attr) ne 'objectclass' ) {
	    foreach my $val ( $entry->get_value($attr) ) {
		print $fh "$attr: $val\n";
	    }

	}
    }

    print $fh "\n";


}


sub disconnect {
    my $self = shift;
    my ( $ldap ) = @_;

    if ( $ldap ) {
	my $mesg = $ldap->unbind;
    }
 ##   if blabla
}


sub readSchema {
    my $self = shift;
    my ( $location, $etcdir ) = @_;
    my $schema;
#    print "readschema: location: |$location| etcdir: |$etcdir|\n";
    if ( $location =~ /^file:\s*(.+)/ ) {
	my $file = $1;
	if ( ! -e $file ) {
	    $file =~ s/\.\/etc\///;
	    $file = "$etcdir/$file";
	}
	$schema = $self->readSchemafromFile($file);
    } elsif ( $location =~ m{^(ldaps?)://\s*(.+):(\d*)} ) {
	$schema = $self->readSchemafromServer($1,$2, $3);
    } elsif ( $location =~ m{^(ldaps?)://\s*(.+)} ) {
	$schema = $self->readSchemafromServer($1,$2,389);
    } else {
	if ( ! -e $location ) {
	    $location = "$etcdir/$location";
	}
	$schema = $self->readSchemafromFile($location);
    }
#    print "schema read\n";
    return ($schema);
}

sub readSchemafromFile {
    my $self = shift;
    my ( $filename ) = @_;
    my $schema;

    $schema = Net::LDAP::Schema->new;
    $schema->parse ( $filename ) or die $schema->error;
    $self->{schema} = $schema;
    return ($schema);
}


sub superclasses {
    my $self = shift;
    my ( $schema, $class ) = @_;
    my $superclass = $class;
    my @classes;

#    use Data::Dumper;
#    print "Schema: \n";
#    print Dumper($schema);

#    print "getting superclasses of *$class*\n";

    push @classes, $class;

#    print $schema->superclass($superclass);

    while ( ( $superclass ) = $schema->superclass ($superclass) ) {
	if ( $superclass ) {
#	    print "found super [$superclass]\n";
	    push @classes, $superclass;
	} else {
	    last;
	}
    }

    return (reverse @classes);

}


sub getmusts {
    my $self = shift;
    my ( $schema, @classes ) = @_;

    my @musts;

    foreach my $class (@classes) {
	foreach my $must ( $schema->must($class) ) {
	    push @musts , lc($$must{name});
	}
    }

    my %seen = ();
    my @uniq = grep { ! $seen{$_} ++} @musts;

    return(@uniq);
}

sub getmays {
    my $self = shift;
    my ( $schema, @classes ) = @_;

    my @mays;

    foreach my $class (@classes) {
	foreach my $may ( $schema->may($class) ) {
	    push @mays , lc($$may{name});
	}
    }

    my %seen = ();
    my @uniq = grep { ! $seen{$_} ++} @mays;

    return(@uniq);
}

sub checkSchema {
    my $self = shift;
    my ($schema, $mode, $rh_object) = @_;
#    my ($schema, $ra_classes, $ra_attributenames, $rh_attributes) = @_;


    # first check if all classes are known:
    my $class;
    my @known_classes = $schema->all_objectclasses;
    my @classes;
    my $is_not_ok = 0;


    foreach $class ( @known_classes ) {
#	print "Known class: $class->{name}\n";
	push @classes,  $class->{name};

#	use Data::Dumper;
#	print Dumper($class);
#	exit();
    }

    my @clean_classes;

    foreach $class ( @{$$rh_object{classes}} ) {
	if ( grep { lc($_) eq lc($class) } @classes ) {
	    push @clean_classes, $class;
	} else {
### todo log error
	    print " Error in schemacheck: unknown object class $class\n";
 	    if ( $mode > 3 ) {
		$is_not_ok++;
	    }
	}
    }

    $$rh_object{classes} = \@clean_classes;

    my @musts = $self->getmusts($schema, @{$$rh_object{classes}});
    my @mays = $self->getmays($schema, @{$$rh_object{classes}});

#    print "Entry has following classes: \n\t@{$$rh_object{classes}}\n";
#    print "  with following must attrs: \n\t@musts\n";
#    print "  with following may attrs: \n\t@mays\n";


    my $may_not_ok = 0;
    my $must_not_ok = 0;

    push @musts, $$rh_object{classinfo}{rdnattr};
#    print "musts: @musts\n";

    foreach my $must (@musts) {
	if ( lc($must) eq "objectclass" ) { next;}
	if ( ! $$rh_object{attributes}{$must} ) {
### todo log error
	    print "Error in schemacheck: missing must: $must\n";
	    if ( $mode <= 3 ) {
		$$rh_object{attributes}{$must}[0] =
		    "value added by dbconnector";
		push @{$$rh_object{attributenames}}, $must;
	    } else {
		$must_not_ok++;
	    }
	}
    }

    if ( $must_not_ok ) { $is_not_ok++; }

    push @mays, @musts;
    foreach my $attr ( keys %{$$rh_object{attributes}} ) {
	if ( ! grep { $attr eq $_ } @mays ) {
	    if ( $mode == 3 ) {
		delete $$rh_object{attributes}{$attr};
		print "Error in schemacheck: not allowed Attribute $attr ".
		    "deleted\n";
	    } else {
### todo log error
		print "Error in schemacheck: Attribute $attr not allowed\n";
		$may_not_ok++;
	    }
	}
    }

    if ( $may_not_ok ) {
	if ( $mode == 1 ) {
	    push @{$$rh_object{classes}}, "attrObject";
	} elsif ( $mode == 2 ) {
	    push @{$$rh_object{classes}}, "extensibleObject";
	} elsif ( $mode == 4 ) {
	    $is_not_ok++;
	}

    }

### TODO check attribute syntaxes

#    use Data::Dumper;
#    print Dumper($rh_object);

    return ($is_not_ok);

}


sub getrdn {
    my $self = shift;
    my ($dn) = @_;
    use Data::Dumper;

    use Net::LDAP::Util qw( ldap_explode_dn);
    my $ra_rdns = ldap_explode_dn($dn);
    my ($key, $val) = each %{@{$ra_rdns}[0]} ;
    if ( each %{@{$ra_rdns}[0]} ) {
	$logger->error("multivalued rdns not yet supported");
    }

#-     print "key: [$key] val: [$val]\n";
    my $rdn = "$key=$val";

    return ($rdn);
}


sub clean_errormessage {
    my $self = shift;
    my ($errormesg) =  @_;
# print "error: \n|$errormesg|\n";
    $errormesg =~ s/\x00//;
    chomp($errormesg);
    return($errormesg);
}


# Changes: TZ 2017-06-01 Add parameter lang. When set, attributes of that language
#                        are returned.
sub incrementValue {
    my $self = shift;
    my ($dn,$attr,$ldap, $control, $lang) =  @_;
    my $number = 0;

    my @args = (
	increment => { $attr => 1 },
	);

    if ( $control ) {
	 push @args, control => [ $control ];
    }

    my $mesg = $ldap->modify($dn, @args );
#			     increment => {
#				 $attr => 1
#			     },
#	);

    if ( $mesg->code() == Net::LDAP::LDAP_SUCCESS ) {
	$logger->debug("LDAP::incrementValue(): incremented $attr");
	my @searchargs = (
	    base => $dn,
	    scope => 'base',
	    attrs => [ $attr ],
	    filter => "(objectclass=*)",
	    );

	if ( $control ) {
	    push @searchargs, control => [ $control ];
	}

	$mesg = $self->ldap_search( $ldap, $lang, @searchargs );
#	    base => $dn,
#			       scope => 'base',
#			       attrs => [ $attr ],
#			       filter => "(objectclass=*)",
#			       control => $control,
#	    );

	if ( $mesg->code() == Net::LDAP::LDAP_SUCCESS ) {
	    if ( $mesg->count() ) {
		my $counterentry=$mesg->pop_entry();
		$number=$counterentry->get_value($attr);
	    }
	    else {
		$logger->error("LDAP::incrementValue(): Error could not search $dn ");
	    }
	}
	else {
	    $logger->error("LDAP::incrementValue(): Error in search at $dn: ".
			   $mesg->error().'('.$mesg->code.')!');
	}
    }
    else {
	$logger->error("LDAP::incrementValue(): could not increment $attr at $dn: ");
	$logger->error('LDAP::incrementValue(): Error '.
		       $mesg->error.'('.$mesg->code.')!');
    }
    $logger->debug("LDAP::incrementValue(): new value of $attr is: $number");

    return $number;
}




# -----------------------------------------------------------------------------
# Name:     moveDn
# Eingabe:  inConnection (Handler) / Ein LDAP-Verbindungs-Handler
#           inSourceDn (string) / Eine DN, die verschoben werden soll
#           inNewDn (string) / Das Ziel, wohin verschoben werden soll
#           inRecursive (boolean) / Rekursiver Aufruf der Funktion
# Funktion: Verschiebt einen DN in einen anderen Teilbaum, bei Angabe von
#           inRecursive auch rekursiv
# Ausgabe:  success (boolean) / Wurde der DN verschoben, dann 1 sonst 0
#
# Changes: PG: 2014-05-28:
#          additional (optional) parameter rh_ldapdef for respecting deref alias
#          MH: 2014-10-16:
#          Omit copyDn and use change_rdn instead, as copyDn excludes Unique Overlay
#          Note, $inRecursive is not respected, as non-recursive makes no sense here
#
# -----------------------------------------------------------------------------
sub moveDn( $$$$$ ) {

  # Eingabevariablen
  my ( $self, $inConnection, $inSourceDn, $inNewDn, $inRecursive,
      $rh_ldapdef ) = @_;

  my $entryName = $self->get_dn_part($inSourceDn,1,undef);

  my $mesg = $self->change_rdn ( $inConnection, $inSourceDn, $entryName, 'deleteoldrdn', $inNewDn, $rh_ldapdef );

  my $success;

  if ( $mesg->code ) {
      $logger->error("moveDn(): could not move $inSourceDn to $inNewDn.");
      $logger->error("LDAP Error: ".$mesg->error ." (".$mesg->code.")");
      $success = 0;
  } else {
      $logger->debug( "moveDn(): moved $inSourceDn to $inNewDn.");
      $success = 1;
  }

  return $success;

}

# -----------------------------------------------------------------------------
# Name:     copyDn
# Eingabe:  inConnection (Handler) / Ein LDAP-Verbindungs-Handler
#           inSourceDn (string) / Eine DN, die kopiert werden soll
#           inNewDn (string) / Das Ziel, wohin kopiert werden soll
#           inRecursive (boolean) / Rekursiver Aufruf der Funktion
# Funktion: Kopiert einen DN in einen anderen Teilbaum, bei Angabe von
#           inRecursive auch rekursiv
# Ausgabe:  success (boolean) / Wurde der DN kopiert, dann 1 sonst 0
#
# Changes: PG: 2014-05-28:
#          additional (optional) parameter rh_ldapdef for respecting deref alias
# ----------------------------------------------------------------------------
sub copyDn( $$$$$ ) {
    my $self = shift;
    # Eingabevariablen
    my ( $inConnection, $inSourceDn, $inNewDn, $inRecursive,
	 $rh_ldapdef ) = @_;


    # Verwendete Variablen
    my $success = 0;
    my $copy;
    my $search;
    my $filter;
    my $entry;
    my $entryName;
    my @tmp;
    my $result;

    # Filterregel
    $filter = "(objectClass=*)";


    # Den Eintrag aus dem Verzeichnis holen
    $result = $self->baseSearch( $inConnection, $inSourceDn, 'base',
				 $filter, $rh_ldapdef );

    if ( $result->code ) {
	&mylogLDAPerror("copyDN(): searching", $result);
    }
#  if( !$search || ($search->code() != 0) ) {
#    $logger->logdie( "addUser(): Der Eintrag " . $inSourceDn . " existiert nicht im Verzeichnis." );
#  }
    else {
	$entry = $result->entry( 0 );
#	@tmp = split( /,/, $inSourceDn );
#	$entryName = $tmp[0];
	$entryName = $self->get_dn_part($inSourceDn,1,undef);
	$logger->debug("copyDN(): working on rdn $entryName, newdn is $inNewDn");
	$entry->dn( $entryName . "," . $inNewDn );


	# Die eigentliche DN kopieren
	if ( $rh_ldapdef->{control} ) {
	    $logger->debug("copyDN(): found control");
	    $result = $inConnection->add( $entry,
					  control => $rh_ldapdef->{control} );
	}
	else {
	    $result = $inConnection->add( $entry );
	}
	if ( $result->code ) {
	    &mylogLDAPerror("copyDN: adding", $result);
	    $logger->error("Entry was: \n".Dumper($entry) );
	    return 0;
	}
#
#    if( !$copy ) {
#      $logger->logdie( "copyDn(): Es ist ein Fehler beim Kopieren der DN " . $inSourceDn . " aufgetreten." );
#
#    }
	else {
	    $logger->debug( "copyDn(): " . $inSourceDn . " wurde nach ".
			   $entry->dn(). " kopiert." );
	    $success = 1;
	}

	$logger->debug("are we recursive: $inRecursive");

	# Bei Rekursion zuerst alle Unterknoten loeschen!
	if( $inRecursive ) {
	    # Filterregel
	    $filter = "(objectClass=*)";

	    # Suche ans LDAP-Verzeichnis senden
	    $result = $self->baseSearch( $inConnection, $inSourceDn, 'one',
					 $filter, $rh_ldapdef );

	    if ( $result->code ) {
		&mylogLDAPerror("copyDN() rekursive: searching", $result);
	    }
#      if( !$search ||($search->code() != 0) ) {
#
#        $logger->logdie( "copyDn(): Bei der Suche unterhalb der DN " . $inSourceDn . " ist ein Fehler aufgetreten." );
#
#      }
	    else {
		$success = 1;
		foreach $entry ( $result->entries ) {

		    $success = $success && $self->copyDn( $inConnection,
							  $entry->dn(),
							  $entryName . "," .
							  $inNewDn,
							  $inRecursive,
							  $rh_ldapdef,
			);
		}
	    }
	}
    }
    # Rueckgabe
    return $success;
}

##################################################################
# Name: baseSearch
#
# Description:
# Performs a base search and has a partial understanding of the rh_ldap-object
# which was introduced in this library later than some basic methods. Is used
# by new version of old method copyDN, moveDN, etc.
#
# Input:
#       $inConnection: LDAP connection object
#       $inSourceDN: BaseDN
#       $scope: search scope: 'base', 'one', or 'sub'
#       $filter: (optional) search filter, defaults to (objectclass = *)
#       $rh_ldapdef: (optional) object defining ldap parameters
#
# Output:
#       $result: search result
#
# Created: PG 2014-05-28
# Changes: TZ 2017-06-01 Add parameter lang. When set, attributes of that language
#                        are returned.
#
######################################################################
sub baseSearch {
    my $self = shift;
    my ( $inConnection, $inSourceDn, $scope, $filter, $rh_ldapdef ) = @_;
    my $result;
    my $lang;

    $filter = "(objectclass=*)" if ! $filter;

    my @args = (
        base      => $inSourceDn,
        scope     => $scope,
        filter    => $filter,
    );
    if ( $rh_ldapdef ) {
	my $deref = $$rh_ldapdef{deref};
	$logger->debug("baseSearch(): searching with deref=$deref");

        push @args, deref => [ $deref ];
#					   sizelimit => $$rh_ldapdef{sizelimit},
#					   timelimit => $$rh_ldapdef{timelimit},
        $lang = $$rh_ldapdef{lang};
    }

    $result = $self->ldap_search( $inConnection, $lang, @args );

    return $result;
}


# -----------------------------------------------------------------------------
# Name:     deleteDn
# Eingabe:  inConnection (Handler) / Ein LDAP-Verbindungs-Handler
#           inDn (string) / Ein DN
#           inRecursive (boolean) / Rekursiver Aufruf der Funktion
# Funktion: Loescht einen DN, bei Angabe von inRecursive auch rekursiv
# Ausgabe:  success (boolean) / Wurde der DN geloescht, dann 1 sonst 0
# -----------------------------------------------------------------------------
sub deleteDn( $$$$ ) {
    my  $self = shift;
    # Eingabevariablen
    my ( $inConnection, $inDn, $inRecursive, $rh_ldapdef ) = @_;


  # Verwendete Variablen
  my $success = 0;
  my $delete;
  my $search;
  my $filter;
  my $entry;
  my $result;

  $logger->debug("In LDAP::deleteDn, dn is $inDn");

  # Bei Rekursion zuerst alle Unterknoten loeschen!
  if( $inRecursive ) {
      $logger->debug("deleteDN() we are in recursive mode here");
      # Filterregel
      $filter = "(objectClass=*)";


      # Suche ans LDAP-Verzeichnis senden
      $result = $self->baseSearch( $inConnection, $inDn, 'one',
				   $filter, $rh_ldapdef );
#    $result = $inConnection->search( ( scope => "one",
#                                       base => $inDn,
#                                       filter => $filter ) );
      if ( $result->code ) {
	  &mylogLDAPerror("deleteDN(): searching", $result);
      }
#    if( $search->code() != 0 ) {
#
#      $logger->logdie( "deleteDn(): Bei der Suche unterhalb der DN " . $inDn . " ist ein Fehler aufgetreten." );
#
#    }
      else {
	  $logger->debug('deleteDN(): number of sub-entries to be deleted first: ' . $result->count() );
	  foreach $entry ( $result->entries ) {
	      $self->deleteDn( $inConnection, $entry->dn(), $inRecursive, $rh_ldapdef );
	  }
      }
  }


  # Die eigentliche DN loeschen
  my @args;
  my $control = $rh_ldapdef->{control};

  if ( $control ) {
      push @args, control => [ $control ];
  }
  $result = $inConnection->delete( $inDn, @args );

  # Fehlerbehandlung
  if ( $result->code ) {
      &mylogLDAPerror("deleteDN(): deleting", $result);
  }
#  if( !$delete || ($delete->code() != 0) ) {
#
#    $logger->logdie( "deleteDn(): Bei der Loeschung der DN " . $inDn . " ist ein Fehler aufgetreten." );
#
#  }
  else {
    $logger->debug( "deleteDn(): successfully deleted DN " . $inDn );
    $success = 1;
  }


  # Rueckgabe
  return $success;

}

sub searchValueFromLDAP {
    my $self = shift;
    my ( $attr, $mapvalue, $mapattr, $uri, $searchtype, $pwtoken, $ldap,
	 $rh_ldapserver, $is_multi ) = @_;

#    print "*** attr: $attr, is_multi: $is_multi\n";
    my $cmp = '';

    if ( ! $ldap ) {
	## we must create Session (and end it when finished)
	## using $uri and $pwtoken
    }


    my $attrval ='';

    if ( ! $self->{entries_found}{$mapvalue} ) {

        ## we could evaluate searchtype here for using substring match etc.
	my $filter = "($mapattr = $mapvalue)";
	$rh_ldapserver->{filter} = $filter;

	my $mesg = $self->doSearch($ldap, $rh_ldapserver);

	my $entry = undef;
	if ($mesg->code == Net::LDAP::LDAP_SUCCESS && $mesg->entries == 1) {

	    $logger->debug("found exactly one entry");
	    $entry = $mesg->pop_entry();
	    if ( ! $is_multi ) {
		$attrval = $entry->get_value( $attr );
#		print "NONMULTI: $attrval\n";
	    }
	    else {
		my @attrvals = $entry->get_value( $attr );
		$attrval = join(' $ ', @attrvals);
#		print "MULTIVALUES: @attrvals\n";
	    }
	    $self->{entries_found}{$mapvalue} = $entry;
	} else {
	    $logger->warn("Entry with search $filter not found");
	}
    }
    else {
	$logger->debug("found entry already stored in Object");
	if ( ! $is_multi ) {
	    $attrval = $self->{entries_found}{$mapvalue}->get_value( $attr );
	}
	else {
	    my @attrvals =
		$self->{entries_found}{$mapvalue}->get_value( $attr );
	    $attrval = join(' $ ', @attrvals);
	}
    }

    if ( $attrval ) {
	$logger->debug("found $attr value $attrval");
    }
    else {
	$logger->warn("found no value in attr $attr");
    }



    return ( $attrval);

}

sub getParentDn_better {

    my $self = shift;
    my ($dn) = @_;


    my $ra_dns = ldap_explode_dn( $dn );


    shift  @{$ra_dns};
#    pop  @{$ra_dns};
    my $parentdn='';
    foreach my $dnpart (  @{$ra_dns} ) {
	my $key = keys  %{$dnpart};
	my $val = values %{$dnpart};

	$parentdn .= "$key=$val,";
    }

    $parentdn =~ s/,$//;
    $logger->debug("parent dn: $parentdn");

    return $parentdn;
}

sub getParentDn {
    my $self = shift;
    my ( $dn ) = @_;
    $logger->debug("getting parent dn of $dn");
### Todo us explodeDN to make this safe!!!
    $dn =~ s/^.+?,//;
    $logger->debug("got parent dn $dn");

    return($dn);

}

sub insertFilterInUri {
    my $self = shift;
    my ( $uri, $filter ) = @_;

    $uri =~ s/^([^?]*\?\?\?)/$1$filter/;

    return $uri;
}

sub mylogLDAPerror
 {
   my ($from, $mesg) = @_;
   $logger->error("Error while $from: Return code: ", $mesg->code);
   $logger->error("\tMessage: ". $mesg->error_name . " :". $mesg->error_text
		  . "MessageID: ". $mesg->mesg_id . "\tDN: ". $mesg->dn);
 }


sub createldaptimefilter {
    my $self = shift;
    my ( $mode,$attr, $timestring ) = @_;
    my $filter='';
    my $operator ='';

#    my ($sec,$min,$hours,$mday,$mon,$year) = localtime();
    my $now = time();

#    print "XXXXXX mode: |$mode| attr: |$attr| timestr: |$timestring|\n";

    if ( $mode eq "OLDER_THAN" ) {
	$operator = '<=';
    }
    elsif ( $mode eq "NEWER_THAN" ) {
	$operator = '>=';
    }
    else {
	print "### unsupported mode |$mode| in createldaptimefilter\n";
	$logger->error("unsupported mode |$mode| in createldaptimefilter");
	return '';
    }

    my ($prefix,$number,$type) = ($timestring =~ /(-?)(\d+)([YMDhms])/);

    my $seconds;

    if ( $type ) {
#    print "number: |$number| type: |$type|\n";

	if ( $type eq 's' ) {
	    $seconds = $number;
	}
	elsif ( $type eq 'm' ) {
	    $seconds = $number * 60;
	}
	elsif ( $type eq 'h' ) {
	    $seconds = $number * 60 * 60;
	}
	elsif ( $type eq 'D' ) {
	    $seconds = $number * 60 * 60 * 24;
	}
	elsif ( $type eq 'M' ) {
	    $seconds = $number * 60 * 60 * 24 * 30;
	}
	elsif ( $type eq 'Y' ) {
	    $seconds = $number * 60 * 60 * 24 * 365;
	}
    }
    else {
	print "### Wrong syntax in time abreviation allowed are only a number followed by 's' 'm' 'h' 'D' 'M' or 'Y'\n";
	$logger->error("Wrong syntax in time abreviation");
	return '';
    }

    my $past;
    if ($prefix eq '-') {
	$past = $now + $seconds;
    } else {
	$past = $now - $seconds;
    }
    my ($nsec,$nmin,$nhours,$nmday,$nmon,$nyear) = gmtime();

    my ($sec,$min,$hours,$mday,$mon,$year) = gmtime($past);
    my $ntime = sprintf("%04d%02d%02d%02d%02d%02dZ",
		      ($nyear+1900),($nmon+1),
		      $nmday,$nhours,$nmin,$nsec);
    my $time = sprintf("%04d%02d%02d%02d%02d%02dZ",
		      ($year+1900),($mon+1),
		      $mday,$hours,$min,$sec);

#    print "now is   : $ntime\n";
#    print "searching: $time\n";

    $filter = "($attr$operator$time)";

    return $filter;
}

sub regexp2ldapfilter {
    my $self = shift;
    my ( $regexp, $attr ) = @_;
    my $filter='';

#    if ( $regexp =~ m|^s/(.+)/.*| ) {
#	$regexp = $1;
#    }


    my $ismulti=0;
    if ( $regexp =~ /\#;/ || $attr =~ /;/ ) {
	$ismulti = 1;
    }

    foreach my $attrib ( split /;/, $attr ) {
	foreach my $re ( split /\#;/, $regexp ) {

	    $re =~ s|(\^?[^\.\*\+\?\(]+).*|$1|;
	    $re =~ s|\\x|\\|g;
	    $re = $self->quoteldapfilter($re);
	    my $part = "*$re*";
	    $part =~ s/^\*\^//;
	    $filter .=  "( $attrib=$part) ";
	}

    }

    if ( $ismulti) {
	$filter = '(| '.$filter.')';
    }
    return $filter;

}

sub quoteldapfilter {
    my $self = shift;
    my ( $string ) = @_;

    $string =~ s/ /\\20/g;
    $string =~ s/>/\\3e/g;
    $string =~ s/</\\3c/g;

    return $string;
}


##################################################################
# Name: get_friendly_dn
#
# Description:
# converts a dn to more readable string by excluding the attribute names
# in the RDNs
#
# Input:
# 1.) $dn: DN to convert
# 2.) $num: (optional) Number of parts that should be returned (counted
#           from left). If number is negative it will cut of that number
#           from the DN (counted from right)
# 3.) $replace: (Optional) String that should be added at the right (replacing
#               the left out RDN parts
# 4.) $sep: (optional) Separator between the RDN Values (Default is: ', ' )
#
# Output:
# converted DN
#
# Created: PG 2013-06-27
# Changes: PG 2013-11-26 added possibility of negative part number
#
######################################################################
sub get_friendly_dn {
    my $self = shift;
    my ( $dn, $num, $replace, $sep ) = @_;

    if ( ! $dn ) {
	$logger->error("get_friendly_dn(): called without dn!!");
	return undef;
    }
    else {
	$logger->debug("get_friendly_dn(): dn: $dn ");
    }
    my $ra_dnparts = ldap_explode_dn($dn);
    my $returndn;
    my $i = $#$ra_dnparts+1;
    $logger->debug("get_friendly_dn(): number of parts $i ");
    $sep = ', ' if ! $sep;

    foreach my $rh_partdef ( @{$ra_dnparts} ) {
	my $part;
	if ( $num && $num > 0 && $i <= $num ) {
	    last;
	}
	elsif ($num && $num < 0 && $i + $num <= 0) {
	    last;
	}
	$i--;
	foreach my $key ( keys %{$rh_partdef} ) {
#	    $logger->debug("get_friendly_dn(): key: $key");
	    $part .='-' if $part;
	    my $val = $rh_partdef->{$key};
#	    $logger->debug("get_friendly_dn(): val: $val");
	    $part .= "$val";
	}
#	$logger->debug("get_friendly_dn(): working on part $part");
	$returndn.=$sep if $returndn;
	$returndn.=$part;
    }

    if ( $replace ) {
	$returndn.=$sep.$replace;
    }

    $logger->debug("get_friendly_dn(): returning: $returndn") if $returndn;

    return $returndn;
}

##################################################################
# Name: cut_dn
#
# Description:
# cuts a dn from left to right
#
# Input:
# 1.) $dn: DN to convert
# 2.) $num: Number of parts that should be cut way (counted from left)
#
# Output:
# new DN (which is a parent dn to the input dn)
#
# Created: PG 2013-06-27
# Changes:
#
######################################################################
sub cut_dn {
    my $self = shift;
    my ( $dn, $num ) = @_;
    if ( ! $dn ) {
	$logger->error("cut_dn(): called without dn!!");
	return undef;
    }
    else {
	$logger->debug("cut_dn(): dn: $dn ");
    }
    if ( ! defined $num || $num == 0 ) {
	$logger->error("cut_dn(): called without number!");
	return undef;
    }
    else {
	$logger->debug("cut_dn(): number: $num ");
    }


    my $ra_dnparts = ldap_explode_dn($dn);
    my $returndn;
    my $i = $num;
    foreach my $rh_partdef ( @{$ra_dnparts} ) {
	my $part;
	if ( $i ) {
	    $i--;
	    next;
	}
	foreach my $key ( keys %{$rh_partdef} ) {
#	    $logger->debug("cut_dn(): key: $key");
	    $part .='+' if $part;
	    my $val = $rh_partdef->{$key};
#	    $logger->debug("cut_dn(): val: $val");
	    $part .= "$key=$val";
	}
#	$logger->debug("cut_dn(): working on part $part");
	$returndn.=',' if $returndn;
	$returndn.=$part;
    }
    $logger->debug("cut_dn(): returning: $returndn");

    return $returndn;

}

##################################################################
# Name: get_dn_part
#
# Description:
# gives back one RDN-Part of a DN
#
# Input:
# 1.) $dn: DN to convert
# 2.) $num: Number of the part that should be returned (counted from left)
# 3.) $isfriendly: (optional) If this flag is set only the RDN attribute
#                  value is returned
#
# Output:
# the RDN
#
# Created: PG 2013-06-27
# Changes:
#
######################################################################

sub get_dn_part {
    my $self = shift;
    my ( $dn, $num, $isfriendly ) = @_;

    if ( ! $dn ) {
	$logger->error("get_dn_part(): called without dn!!");
	return undef;
    }
    else {
	$logger->debug("get_dn_part(): dn: $dn ");
    }
    if ( ! defined $num || $num == 0 ) {
	$logger->error("get_dn_part(): called without number!");
	return undef;
    }
    else {
	$logger->debug("get_dn_part(): number: $num ");
    }

    my $ra_dnparts = ldap_explode_dn($dn);
    my $part;
    my $c = $#$ra_dnparts+1;
    my $i = 1;
    $logger->debug("get_dn_part(): dn: $dn ");
    $logger->debug("get_dn_part(): number of parts $c ");

    foreach my $rh_partdef ( @{$ra_dnparts} ) {

	if ( $i == $num ) {
	    foreach my $key ( keys %{$rh_partdef} ) {
		$part .='-' if $part;
		my $val = $rh_partdef->{$key};
		if ( $isfriendly) {
		    $part .= "$val";
		}
		else {
		    $part .= "$key=$val";
		}
	    }
	    last;
	}
	$i++;
    }

    $logger->debug("get_dn_part(): returning: $part");

    return $part;
}

##################################################################
# Name: count_dn_parts
#
# Description:
# counts the number of RDNs in a DN
#
# Input:
# 1.) $dn: DN to count
#
# Output:
# Number of RDNs
#
# Created: PG 2013-06-27
# Changes:
#
######################################################################
sub count_dn_parts {
    my $self = shift;
    my ( $dn ) = @_;

    if ( ! $dn ) {
	$logger->error("count_dn_parts(): called without dn!!");
	return undef;
    }
    else {
	$logger->debug("count_dn_parts(): dn: $dn ");
    }
    my $ra_dnparts = ldap_explode_dn($dn);
    my $i = $#$ra_dnparts+1;

    return $i;
}

##################################################################
# Name: has_subordinates
#
# Description:
# finds out if entry is leaf or nonleaf object
#
# Input:
# 1.) $entry: a Net::LDAP::Entry object (including operational attributes)
# 2.) $ldap: a Net::LDAP Connection object
# Output:
#     $isnonleaf: 1 if entry is non leaf and 0 if leaf
#
#
# Created: PG 2013-07-05
# Changes:
#
######################################################################
sub has_subordinates {
    my $self = shift;
    my ( $entry, $ldap ) = @_;

    if ( ! $entry ) {
	$logger->error("has_subordinates(): called without entry!!");
	return undef;
    }
    else {
	$logger->debug("has_subordinates(): inspecting entry: ".$entry->dn());
    }
    my $isnonleaf=undef;

    my $hasoperationalattrs = $self->has_operational_attrs($entry);

    if ( ! $hasoperationalattrs ) {
	$entry = $self->reload_entry_with_operationals($entry,$ldap)
    }

    my $hassubordinates = $entry->get_value('hassubordinates');
    if ( $hassubordinates ) {
	if ( $hassubordinates eq 'TRUE' ) {
	    $isnonleaf = 1;
	}
	else {
	    $isnonleaf = 0;
	}
    }
    else {

	$logger->debug("has_subordinates(): no value for attribute hassubordinates found in entry: ".$entry->dn());
	$logger->debug("has_subordinates(): will search for children");
	my @entries = $self->get_children($entry,$ldap);
	if ( $entries[0] ) {
	    $isnonleaf = 1;
	}
	else {
	    $isnonleaf = 0;
	}
    }


    return $isnonleaf;
}


##################################################################
# Name: change_rdn
#
# Description:
# Changes the rdn of an entry
#
# Input:
# 1.) $ldap: a Net::LDAP connection object
# 2.) $dn: dn of the entry to be renamed
# 3.) $newrdn: New RDN
# 4.) $isdeleteoldrdn: (optional)  Flag if set the old rdnvalue will be deleted
# 5.) $newsuperior: (optional) new basedn whero to move the entry
# :
# Output:
#     $mesg: Net::LDAP::Message object returned by moddn
#
#
# Created: PG 2013-07-05
# Changes:
#
######################################################################
sub change_rdn {
    my $self = shift;
    my ( $ldap, $dn, $newrdn, $isdeleteoldrdn, $newsuperior, $rh_ldapdef ) = @_;

    if ( ! $dn ) {
	$logger->error("change_rdn(): called without dn!!");
	return undef;
    }
    else {
	$logger->debug("change_rdn(): called with dn: $dn");
    }
    my @args = (
	newrdn => $newrdn,
	deleteoldrdn => $isdeleteoldrdn,
	newsuperior => $newsuperior,
	);
    my $control = $rh_ldapdef->{control};

    if ( $control ) {
	 push @args, control => [ $control ];
    }

    my $mesg = $ldap->moddn ( $dn, @args);
#			      newrdn => $newrdn,
#			      deleteoldrdn => $isdeleteoldrdn,
#			      newsuperior => $newsuperior,
#			      control => $control,
#	);

    return $mesg;
}


##################################################################
# Name: has_operational_attrs
#
# Description:
# finds out if the entry object has operational attributes
#
# Input:
# 1.) $entry: a Net::LDAP::Entry object
#
# Output:
#     $has_operational: 1 if entry has operational attributes and 0 if not
#
#
# Created: PG 2013-07-05
# Changes:
#
######################################################################
sub has_operational_attrs {
    my $self = shift;
    my ( $entry ) = @_;
    my $has_operational;

    ### this operational attribute is mandatory:
    my $modifyersname = $entry->get_value('modifiersName');

    if ( $modifyersname ) {
	$has_operational = 1;
    }
    else {
	$has_operational = 0;
    }

    return $has_operational;

}

##################################################################
# Name: reload_entry_with_operationals
#
# Description:
# reloads an entry with all operational attributes
#
# Input:
# 1.) $entry: a Net::LDAP::Entry object
# 2.) $ldap: a Net::LDAP Connection object
#
# Output:
#     $newentry:  a Net::LDAP::Entry object with operational attrs
#
#
# Created: PG 2013-07-05
# Changes: TZ 2017-06-01 Add parameter lang. When set, attributes of that language
#                        are returned.
#
######################################################################
sub reload_entry_with_operationals {
    my $self = shift;
    my ( $entry, $ldap, $control, $lang ) = @_;

    $logger->debug("reload_entry_with_operationals() started");

    my $newentry;
    my $attrs = ['*','+'];
    my $basedn = $entry->dn();
    my $scope = 'base';
    my $filter = '(objectclass=*)';

    my @args = (
	base      => $basedn,
	scope     => $scope,
	filter    => $filter,
	attrs     => $attrs,
	);
    if ( $control ) {
	push @args, control => [ $control ];
    }

    my $mesg = $self->ldap_search( $ldap, $lang, @args );

#	base      => $basedn,
#	scope     => $scope,
#	filter    => $filter,
#	attrs     => $attrs,
#	control   => $control,
#	);
    $newentry = $mesg->pop_entry;

    return $newentry;

}

##################################################################
# Name: get_children
#
# Description:
# searches children of an entry
#
# Input:
# 1.) $entry: a Net::LDAP::Entry object
# 2.) $ldap: a Net::LDAP Connection object
#
# Output:
#     @children:  array of Net::LDAP::Entry objects
#
#
# Created: PG 2013-07-05
# Changes: TZ 2017-06-01 Add parameter lang. When set, attributes of that language
#                        are returned.
#
######################################################################
sub get_children {
    my $self = shift;
    my ( $entry, $ldap, $control, $lang ) = @_;

    $logger->debug("get_children(): started");
    my @children;
    my $attrs = ['*','+'];
    my $basedn = $entry->dn();
    my $scope = 'one';
    my $filter = '(objectclass=*)';
    my @args = (
	base      => $basedn,
	scope     => $scope,
	filter    => $filter,
	attrs     => $attrs,
	);
    if ( $control ) {
	 push @args, control => [ $control ];
    }

    my $mesg = $self->ldap_search( $ldap, $lang, @args );
#	base      => $basedn,
#	scope     => $scope,
#	filter    => $filter,
#	attrs     => $attrs,
#	control   => $control,
#	);
    @children = $mesg->entries();

    return @children;
}

##################################################################
# Name: create_alias
#
# Description:
# creates an aliasobject
#
# Input:
# 1.) $aliasobjectdn: DN of the alias object
# 2.) $aliasdn: DN of the entry pointed to by alias object
#
# Output:
#     $error: Error message
#
#
#
# Created: PG 2013-08-14
# Changes:
#
######################################################################
sub create_alias {
    my $self = shift;
    my ( $aliasobjectdn, $aliasdn, $ldap ) = @_;
    my $error='';
    $logger->debug("create_alias() started");
    my $entry = Net::LDAP::Entry->new;
    my $rdn = $self->get_dn_part($aliasobjectdn,1,undef);
    my ($attr, $val) = split /=/, $rdn;
    $entry->dn($aliasobjectdn);
    $entry->add( 'objectClass' => [ 'top', 'alias', 'extensibleObject' ] );
    $entry->add( $attr => $val );
    $entry->add( 'aliasedObjectName' => $aliasdn );

    my $mesg = $entry->update($ldap);
    if ( $mesg->code != Net::LDAP::LDAP_SUCCESS ) {
	$error = "Fehler beim Anlegen des Alias: ".$mesg->error;
	$logger->debug("create_alias(): error while creating aliasentry $aliasobjectdn: ".$mesg->error);
    }
    else {
	$logger->debug("create_alias(): created aliasentry $aliasobjectdn");
    }

    return $error;

}


##################################################################
# Name: clear_entry_attribute
#
# Description:
# deletes temporary attributes from a local ldap entry structure
#
# Input:
# 1.) $entry: Net::LDAP::Entry object
# 2.) $attribute: name of the temporary attribute
#
# Output:
#     no output just changes the entry object
#
#
# Created: PG 2013-08-17
# Changes:
#
######################################################################
sub clear_entry_attribute {
    my $self = shift;
    my ( $entry, $attr  ) = @_;
    $logger->debug("clear_entry_attribute(): clearing attr: $attr");
#    print "clear_entry_attribute()\n";
    my $count = 0;
    foreach my $change ( @{$entry->{'changes'}} ) {
#	$logger->debug("clear_entry_attribute(): change: ".Dumper($change));
	if ( $change eq 'delete' ) {
#	    print "clear_entry_attribute(): change: "
#		.Dumper($change)."\n";
#	    print "clear_entry_attribute(): deletes are: $entry->{'changes'}[$count+1]\n";
	    my $isdelete = 0;
	    foreach my $deleteattr ( @{$entry->{'changes'}[$count+1]} ) {
#		print "attr: $attr\n";
		if ( $attr eq $deleteattr ) {
		    delete $entry->{'changes'}[$count+1];
		    delete $entry->{'changes'}[$count];
		    $isdelete = 1;
		}
	    }
	}
	$count++;
    }

}





1;
