#######################################################################
# Copyright: DAASI International GmbH 2018. All rights reserved.
#
# This is Open Source Software
# License: Artistic License 2.0
#
# Author: Peter Gietz, DAASI International GmbH, www.daasi.de
# For questions please mail to peter.gietz@daasi.de
#######################################################################
#!/usr/bin/perl -w


package DAASIlib::Objects;

use strict;
use warnings;
use vars qw($VERSION);


use Log::Log4perl qw(:levels);

my $logger = Log::Log4perl->get_logger("main"); 

our $VERSION = "0.10";

use DAASIlib::Data;
my $data = new DAASIlib::Data;
use DAASIlib::LDAP;
my $libldap = new DAASIlib::LDAP;
use DAASIlib::DBI;
my $libdbi = new DAASIlib::DBI;

use Data::Dumper;
use DAASIlib::Convert;
my $conv = new DAASIlib::Convert;


sub new {
    my $class;
    my $self;

    my ( $this , $param ) = @_;
    $class = ref($this) || $this;

    $self={
	data => {},
    };

    bless($self , $class );

    $self->{data}{lastrecordnumber} = -1;

    return $self;

}


sub initLdapModifyObject {
    my $self = shift;
    my ($fpout, $keyattr, $ldap, $rh_ldapserver, $rh_tobedeleted, 
	$is_net_ldap_ldif, $is_verbose ) = @_;
    $logger->debug("Init LdapModifyObject");
    $self->{data}{'fpout'} = $fpout;
    $self->{data}{'ldap'} = $ldap;
    $self->{data}{'keyattr'} = $keyattr;
    $self->{data}{'ldapserver'} = $rh_ldapserver;
    $self->{data}{'tobedeleted'} = $rh_tobedeleted,
    $self->{data}{'is_net_ldap_ldif'} = $is_net_ldap_ldif ;
    $self->{data}{'is_verbose'} = $is_verbose ;

}

sub initLdapDeleteObject {
    my $self = shift;
    my ($fpout, $keyattr, $deletemode, $ldap, $rh_ldapserver,  
	$is_net_ldap_ldif, $is_verbose ) = @_;
    $logger->debug("Init LdapDeleteObject");
    $self->{data}{'fpout'} = $fpout;
    $self->{data}{'ldap'} = $ldap;
    $self->{data}{'keyattr'} = $keyattr;
    $self->{data}{'delmode'} = $deletemode;
    $self->{data}{'ldapserver'} = $rh_ldapserver;
    $self->{data}{'is_net_ldap_ldif'} = $is_net_ldap_ldif ;
    $self->{data}{'is_verbose'} = $is_verbose ;

}




sub doLdapModify {
    my $self = shift;
    my ($rh_outputobject, $rh_attrs) = @_;
    use Data::Dumper;
#    $self->{data}{'outputobject'} = $rh_outputobject;

    my $entrynumber = $rh_outputobject->{attributes}{$self->{data}{'keyattr'}}[0];

    $logger->debug("doLdapModify(): will modify entry with ",
		   $self->{data}{'keyattr'},
		   " = $entrynumber");

    my $newentry = $libldap->createEntryFromObject($rh_outputobject, '/^__/', 1);

    $logger->debug("doLdapModify(): created new LDAP entry: ",
		   $newentry->dn() );

#    my @nonoattrs = qw( hisuuid uid hisNameAtBirth userpassword sambaSID fhfCreationDate
#			uidNumber gidNumber fhfMailPublication hisMatriculationNumber
#			fhfDeactivated sambaLMPassword sambaNTPassword );
#
#    my @sourceattrs = $newentry->attributes;
#    @sourceattrs = $data->findNewValues(\@sourceattrs, \@nonoattrs, 1);

    ### todo make this configurable:
#    my @sourceattrs = qw( sn givenName cn displayName fhfCodeOfMembership fhfCodeOfMembershipText gecos );
    my @sourceattrs = ();
    if ( $rh_attrs ) {
	@sourceattrs = @{$rh_attrs};
    }

    $logger->debug("doLdapModify(): sourceattrs: |@sourceattrs|");

#    print "Newentry: ", $newentry->dn(), "\n";

    my $oldentry = $self->doFindOldEntry($entrynumber);

    $self->{is_changed} = 0;
    $self->{is_newentry} = 0;

    if ($oldentry) {
#    print "* Newentry: ", $newentry->dn(), "\n";
#	print Dumper($newentry);
#	print "found Old entry: ", $oldentry->dn(), "\n";
#	print Dumper($oldentry);
	$logger->debug("doLdapModify(): Found corresponding old entry: ", 
		       $oldentry->dn());
	$libldap->updateFromEntry($newentry, $oldentry, 
				  @sourceattrs);

#	print "modified Oldentry:\n";
#	print Dumper($oldentry);

	if ( @{$oldentry->{changes}} ) {	
	    $logger->debug("doLdapModify(): Found changes between old and new entry");
	    my $mesg = $oldentry->update( $self->{data}{'ldap'} );
	    if ( $mesg->code != Net::LDAP::LDAP_SUCCESS ) {
		$logger->error ( "doLdapModify(): ERROR in entry->update(): ", 
				 $mesg->error );
	    }
	    else {
		$logger->debug ("doLdapModify(): successful entry->update()");
		$self->{is_changed} = 1;
	    }
	}
    }
    else {
	$logger->debug("doLdapModify(): No old entry thus will add newentry");
	my $mesg = $newentry->update ( $self->{data}{'ldap'}  );
	if ( $mesg->code != Net::LDAP::LDAP_SUCCESS ) {
	    $logger->error ( "doLdapModify(): ERROR in entry->update(): ", 
			     $mesg->error );
	}
	else {
	    $logger->debug ("doLdapModify(): successful entry->update()");
	    $self->{is_changed} = 1;
	    $self->{is_newentry} = 1;
	}
	### do a simple add
    }

    return $entrynumber;

}

sub doLdapDeletion {
    my $self = shift;
    my ($key) = @_;
    use Data::Dumper;

    my $isdelete=0;
    my $action;
    my $deactivateattrs;

    if ( $self->{data}{'delmode'} eq 'delete' ) {
	$isdelete=1;
	$action = 'delete';
    }
    elsif ( $self->{data}{'delmode'} eq 'nodelete' ) {
	$action = '';
    }
    elsif ( $self->{data}{'delmode'} =~ /^deactivateattrs_(.*)$/ ) {
	$self->setdeactivateattrs($1);
	$action = 'deactivate';
	$logger->debug("boolattrs: @{$self->{data}{boolattrs}}");
	$logger->debug("stampattrs: @{$self->{data}{stampattrs}}");
    }
    else {
	$logger->error("unknown mode in doLdapDeletion");
    }


    if ( $action ) {
	my $oldentry = $self->doFindOldEntry($key);
	if ($oldentry) {
	    
	    my $dn = $oldentry->dn();
	    
	    if ($isdelete) {
		$libldap->deleteDn($dn, $self->{data}{'ldap'}, 0);
	    }
	    else {
		$self->deactivateEntry($oldentry)
		}
	    
	    if ( ! @{$oldentry->{changes}} ) {
		$logger->error("no changes made in entry to $action"); ;
	    }
	    my $mesg = $oldentry->update( $self->{data}{'ldap'} );
	    unless ( $mesg->code == Net::LDAP::LDAP_SUCCESS ) {
		$logger->error("Error in LDAP update of entry $dn: ".$mesg->error);
	    }
	}
	else {
	    $logger->error("doLdapDeletion tried to $action entry $key, although it dies not exist");
	}
    }
    else {
#	$logger->debug("Would have deleted entry with key if ldapdeletemode were not nodelete");
    }

    return;

}


sub deactivateEntry {
    my $self = shift;
    my ( $entry ) = @_;

    my $timestamp = $conv->currentDateAsGeneralizedTime();

    foreach ( @{$self->{data}{boolattrs}} ) {
	$entry->replace( $_ => 'TRUE' );
    }

    foreach (@{$self->{data}{stampattrs}} ) {
	$entry->replace( $_ => $timestamp );
    }

}

sub setdeactivateattrs {
    my $self = shift;
    my ( $deactivateattrs ) = @_;

    my @pairs = split /_/, $deactivateattrs;

    my @settotrueattrs;
    my @settimestampattrs;

    foreach (@pairs) {
	my ($bool,$stamp) = split /:/;
	push @settotrueattrs, $bool;
	if ( $stamp) {
	    push @settimestampattrs, $stamp;
	}
    }
	
    $self->{data}{boolattrs} = \@settotrueattrs;
    $self->{data}{stampattrs} = \@settimestampattrs;

    return;

}


sub doFindOldEntry {
    my $self = shift;
    my ( $entrynumber ) = @_;
 
    my $keyattr = $self->{data}{'keyattr'};
    my $oldentry = undef;
    my $filter = "( " .$self->{data}{'keyattr'} ."="
	.$entrynumber." )";
	$self->{data}{ldapserver}->{filter} = $filter;
	$oldentry = $libldap->sessionSearchEntry($self->{data}{ldap}, 
						 $self->{data}{ldapserver});
    return $oldentry;
} 

sub initHisObject {

    my $self = shift;
    my ($fpout, # $rh_output_object, 
	$ldap, $rh_ldapserver, 
	$ra_alreadyseen, $hisconnector_mode, $rh_hisattributes,
	$sqlinputuri, $sqlstatement, $is_net_ldap_ldif, $is_verbose ) = @_;

    $logger->debug("Init HISObject");
    $self->{data}{'fpout'} = $fpout;
#    $self->{data}{'outputobject'} = $rh_output_object;
    $self->{data}{'ldap'} = $ldap;
    $self->{data}{'ldapserver'} = $rh_ldapserver;
    $self->{data}{'alreadyseen'} = $ra_alreadyseen;
    $self->{data}{'hisconnector_mode'} = $hisconnector_mode;
    $self->{data}{'hisattributes'} = $rh_hisattributes;
    if ( $sqlinputuri ) {
	$self->{data}{'sqlinputuri'} = $sqlinputuri;
    }
    else {
	### just for Test purposes:
	$self->{data}{'sqlinputuri'} = 'dbi:CSV:f_dir=./csvdata/:autocommit=1'
    }
    $self->{data}{'sqlstatement'} = $sqlstatement;
    $self->{data}{'is_net_ldap_ldif'} = $is_net_ldap_ldif ;
    $self->{data}{'is_verbose'} = $is_verbose ;
}

sub doHisConnector {
    my $self = shift;
    my ( $rh_output_object) = @_;
    
    $self->{data}{'outputobject'} = $rh_output_object;


#    print "doHisConnector\n";
    my $errors = $self->doHisCheckConnectortype();

    if ( $errors ) {
	$logger->error("wrong his connector mode for this object");
	return();
    }

    if ( $self->{data}{hisconnector_mode} eq 'initperson' ) {
#	print "calling doHisConnectorPerson\n";
        $self->doHisConnectorPerson();
    } 
    elsif ( $self->{data}{hisconnector_mode} eq 'initrolle' ) {
#	print "calling doHisConnectorRolle\n";
	$self->doHisConnectorRolle();
    } 
    else {
#	print "calling doHisConnectorUpdater\n";
	$self->doHisConnectorUpdater();
    }

}

sub doHisConnectorPerson {
    my $self = shift;

    $logger->debug( "new His Person: ",
		    "P_id: $self->{data}{outputobject}->{attributes}{__hispid}[0], ",
		    "P_optype: $self->{data}{outputobject}->{attributes}{__hispoperationtype}[0], ",
		    "P_crtime: $self->{data}{outputobject}->{attributes}{__hispcreationtime}[0]");
    if ( $self->{data}{outputobject}->{attributes}{sn}[0] ) {
	$logger->debug( "  name: $self->{data}{outputobject}->{attributes}{sn}[0]" );
    }

    my $oldentry = $self->doHisFindOldEntry();

    my ($is_change, $newentry) 
	= $self->doHisCreateEntryFromObject($oldentry, 'meta_person');

    if ( ! $newentry) {
	$logger->error("doHisConnectorPerson: Couldn't create new "
		       ."meta person entry, thus returning");
	print "**** Couldn't create new meta person entry, thus returning\n"
	     if $self->{data}{'is_verbose'};
	return();
    }

    $self->doHisupdateEntry($is_change, $newentry, 'meta_person');

}



sub doHisConnectorRolle {
    my $self = shift;

    my $oldentry = $self->doHisFindOldEntry();

    if (! $oldentry ) {
	$logger->error("old entry not found while processing role attributes\n");
    }
    my ($is_change, $newentry) 
	= $self->doHisCreateEntryFromObject($oldentry, 'meta_rolle');

    if ( ! $newentry) {
	print "*** Couldn't create new entry, thus returning\n"
	     if $self->{data}{'is_verbose'};
	return();
    }

    $self->doHisupdateEntry($is_change, $newentry, 'meta_rolle');
}



sub doHisConnectorUpdater {
    my $self = shift;

    if (! grep { $self->{data}{outputobject}->{attributes}{__hispid}[0] eq $_} 
	 @{$self->{data}{alreadyseen}} ) {

	push @{$self->{data}{alreadyseen}}, 
	    $self->{data}{outputobject}->{attributes}{__hispid}[0];

	$self->doHisConnectorPerson();
    }
    # now comes the role attributes
    $self->doHisConnectorRolle();
}


sub enhanceHisEntry {
    my $self = shift;
    my ($newentry) = @_;

#    my $enhancedstatement = 
#    my $newcondition;
    my $matriculationnumber = $newentry->get_value('hismatriculationnumber');
    if (! $matriculationnumber) {
	$logger->error("No matriculationnumber in entry!!");
	print "No matriculationnumber in entry!!\n";
#	use Data::Dumper;
	$logger->debug( Dumper($newentry));
	return ($newentry);
    }
    elsif (length($matriculationnumber) < 7 ) {
	$logger->error("matriculationnumber in entry IS TO SHORT: $matriculationnumber !!");
	return ($newentry);
    }

    $matriculationnumber = substr($matriculationnumber,5);
    my $sqlstatement = 
	$self->amendHisSqlStatement($self->{data}{sqlstatement},
				    $matriculationnumber);


#    print "SQLstatement: $sqlstatement\n";
#    print "matriculationnumber: $matriculationnumber\n";

    my $result = $libdbi->performSearch($self->{data}{'sqlinputuri'},
					$sqlstatement,
					'inputuri');



#    use Data::Dumper;
#    print Dumper($result);
#    exit();

    return ($newentry);
}

sub amendHisSqlStatement {
    my $self = shift;

    my ($sqlstatement,$matriculationnumber ) = @_;

    my $newcondition = "p.studentnumber=$matriculationnumber ";


    if ( $sqlstatement =~ /\.createdtimestamp>/ ) {
	$sqlstatement =~ s/AND +p\.createdtimestamp>'.+?'/AND $newcondition/;
    }
    else {
	$sqlstatement = $libdbi->enhanceSQLStatement($sqlstatement,
						     $newcondition); 
    }

    if ( $sqlstatement =~ /ORDER BY .+/ ) {
	$sqlstatement =~ s/ORDER BY .+/ORDER BY p.createdtimestamp/;
    }
    else {
	$sqlstatement .= ' ORDER BY p.createdtimestamp';
    }


    return ($sqlstatement);

}

sub doHisupdateEntry {
    my $self = shift;
    my ($is_change, $newentry, $histable ) = @_;

#    my ($is_change, $newentry, $rh_output_object, $ldap, 
#	$is_net_ldap_ldif, $fpout, $histable, $rh_hisattributes ) = @_;

    $logger->debug("$histable back from doHisCreateEntryFromObject DN: ", 
		   $newentry->dn() ? $newentry->dn() : ' ', 
		   " is_change: $is_change");

    my $allmusts_exist = 1;
#    my $ra_mustattrs = $data->arrayToLower($self->{data}{hisattributes}->{mustattr});
    my $ra_mustattrs = $self->{data}{hisattributes}->{mustattr};
    my @entryattrs = $newentry->attributes ( nooptions => 1 );

     my @missingmustattrs = $data->findNewValues($ra_mustattrs,
						 \@entryattrs, 1);

    if ( $missingmustattrs[0] ) {
	my $gn = $newentry->get_value('givenName');
	my $sn = $newentry->get_value('sn');
	my $matr = $newentry->get_value('hisMatriculationNumber');
	my $dispn = $newentry->get_value('displayName');
	my $cos = $newentry->get_value('hisCourseOfStudy');
	my $fos = $newentry->get_value('hisFiledOfStudy');

	$logger->warn("error in entry: ",$newentry->dn(),
	"\n\t\tname: ", $gn ? $gn : '', " ", $sn ? $sn : '', " matr: ", 
		      $matr ? $matr : '');
	$logger->warn("Some Must attributes are missing: @missingmustattrs \nentry only contains @entryattrs");
	
#	        "\n\t\tdisplayname: $dispn, course: $cos, field: $fos\n";
#	print "Some Must attributes are missing: @missingmustattrs\n";
#	print "mustattrs: @{$ra_mustattrs}\n";
#	print "entryattrs: @entryattrs\n";

#	use Data::Dumper;
#	print "entry: \n\t", Dumper($newentry);

#	print "\n\n#################################\n\n";
#	print "outputobject: ", Dumper($self->{data}{'outputobject'}),"\n\n";
#	exit();

	if ( $self->{data}{hisconnector_mode} =~ /amend/ ) {
	    $logger->info("trying to amend entry");
#	    print "Trying to amend\n";
#	    exit();
	    $newentry = $self->enhanceHisEntry($newentry);
	    @entryattrs = $newentry->attributes ( nooptions => 1 );
	    @missingmustattrs = $data->findNewValues($ra_mustattrs,\@entryattrs);
	    if ( $missingmustattrs[0] ) {
		$logger->error("Some Must attributes are still missing after enhanceHisEntry: @missingmustattrs");
	    }
	    else {
		$logger->debug("Now all Must attributes are there");
	    }
	}
    }
    else {
	    $logger->debug("No Must attributes are missing");
#	print "no musts are missing\n";
    }

    if ($is_change > 0) {
	if ($is_change > 98) {
#	    print "ModRDN detected!!\n";
	    if ( $histable eq 'meta_rolle') {
		$logger->error("Unexpected RDN change in $histable attributes");
	    }
	    my $newrdnval 
		= $newentry->get_value($self->{data}{outputobject}->{classinfo}->{rdnattr});
	    my $newdn = "$self->{data}{outputobject}->{classinfo}->{rdnattr}="
		. $newrdnval
		. ",$self->{data}{outputobject}->{classinfo}->{basedn}";
	    my $olddn = $newentry->dn();
	    print "  * Changing RDN from $olddn to $newdn\n"
		if $self->{data}{'is_verbose'};
	    $logger->info("Changing RDN from $olddn to $newdn");  

	    $self->{data}{ldap}->delete( $newentry );
	    $newentry->dn( $newdn );
	    $newentry->changetype( "add" ); 
	}

	
	$logger->debug("Perfoming update on $histable attrs in entry",
		       $newentry->dn()? $newentry->dn() : ' ' );
	my $error = $libldap->updateEntry($newentry,$self->{data}{ldap});


#	print Dumper($self->{data}{hisattributes});

	if ( @{$self->{data}{hisattributes}->{checkvalue}}[0] ) {
	    $self->checkrules($newentry->dn());
	}
	### Workaround, that will be made obsolete by amend entry:
#	$self->checkdisplayname( $newentry->dn() );

#	if ( $error ) {
#	    print "error in ldap update: $error\n";
#	    use Data::Dumper;
#	    print Dumper($newentry);
#	    exit();
#	}

	if ( $self->{data}{fpout}) {
	    if ( $self->{data}{is_net_ldap_ldif} ) {
		$self->{data}{fpout}->write_entry($newentry);
	    }
	    else {
		$libldap->printEntry($newentry, $self->{data}{fpout});
	    }
	}
    } else { 
	if ( $is_change == 0) {
	    $logger->debug("No $histable attrs to change in entry ",
			   $newentry->dn()? $newentry->dn() : ' ');
	} 
	else {
	    $logger->error("Negative value in is_change means something went wrong");
	}
    }
}


sub checkrules {
    my $self = shift;
    my ($dn) = @_;

    my $ra_checkvalues = $self->{data}{hisattributes}->{checkvalue};
    $self->{data}{ldapserver}->{scope } = "base";
    $self->{data}{ldapserver}->{basedn} = $dn;
    $self->{data}{ldapserver}->{filter} = '';
    my $entry = $libldap->sessionSearchEntry($self->{data}{ldap}, 
					     $self->{data}{ldapserver});

    if ($entry) {
	foreach my $rule ( @{$ra_checkvalues} ) {
#	    print "Found Rule: $rule\n";
	    $self->checkrule($rule,$entry);
	}
    }
    else {
	$logger->error("couln't read entry while checking rules"); 
    }

#    exit;
    
}

sub checkrule {
    my $self = shift;
    my ($rule, $entry) = @_;

#    print "Checking rule $rule\n";

    my ($target, $ruleval) = split /=/, $rule;
    my $is_add = 0;
    if ( $target =~ s/\+// ) {
	$is_add =1;
    }

#    print "Target: $target, is_add: $is_add, ruleval: $ruleval\n";
    my $expectedval = '';

    foreach my $valpart (split /\./, $ruleval ) {
#	print "valpart: $valpart\n"; 
	if ( $valpart =~ s/\'//g ) {
	    $expectedval.= $valpart;
	}
	else {
	    my $attrval=$entry->get_value($valpart);
	    if ( $attrval ) {
		$expectedval.=$attrval;
	    }
	    else {
		$logger->error("couldn't retrieve Attribute $valpart, while checking rules");
	    }
	}
    }
    my $oldval = $entry->get_value($target);
#    print "newvalue for attr $target: $expectedval (oldvalue: $oldval)\n";
    if ( $oldval ne $expectedval ) {
	$logger->warn("$target ($oldval) not as expected ($expectedval), will amend that by ", $is_add ? "adding " : "replacing", "value");
	if ( $is_add ) {
	    $entry->add ( $target => $expectedval );
	}
	else {
	    $entry->replace ( $target => $expectedval );
	}
	my $error = $libldap->updateEntry($entry,$self->{data}{ldap});
	$logger->debug("Amendments done, error: $error ");
    }
    else {
	$logger->debug("No amendments to be done");
    }
}


sub doHisCreateEntryFromObject {
    my $self = shift;

    my ( $oldentry, $table) = @_;
#   my ( $rh_object, $oldentry, $rh_hisattributes, $hisconnector_mode, $table) = @_;

#    print "Outputobject at start of doHisCreateEntryFromObject\n"; 
#    print Dumper(%{$self->{data}{outputobject}->{attributes}});


    my $x = 0;

    my $is_change=0;
    my $entry;
    my $dn;
    my $is_old_entry= 0;
    my $is_eval_flags = 0;
    my %flaginfo;
    
    $logger->debug("doHisCreateEntryFromObject called with $self->{data}{hisconnector_mode} $table");
    
    if ( $oldentry ) {
	$entry = $oldentry->clone();
	$dn = $entry->dn();
	$is_old_entry= 1;
    }
    else {
	$entry = Net::LDAP::Entry->new;
	$dn = $libldap->createDnFromObject($self->{data}{outputobject});
	if (! $dn) { 
	    $logger->error("ERROR could not create DN in doHisCreateEntryFromObject");
	    $is_change = -102
	    }
	else {
	    $entry->dn($dn);
	    $entry->add( "objectClass" => $self->{data}{outputobject}->{classes} );
	    $is_change = 1;
	}
	$is_old_entry= 0;
    }
    
    
    if ( $self->{data}{hisconnector_mode} ne 'ignoreflags' ) {
	$is_eval_flags = 1;
	%flaginfo = $self->createHisFlagInfo();
##$rh_object, $rh_hisattributes);
    }
    
    my @activeattributes;
    
    my $is_personattrs = 0;
    my $his_changemode;
    
    my $rdnattr = $self->{data}{outputobject}->{classinfo}{rdnattr};
    my $is_rdn_change = 0;
    
    my %singleattrinfo  = $self->doHisCreateHashFromArray('singlevalue');
    my %neverchangeinfo = $self->doHisCreateHashFromArray('neverchange');
    my %noflagattrs = $self->doHisCreateHashFromArray('noflagattr');
#    my %neverchangeinfo = &doHisCreateHashFromArray($rh_hisattributes, 
#						    "neverchange");
    
    if ( $table eq "meta_person" ) {
	@activeattributes = @{$self->{data}{hisattributes}->{personattr}};
	$is_personattrs = 1;
	$his_changemode 
	    = $self->{data}{outputobject}->{attributes}{__hispoperationtype}[0];
    } 
    else {
	@activeattributes = @{$self->{data}{hisattributes}->{roleattr}};
	$his_changemode 
	    = $self->{data}{outputobject}->{attributes}{__hisroperationtype}[0];
    }

    push @activeattributes , @{$self->{data}{hisattributes}->{mustattr}};
    
    @activeattributes = $data->makeUnique( @activeattributes );
    
    if ( ! $his_changemode ) {
	$logger->error("couldn't find his operationtype in $table");
    }

    my $entry_changetype = $self->doHisGetEntryChangetype($his_changemode,
							  $is_old_entry,
							  $is_personattrs);
    
    if (  $entry_changetype) { 
	$entry->changetype($entry_changetype);
    }
    
#    print "Active Attrs: @activeattributes\n";
#    print "Outputobject at start of loop in doHisCreateEntryFromObject\n"; 
#    print Dumper(%{$self->{data}{outputobject}->{attributes}});
#    print "xxx postalCode: $self->{data}{outputobject}->{attributes}{postalCode}[0]\n";

#    exit();
    foreach my $attr ( @activeattributes ) {
#-	print "* checking attribute $attr\n";
	my $do_it=0;
	my @oldvalues=();
	my @newvalues=();
	my @isectvalues=();
	my $foundlowercase = 0;
	my $found = 0;
	my $is_caseignore = 0;

#	print "x: $x attr: $attr\n";
#	if ( $attr =~ /postalcode/i ) {
#	    print "Postalcode : $attr\n";

#	    print "normal: @{$self->{data}{outputobject}->{attributes}{$attr}}[0]\n";
#	    print "lowercase: @{$self->{data}{outputobject}->{attributes}{lc($attr)}}[0]\n";

#	}
	
	if ( @{$self->{data}{outputobject}->{attributes}{$attr}}[0] ) {
#	    print "found Attr\n";
#	    print "1xxx postalCode: $self->{data}{outputobject}->{attributes}{postalCode}[0]\n";

	    $found = 1;
	}
	if ( @{$self->{data}{outputobject}->{attributes}{lc($attr)}}[0] ) {
#	    print "found lc(Attr)\n";
#	    print "2yyy postalCode: $self->{data}{outputobject}->{attributes}{postalCode}[0]\n";

	    $is_caseignore= 1;
	    $foundlowercase = 1; 
	}
	
	if (  $found  || $foundlowercase ) {
#	    print "if found\n";
#	    print "3zzz postalCode: $self->{data}{outputobject}->{attributes}{postalCode}[0]\n";

#	    print "   $attr exists in object\n"; # if ( $attr =~ /postalcode/i );
	    if ( $is_old_entry ) {
		@oldvalues = $oldentry->get_value($attr);
		@newvalues = 
		    $data->findNewValues($self->{data}{outputobject}->{attributes}{$attr},
					 \@oldvalues, 0);
		@isectvalues = 
		    $data->findIntersection($self->{data}{outputobject}->{attributes}{$attr},
					    \@oldvalues, 0);
#-		print "   oldvalues: [@oldvalues]\n";
#-		print "   isectvalues: [@isectvalues]\n";
		
	    }
	    else {
		if ( $foundlowercase ) {
		    @newvalues = 
			@{$self->{data}{outputobject}->{attributes}{lc($attr)}};
		} 
		else {
		    @newvalues = 
			@{$self->{data}{outputobject}->{attributes}{$attr}};
		}
	    }

#	    print " before if newvalues: $self->{data}{outputobject}->{attributes}{postalCode}[0]\n"; # if ! $x;
#	    print "oldvalues: @oldvalues\nnew values: @newvalues\n";

	    if (@newvalues) {
#-		print "$attr:   newvalues: [@newvalues]\n";

		if ( lc($attr) eq lc($rdnattr) ) {
#-		    print "  this is a RDN attr!\n";
		    $is_rdn_change = 1;
		}

##		print "is_evalflags: |$is_eval_flags|\n";

		if ( $noflagattrs{$attr} ||
		     $noflagattrs{lc($attr)} ) {
#		    print "noflagattr $attr\n";
		     $do_it=1;
		}
		elsif ( $is_eval_flags ) {
#-		    print "   * evaluating flag of [$attr]: [$flaginfo{$attr}]\n";

#		    print Dumper %flaginfo;
#		    exit();
		    if ( $flaginfo{$attr} || $flaginfo{lc($attr)} ) { 
#			print "   Flag of $attr is set\n";
			$do_it=1; 
		    }
		    else {
			print "   Flag of $attr is NOT set\n";
		    }

		} else {
		    $do_it=1;
		}
	    }
##	    print "doit: |$do_it| isoldentry: |$is_old_entry| hischangemode: |$his_changemode|\n";
	    if ( ! $is_old_entry && $his_changemode eq "MOD" ) {
		$do_it = 1;
	    }
##print "doit: |$do_it|\n";	    
	    if ($do_it) {
##		print "   something to do changemod: $his_changemode\n" ;

##		print " $self->{data}{outputobject}->{attributes}{postalCode}[0]\n";

		if ( $his_changemode eq "ADD" ) {
#-		    print "   changetype ADD, is meta_person: $is_personattrs, is_old_entry: $is_old_entry\n"; 
		    if ( @oldvalues) {
			if ( ! $neverchangeinfo{$attr} && 
			     ! $neverchangeinfo{lc($attr)} ) {
			    $is_change++;

			    if ( $singleattrinfo{$attr} ||
				 $singleattrinfo{lc($attr)}) {
#-				print "  replacing   $attr => [@newvalues]\n";
				$entry->replace ( $attr => \@newvalues);
			    } 
			    else {
#-				print "  adding   $attr => [@newvalues]\n";
				$entry->add ( $attr => \@newvalues);
			    }
			    
			} 
			else {
			    $logger->debug("Didn't change $attr, because it should never be changed");
			}
		    } 
		    else {
#-			print "  adding   $attr => [@newvalues]\n";
			$is_change++;
			$entry->add ( $attr => \@newvalues);
		    }
		    
		}
		elsif ( $his_changemode eq "MOD" ) {
#-		    print "  changetype MOD, is meta_person: $is_personattrs, is_old_entry: $is_old_entry\n"; 
		    if ( @oldvalues) {
			if ( ! $neverchangeinfo{$attr} &&
			     ! $neverchangeinfo{lc($attr)}) {
			    $is_change++;
##			    if ( $singleattrinfo{$attr} ) {
##-				print "  replacing   $attr => [@newvalues]\n";
			    if ( $is_rdn_change ) {
#-				print "replacing RDN attr!!\n";
				$is_change += 100;
			    }
			    $entry->replace ( $attr => \@newvalues);
##			    } 
##			    else {
##-				print "  adding   $attr => [@newvalues]\n";
##				$entry->add ( $attr => \@newvalues);
#			    }
			} else {
			    $logger->debug("Didn't change $attr, because it should never be changed");
			}
		    } 
		    else {
			$is_change++;
#-			print "  adding   $attr => [@newvalues]\n";
			$entry->add ( $attr => \@newvalues);
		    }
		    
		}
		elsif ( $his_changemode eq "DEL" ) {
#-		    print "  changetype DEL meta_person: $is_personattrs, is_old_entry: $is_old_entry\n"; 
		    if ( @oldvalues ) {
			if ( ! $neverchangeinfo{$attr} &&
			     ! $neverchangeinfo{lc($attr)} ) {
#			    if ( $singleattrinfo{$attr} ) {
			    $is_change++;
#-			    print "  deleting   $attr => [@isectvalues]\n";
			    $entry->delete( $attr => \@isectvalues);
#			    } 
#			    else {
#				$entry->add ( $attr => \@newvalues);
#			    }
			}
			else {
			    $logger->warn("Didn't change $attr, because it should never be changed");
			    print " * Didn't change $attr, because it should never be changed\n"  if $self->{data}{'is_verbose'};
			}
		    } 
#		    else {
#			;
#####			$is_change--;
#		    }
		}
	    }
	}
	$x++;
#	print "At end of loop: $self->{data}{outputobject}->{attributes}{postalCode}[0]\n";
#	exit();
    }


#    print "Entry at the end of doHisCreateEntry (is_change is $is_change):\n";

#print "after loop: $self->{data}{outputobject}->{attributes}{postalCode}[0]\n";

#    print Dumper ($entry);
#    exit();


    if ( ! $entry ) {
	$logger->error("doHisCreateEntry couldn't create new entry: is_change: $is_change, hischangemode: $his_changemode, is_personattr: $is_personattrs, is_oldentry: $is_old_entry");
    }


    if ( ! $entry_changetype) { 
	$is_change = 0;
    }

    return ($is_change, $entry);
}


sub doHisCheckConnectortype {
    my $self = shift;

#    my ($rh_object, $hisconnector_mode) = @_;

    my $is_error = 0;

    if ( $self->{data}{hisconnector_mode} ne 'initrolle' ) {
	if ( !  $self->{data}{outputobject}->{attributes}{__hispid}[0] ) {
	    $logger->error("missing value of __his>p<id in HIS mode $self->{data}{hisconnector_mode}, thus returning");
	    $is_error++;
	}
	if ( ! $self->{data}{outputobject}->{attributes}{__hispoperationtype}[0] ) {
	    $logger->error("missing value of __his>p<operationtype in HIS mode $self->{data}{hisconnector_mode}, thus returning");
	    $is_error++;
	}
    }

    if ( $self->{data}{hisconnector_mode} ne 'initperson') {
	if ( !  $self->{data}{outputobject}->{attributes}{__hisrid}[0] ) {
	    $logger->error("missing value of __his>r<id in HIS mode $self->{data}{hisconnector_mode}, thus returning");
	    $is_error++;
	}
	if ( !  $self->{data}{outputobject}->{attributes}{__hisroperationtype}[0] ) {
	    $logger->error("missing value of __his>r<operationtype in HIS mode $$self->{data}{hisconnector_mode}, thus returning");
	    $is_error++;
	}

    }
	
    return ($is_error);
}

sub doHisGetEntryChangetype {
    my $self=shift;
    my ($his_changemode, $is_old_entry, $is_personattrs)= @_;

    my $entry_changetype = "";

    $logger->debug("doHisGetEntryChangetype called with his_changemod: [",
		   $his_changemode ? $his_changemode : "none", 
		   "], is_old_entry: [",
		   $is_old_entry ? $is_old_entry : "0", 
		   "], is_personattrs: [",
		   $is_personattrs ? $is_personattrs : "0", "]");

    if ( $his_changemode eq "ADD" ) {
	if ( $is_personattrs ) {
	    if ($is_old_entry ) {
		$logger->error("Entry already exists although ".
			       "Metaperson ADD mode");
#		$entry->changetype('modify');
		$entry_changetype = 'modify';
	    }
	    else {
		$entry_changetype = 'add';
	    }
	} 
	else {
	    $logger->warn("His Rolee entry with ADD mode changed to MOD mode");
	    $entry_changetype = 'modify';
	}
    }
    elsif ( $his_changemode eq "MOD" ) {
	if ( $is_personattrs ) {
	    if (! $is_old_entry ) {
		$logger->error("Entry does not exists although ".
			       "Metaperson MOD mode");
		$entry_changetype='add';
	    }
	    else {
#			    $entry->changetype = "add";
		$entry_changetype='modify';
	    }
	} 
	else {
	    $entry_changetype = 'modify';
	}
    }
    elsif ( $his_changemode eq "DEL" ) {
	if ( $is_personattrs ) {
	    if (! $is_old_entry ) {
		$logger->error("Entry does not exists although ".
			       "Metaperson DEL mode");
		$entry_changetype= '';
	    }
	    else {
		$entry_changetype = 'modify';
	    }
	} 
	else {
	    $entry_changetype = 'modify';
	}
    }

    return($entry_changetype);
}

sub doHisFindOldEntry {
    my $self = shift;

#    my ($rh_output_object, $rh_ldapserver, $ldap) = @_;

    my $oldentry = undef;
    my $mappingattr = $self->{data}{outputobject}->{attributes}{__hisMatchingAttribute}[0];

    if ( $mappingattr ) {
	my $filter = "( & (objectclass=hisPerson)"
	             . " ($mappingattr="
		     . $self->{data}{outputobject}->{attributes}{$mappingattr}[0]
		     . "))";
	$self->{data}{ldapserver}->{scope } = "sub";
	$self->{data}{ldapserver}->{basedn} = $self->{data}{outputobject}->{classinfo}{'basedn'};
	$self->{data}{ldapserver}->{filter} = $filter;
	$oldentry = $libldap->sessionSearchEntry($self->{data}{ldap}, 
						 $self->{data}{ldapserver});

    }
    return $oldentry;
}


sub doHisCreateHashFromArray {
    my $self = shift;

    my ($attrtype) = @_;
    my %info;

    foreach my $attr ( @{$self->{data}{hisattributes}->{$attrtype}}) {
	$info{$attr} = 1;
    }

    return(%info);
}


sub createHisFlagInfo {
    my $self=shift;

#    my ($rh_object, $rh_hisattributes) = @_;
    my %flaginfo;

#    use Data::Dumper;
#    print Dumper($rh_object);

    my $flagattrname;


    foreach my $attr (keys %{$self->{data}{outputobject}->{attributes}}) {
#	print "searching flag of attr: $attr\n";
	if ( $attr !~ /_flag$/ ) {
	    if ( $attr =~ /^__/ ) {
		$flagattrname = "${attr}_flag";
	    } else  {
		$flagattrname = "__${attr}_flag";
	    }
	    if ( $self->{data}{outputobject}->{attributes}{$flagattrname} ) {
		$flaginfo{$attr} = 1;
#		print "Found and set\n";
	    }
	}
#	print "$_: @{$rh_object->{attributes}{$_}}\n";

    }

#    my @mustattrs = @{$conf->{data}{hisattributes}{mustattr}};
    my @mustattrs = @{$self->{data}{hisattributes}->{mustattr}};
# ( 'uid', 'hisUuid', 'sn', 'cn', 'givenName', 'displayName', etc );
    foreach my $mustattr ( @mustattrs ) {
#	print "adding mustattr: $mustattr\n";
	$flaginfo{$mustattr} = 1;
    }
    


    return( %flaginfo);
}





sub readBiblInput {
    my $self = shift;
    my ($fp, $sep, $quote, $defaultclass) = @_; 

    my %object;   
    my ( $field, $i );
    my ( $inline);

    my $recordnumber=$self->{data}{lastrecordnumber};
    my $read_data = 0;
    my $current_record = 0;
    my $allreadyexists = 0;


    my $isstar =0;
    my $commentdata = '';

#    print "### Reading entry\n";

    while ($inline = <$fp>) {
#	print "readline: $inline\n";
	if ( eof($fp) ) {
	    print "encountered eof\n";
	    return (0);
	}
        chomp($inline);

	if ($inline =~ /^#/ ) {
	    next;
	}
	elsif ( $inline =~ /^\s*$/ ) {
	    if ( $read_data) {
#	    print "returning (inline is [$inline])\n";
		$object{class} = $defaultclass;
		if ( $isstar ) {
		    $commentdata = '* '.$commentdata;
		}

		if ($commentdata) {
		    $object{attributes}{x}[0]=$commentdata; 
		}
#	    use Data::Dumper;
#	    print Dumper(%object);
		$self->{data}{lastrecordnumber} = $current_record;
		last;
	    }
	    else {
		next;
	    }
	}

	$inline =~ s/\s*$//;

	if ( ! $read_data) {
	    if  ($inline =~ /^(\d+) *$/ ) {
		$current_record  = $1; 
	    }
	    elsif ($inline =~ /^(\d+)\*/ ) {
		$isstar = 1;
		$current_record  = $1;
	    }
	    elsif ($inline =~ /^(\d+)(.+)/ ) {
		$current_record  = $1;
		print "found comment [$2] in entry no. $current_record\n"; 
		$commentdata = $2;
	    }
	    if ( $current_record != $recordnumber ) {
#		print "new record number $current_record\n";
		if ($recordnumber && 
		    $current_record -1 != $recordnumber && 
		    $recordnumber != -1) {
		    print "current recordnumber $current_record is not consecutive ($recordnumber)\n";
		}
	    }
	    else {
		print "recordnumber $current_record already exists\n";
		$allreadyexists = 1;
	    }
	    if ($allreadyexists) {
		$object{attributes}{'num'}[0] = $current_record."a";
		$allreadyexists =0;
	    }
	    else {
		$object{attributes}{'num'}[0] = $current_record; 
	    }
	    $read_data = 1;
	}
	elsif ( $inline =~ /^$current_record (\w\w?):\w*$/ ) {
	    $read_data++;
#	     print " empty field \n";
	}
	elsif ( $inline =~ /^$current_record (\w\w?): ?(.*)/ ) {
	    $read_data++;
	    $object{attributes}{$1}[0]=$2; 
#	    print "data found: $1: $2\n";
	} 
	else {
	    print "wrongly formatted line [$inline] (recnum: $current_record)\n";
	}
    }

    return (\%object);

}


1;
